/* Generated by Cython 0.29.32 */

#ifndef __PYX_HAVE__pygame_sdl2__event
#define __PYX_HAVE__pygame_sdl2__event

#include "Python.h"

#ifndef __PYX_HAVE_API__pygame_sdl2__event

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#ifndef DL_IMPORT
  #define DL_IMPORT(_T) _T
#endif

__PYX_EXTERN_C PyObject *event_queue;

#endif /* !__PYX_HAVE_API__pygame_sdl2__event */

/* WARNING: the interface of the module init function changed in CPython 3.5. */
/* It now returns a PyModuleDef instance instead of a PyModule instance. */

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initevent(void);
#else
PyMODINIT_FUNC PyInit_event(void);
#endif

#endif /* !__PYX_HAVE__pygame_sdl2__event */
