/* Generated by Cython 0.29.32 */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_32"
#define CYTHON_HEX_VERSION 0x001D20F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_HEX >= 0x07030900)
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pygame_sdl2__rwobject
#define __PYX_HAVE_API__pygame_sdl2__rwobject
/* Early includes */
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <stddef.h>
#include "SDL.h"
#include <stdlib.h>
#include "python_threads.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "src/pygame_sdl2/rwobject.pyx",
  "stringsource",
  "type.pxd",
};
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()


/*--- Type declarations ---*/
struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl;
struct __pyx_opt_args_11pygame_sdl2_8rwobject_to_rwops;
struct __pyx_t_11pygame_sdl2_8rwobject_SubFile;
struct __pyx_t_11pygame_sdl2_8rwobject_BufFile;

/* "pygame_sdl2/rwobject.pxd":21
 * from sdl2 cimport SDL_RWops
 * 
 * cdef SDL_RWops *to_rwops(filelike, mode=*) except NULL             # <<<<<<<<<<<<<<
 * 
 */
struct __pyx_opt_args_11pygame_sdl2_8rwobject_to_rwops {
  int __pyx_n;
  PyObject *mode;
};

/* "pygame_sdl2/rwobject.pyx":133
 *     return 0
 * 
 * cdef struct SubFile:             # <<<<<<<<<<<<<<
 *     SDL_RWops *rw
 *     Sint64 base
 */
struct __pyx_t_11pygame_sdl2_8rwobject_SubFile {
  SDL_RWops *rw;
  Sint64 base;
  Sint64 length;
  Sint64 tell;
};

/* "pygame_sdl2/rwobject.pyx":188
 *     return 0
 * 
 * cdef struct BufFile:             # <<<<<<<<<<<<<<
 *     Py_buffer view
 *     Uint8 *base
 */
struct __pyx_t_11pygame_sdl2_8rwobject_BufFile {
  Py_buffer view;
  Uint8 *base;
  Uint8 *here;
  Uint8 *stop;
};

/* "pygame_sdl2/rwobject.pyx":381
 *     return to_rwops(filelike)
 * 
 * cdef class RWopsImpl(object):             # <<<<<<<<<<<<<<
 *     cdef SDL_RWops *ops
 *     cdef int closed
 */
struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl {
  PyObject_HEAD
  struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_vtab;
  SDL_RWops *ops;
  int closed;
};



struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl {
  SDL_RWops *(*get_rwops)(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *);
  PyObject *(*set_rwops)(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *, SDL_RWops *);
  PyObject *(*clear_rwops)(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *);
};
static struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_vtabptr_11pygame_sdl2_8rwobject_RWopsImpl;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* DivInt[ptrdiff_t].proto */
static CYTHON_INLINE ptrdiff_t __Pyx_div_ptrdiff_t(ptrdiff_t, ptrdiff_t);

/* UnaryNegOverflows.proto */
#define UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static SDL_RWops *__pyx_f_11pygame_sdl2_8rwobject_9RWopsImpl_get_rwops(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_11pygame_sdl2_8rwobject_9RWopsImpl_set_rwops(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self, SDL_RWops *__pyx_v_ops); /* proto*/
static PyObject *__pyx_f_11pygame_sdl2_8rwobject_9RWopsImpl_clear_rwops(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self); /* proto*/

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'libc.stddef' */

/* Module declarations from 'sdl2' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'pygame_sdl2.rwobject' */
static PyTypeObject *__pyx_ptype_11pygame_sdl2_8rwobject_RWopsImpl = 0;
static SDL_RWops *__pyx_f_11pygame_sdl2_8rwobject_to_rwops(PyObject *, struct __pyx_opt_args_11pygame_sdl2_8rwobject_to_rwops *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_11pygame_sdl2_8rwobject_set_error(PyObject *); /*proto*/
static Sint64 __pyx_f_11pygame_sdl2_8rwobject_python_size(SDL_RWops *); /*proto*/
static Sint64 __pyx_f_11pygame_sdl2_8rwobject_python_seek(SDL_RWops *, Sint64, int); /*proto*/
static size_t __pyx_f_11pygame_sdl2_8rwobject_python_read(SDL_RWops *, void *, size_t, size_t); /*proto*/
static size_t __pyx_f_11pygame_sdl2_8rwobject_python_write(SDL_RWops *, void const *, size_t, size_t); /*proto*/
static int __pyx_f_11pygame_sdl2_8rwobject_python_close(SDL_RWops *); /*proto*/
static Sint64 __pyx_f_11pygame_sdl2_8rwobject_subfile_size(SDL_RWops *); /*proto*/
static Sint64 __pyx_f_11pygame_sdl2_8rwobject_subfile_seek(SDL_RWops *, Sint64, int); /*proto*/
static size_t __pyx_f_11pygame_sdl2_8rwobject_subfile_read(SDL_RWops *, void *, size_t, size_t); /*proto*/
static int __pyx_f_11pygame_sdl2_8rwobject_subfile_close(SDL_RWops *); /*proto*/
static Sint64 __pyx_f_11pygame_sdl2_8rwobject_buffile_size(SDL_RWops *); /*proto*/
static Sint64 __pyx_f_11pygame_sdl2_8rwobject_buffile_seek(SDL_RWops *, Sint64, int); /*proto*/
static size_t __pyx_f_11pygame_sdl2_8rwobject_buffile_read(SDL_RWops *, void *, size_t, size_t); /*proto*/
static size_t __pyx_f_11pygame_sdl2_8rwobject_buffile_write(SDL_RWops *, void const *, size_t, size_t); /*proto*/
static int __pyx_f_11pygame_sdl2_8rwobject_buffile_close(SDL_RWops *); /*proto*/
static SDL_RWops *__pyx_f_11pygame_sdl2_8rwobject_RWopsFromPython(PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "pygame_sdl2.rwobject"
extern int __pyx_module_is_main_pygame_sdl2__rwobject;
int __pyx_module_is_main_pygame_sdl2__rwobject = 0;

/* Implementation of 'pygame_sdl2.rwobject' */
static PyObject *__pyx_builtin_property;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_IOError;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_OSError;
static const char __pyx_k_b[] = "b";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_bf[] = "bf";
static const char __pyx_k_io[] = "io";
static const char __pyx_k_rb[] = "rb";
static const char __pyx_k_rv[] = "rv";
static const char __pyx_k_rw[] = "rw";
static const char __pyx_k_sf[] = "sf";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_read[] = "read";
static const char __pyx_k_seek[] = "seek";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_tell[] = "tell";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_view[] = "view";
static const char __pyx_k_RWops[] = "RWops";
static const char __pyx_k_ascii[] = "ascii";
static const char __pyx_k_bytes[] = "bytes_";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_cmode[] = "cmode";
static const char __pyx_k_cname[] = "cname";
static const char __pyx_k_data1[] = "data1";
static const char __pyx_k_dname[] = "dname";
static const char __pyx_k_rw_in[] = "rw_in";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_IOBase[] = "IOBase";
static const char __pyx_k_closed[] = "closed";
static const char __pyx_k_ctypes[] = "ctypes";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_fileno[] = "fileno";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_hidden[] = "hidden";
static const char __pyx_k_holder[] = "_holder";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_length[] = "length";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_offset[] = "offset";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_whence[] = "whence";
static const char __pyx_k_IOError[] = "IOError";
static const char __pyx_k_OSError[] = "OSError";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_unicode[] = "unicode_";
static const char __pyx_k_unknown[] = "unknown";
static const char __pyx_k_SEEK_CUR[] = "SEEK_CUR";
static const char __pyx_k_SEEK_END[] = "SEEK_END";
static const char __pyx_k_SEEK_SET[] = "SEEK_SET";
static const char __pyx_k_c_void_p[] = "c_void_p";
static const char __pyx_k_filelike[] = "filelike";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_property[] = "property";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_readable[] = "readable";
static const char __pyx_k_readinto[] = "readinto";
static const char __pyx_k_seekable[] = "seekable";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_truncate[] = "truncate";
static const char __pyx_k_writable[] = "writable";
static const char __pyx_k_RWopsImpl[] = "RWopsImpl";
static const char __pyx_k_RawIOBase[] = "RawIOBase";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_file_type[] = "file_type";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_RWops_seek[] = "RWops.seek";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_fsencoding[] = "fsencoding";
static const char __pyx_k_get_closed[] = "get_closed";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_RWops_close[] = "RWops.close";
static const char __pyx_k_RWops_write[] = "RWops.write";
static const char __pyx_k_RWops___init[] = "RWops.__init__";
static const char __pyx_k_RWops_closed[] = "RWops.closed";
static const char __pyx_k_RWops_fileno[] = "RWops.fileno";
static const char __pyx_k_check_closed[] = "check_closed";
static const char __pyx_k_rw_in_object[] = "rw_in_object";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_Could_not_read[] = "Could not read: {}";
static const char __pyx_k_Could_not_seek[] = "Could not seek: {}";
static const char __pyx_k_RWops_readable[] = "RWops.readable";
static const char __pyx_k_RWops_readinto[] = "RWops.readinto";
static const char __pyx_k_RWops_seekable[] = "RWops.seekable";
static const char __pyx_k_RWops_truncate[] = "RWops.truncate";
static const char __pyx_k_RWops_writable[] = "RWops.writable";
static const char __pyx_k_Could_not_write[] = "Could not write: {}";
static const char __pyx_k_RWops_from_file[] = "RWops_from_file";
static const char __pyx_k_RWops_is_buffer[] = "RWops_is_buffer";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_Could_not_open_r[] = "Could not open {!r}: {}";
static const char __pyx_k_RWops_is_subfile[] = "RWops_is_subfile";
static const char __pyx_k_RWops_from_buffer[] = "RWops_from_buffer";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_pygame_sdl2_compat[] = "pygame_sdl2.compat";
static const char __pyx_k_RWops_create_subfile[] = "RWops_create_subfile";
static const char __pyx_k_RWops_unwrap_subfile[] = "RWops_unwrap_subfile";
static const char __pyx_k_pygame_sdl2_rwobject[] = "pygame_sdl2.rwobject";
static const char __pyx_k_get_sdl_rwops_pointer[] = "get_sdl_rwops_pointer";
static const char __pyx_k_getfilesystemencoding[] = "getfilesystemencoding";
static const char __pyx_k_Unknown_value_for_whence[] = "Unknown value for 'whence'";
static const char __pyx_k_RWops_is_file_like_object[] = "RWops_is_file_like_object";
static const char __pyx_k_RWops_from_file_like_object[] = "RWops_from_file_like_object";
static const char __pyx_k_RWops_get_sdl_rwops_pointer[] = "RWops.get_sdl_rwops_pointer";
static const char __pyx_k_src_pygame_sdl2_rwobject_pyx[] = "src/pygame_sdl2/rwobject.pyx";
static const char __pyx_k_I_O_operation_on_closed_RWops[] = "I/O operation on closed RWops.";
static const char __pyx_k_r_is_not_a_filename_or_file_lik[] = "{!r} is not a filename or file-like object.";
static const char __pyx_k_self_ops_cannot_be_converted_to[] = "self.ops cannot be converted to a Python object for pickling";
static const char __pyx_k_Invalid_value_passed_in_for_name[] = "Invalid value passed in for name";
static const char __pyx_k_Passed_in_RWops_object_is_closed[] = "Passed in RWops object is closed";
static const char __pyx_k_Passed_in_RWops_object_is_not_Su[] = "Passed in RWops object is not SubFile";
static const char __pyx_k_Passed_in_object_does_not_suppor[] = "Passed in object does not support buffer protocol";
static const char __pyx_k_Passed_in_object_should_be_RWops[] = "Passed in object should be RWops";
static PyObject *__pyx_n_s_AttributeError;
static PyObject *__pyx_kp_s_Could_not_open_r;
static PyObject *__pyx_kp_s_Could_not_read;
static PyObject *__pyx_kp_s_Could_not_seek;
static PyObject *__pyx_kp_s_Could_not_write;
static PyObject *__pyx_n_s_IOBase;
static PyObject *__pyx_n_s_IOError;
static PyObject *__pyx_kp_s_I_O_operation_on_closed_RWops;
static PyObject *__pyx_kp_s_Invalid_value_passed_in_for_name;
static PyObject *__pyx_n_s_OSError;
static PyObject *__pyx_kp_s_Passed_in_RWops_object_is_closed;
static PyObject *__pyx_kp_s_Passed_in_RWops_object_is_not_Su;
static PyObject *__pyx_kp_s_Passed_in_object_does_not_suppor;
static PyObject *__pyx_kp_s_Passed_in_object_should_be_RWops;
static PyObject *__pyx_n_s_RWops;
static PyObject *__pyx_n_s_RWopsImpl;
static PyObject *__pyx_n_s_RWops___init;
static PyObject *__pyx_n_s_RWops_close;
static PyObject *__pyx_n_s_RWops_closed;
static PyObject *__pyx_n_s_RWops_create_subfile;
static PyObject *__pyx_n_s_RWops_fileno;
static PyObject *__pyx_n_s_RWops_from_buffer;
static PyObject *__pyx_n_s_RWops_from_file;
static PyObject *__pyx_n_s_RWops_from_file_like_object;
static PyObject *__pyx_n_s_RWops_get_sdl_rwops_pointer;
static PyObject *__pyx_n_s_RWops_is_buffer;
static PyObject *__pyx_n_s_RWops_is_file_like_object;
static PyObject *__pyx_n_s_RWops_is_subfile;
static PyObject *__pyx_n_s_RWops_readable;
static PyObject *__pyx_n_s_RWops_readinto;
static PyObject *__pyx_n_s_RWops_seek;
static PyObject *__pyx_n_s_RWops_seekable;
static PyObject *__pyx_n_s_RWops_truncate;
static PyObject *__pyx_n_s_RWops_unwrap_subfile;
static PyObject *__pyx_n_s_RWops_writable;
static PyObject *__pyx_n_s_RWops_write;
static PyObject *__pyx_n_s_RawIOBase;
static PyObject *__pyx_n_s_SEEK_CUR;
static PyObject *__pyx_n_s_SEEK_END;
static PyObject *__pyx_n_s_SEEK_SET;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unknown_value_for_whence;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_ascii;
static PyObject *__pyx_n_s_b;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_bf;
static PyObject *__pyx_n_s_bytes;
static PyObject *__pyx_n_s_c_void_p;
static PyObject *__pyx_n_s_check_closed;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_closed;
static PyObject *__pyx_n_s_cmode;
static PyObject *__pyx_n_s_cname;
static PyObject *__pyx_n_s_ctypes;
static PyObject *__pyx_n_s_data1;
static PyObject *__pyx_n_s_decode;
static PyObject *__pyx_n_s_dname;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_file_type;
static PyObject *__pyx_n_s_filelike;
static PyObject *__pyx_n_s_fileno;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fsencoding;
static PyObject *__pyx_n_s_get_closed;
static PyObject *__pyx_n_s_get_sdl_rwops_pointer;
static PyObject *__pyx_n_s_getfilesystemencoding;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_hidden;
static PyObject *__pyx_n_s_holder;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_io;
static PyObject *__pyx_n_s_length;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_offset;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_property;
static PyObject *__pyx_n_s_pygame_sdl2_compat;
static PyObject *__pyx_n_s_pygame_sdl2_rwobject;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_r;
static PyObject *__pyx_kp_s_r_is_not_a_filename_or_file_lik;
static PyObject *__pyx_n_b_rb;
static PyObject *__pyx_n_s_rb;
static PyObject *__pyx_n_s_read;
static PyObject *__pyx_n_s_readable;
static PyObject *__pyx_n_s_readinto;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_rv;
static PyObject *__pyx_n_s_rw;
static PyObject *__pyx_n_s_rw_in;
static PyObject *__pyx_n_s_rw_in_object;
static PyObject *__pyx_n_s_seek;
static PyObject *__pyx_n_s_seekable;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_kp_s_self_ops_cannot_be_converted_to;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_sf;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_kp_s_src_pygame_sdl2_rwobject_pyx;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_tell;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_truncate;
static PyObject *__pyx_n_s_unicode;
static PyObject *__pyx_n_s_unknown;
static PyObject *__pyx_kp_s_utf_8;
static PyObject *__pyx_n_s_view;
static PyObject *__pyx_n_s_whence;
static PyObject *__pyx_n_s_writable;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_check_closed(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_2get_closed(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_4close(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_6seek(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self, PyObject *__pyx_v_offset, PyObject *__pyx_v_whence); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_8readinto(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self, PyObject *__pyx_v_b); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_10write(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self, PyObject *__pyx_v_b); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_12get_sdl_rwops_pointer(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_2close(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_4closed(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_6fileno(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_8readable(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_10seek(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_offset, PyObject *__pyx_v_whence); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_12seekable(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_14truncate(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_size); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_16writable(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_18readinto(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_b); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_20write(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_b); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_22get_sdl_rwops_pointer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_RWops_from_file(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_mode); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_2RWops_from_file_like_object(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_filelike); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_4RWops_is_file_like_object(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_6RWops_from_buffer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_b, PyObject *__pyx_v_mode); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_8RWops_is_buffer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_10RWops_create_subfile(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object, PyObject *__pyx_v_base, PyObject *__pyx_v_length); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_12RWops_is_subfile(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object); /* proto */
static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_14RWops_unwrap_subfile(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object); /* proto */
static PyObject *__pyx_tp_new_11pygame_sdl2_8rwobject_RWopsImpl(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_2;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_codeobj__11;
static PyObject *__pyx_codeobj__14;
static PyObject *__pyx_codeobj__16;
static PyObject *__pyx_codeobj__18;
static PyObject *__pyx_codeobj__20;
static PyObject *__pyx_codeobj__22;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__30;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__34;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__38;
static PyObject *__pyx_codeobj__40;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__46;
static PyObject *__pyx_codeobj__48;
static PyObject *__pyx_codeobj__50;
static PyObject *__pyx_codeobj__52;
/* Late includes */

/* "pygame_sdl2/rwobject.pyx":61
 * 
 * 
 * cdef set_error(e):             # <<<<<<<<<<<<<<
 *     cdef char *msg
 *     e = str(e)
 */

static PyObject *__pyx_f_11pygame_sdl2_8rwobject_set_error(PyObject *__pyx_v_e) {
  char *__pyx_v_msg;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_error", 0);
  __Pyx_INCREF(__pyx_v_e);

  /* "pygame_sdl2/rwobject.pyx":63
 * cdef set_error(e):
 *     cdef char *msg
 *     e = str(e)             # <<<<<<<<<<<<<<
 *     msg = <char *> e
 *     SDL_SetError("%s", msg)
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_e); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_e, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":64
 *     cdef char *msg
 *     e = str(e)
 *     msg = <char *> e             # <<<<<<<<<<<<<<
 *     SDL_SetError("%s", msg)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_v_e); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 64, __pyx_L1_error)
  __pyx_v_msg = ((char *)__pyx_t_2);

  /* "pygame_sdl2/rwobject.pyx":65
 *     e = str(e)
 *     msg = <char *> e
 *     SDL_SetError("%s", msg)             # <<<<<<<<<<<<<<
 * 
 * cdef Sint64 python_size(SDL_RWops *context) with gil:
 */
  (void)(SDL_SetError(((char const *)"%s"), __pyx_v_msg));

  /* "pygame_sdl2/rwobject.pyx":61
 * 
 * 
 * cdef set_error(e):             # <<<<<<<<<<<<<<
 *     cdef char *msg
 *     e = str(e)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.set_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":67
 *     SDL_SetError("%s", msg)
 * 
 * cdef Sint64 python_size(SDL_RWops *context) with gil:             # <<<<<<<<<<<<<<
 *     f = <object> context.hidden.unknown.data1
 * 
 */

static Sint64 __pyx_f_11pygame_sdl2_8rwobject_python_size(SDL_RWops *__pyx_v_context) {
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_cur = NULL;
  PyObject *__pyx_v_rv = NULL;
  Sint64 __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Sint64 __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("python_size", 0);

  /* "pygame_sdl2/rwobject.pyx":68
 * 
 * cdef Sint64 python_size(SDL_RWops *context) with gil:
 *     f = <object> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_context->hidden.unknown.data1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_f = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":70
 *     f = <object> context.hidden.unknown.data1
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         cur = f.tell()
 *         f.seek(0, 2)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "pygame_sdl2/rwobject.pyx":71
 * 
 *     try:
 *         cur = f.tell()             # <<<<<<<<<<<<<<
 *         f.seek(0, 2)
 *         rv = f.tell()
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_tell); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 71, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_cur = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "pygame_sdl2/rwobject.pyx":72
 *     try:
 *         cur = f.tell()
 *         f.seek(0, 2)             # <<<<<<<<<<<<<<
 *         rv = f.tell()
 *         f.seek(cur, 0)
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_seek); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 72, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pygame_sdl2/rwobject.pyx":73
 *         cur = f.tell()
 *         f.seek(0, 2)
 *         rv = f.tell()             # <<<<<<<<<<<<<<
 *         f.seek(cur, 0)
 *     except:
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_tell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 73, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_rv = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "pygame_sdl2/rwobject.pyx":74
 *         f.seek(0, 2)
 *         rv = f.tell()
 *         f.seek(cur, 0)             # <<<<<<<<<<<<<<
 *     except:
 *         return -1
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_seek); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_cur, __pyx_int_0};
        __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 74, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_cur, __pyx_int_0};
        __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 74, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 74, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_INCREF(__pyx_v_cur);
        __Pyx_GIVEREF(__pyx_v_cur);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_cur);
        __Pyx_INCREF(__pyx_int_0);
        __Pyx_GIVEREF(__pyx_int_0);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_int_0);
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 74, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pygame_sdl2/rwobject.pyx":70
 *     f = <object> context.hidden.unknown.data1
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         cur = f.tell()
 *         f.seek(0, 2)
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "pygame_sdl2/rwobject.pyx":75
 *         rv = f.tell()
 *         f.seek(cur, 0)
 *     except:             # <<<<<<<<<<<<<<
 *         return -1
 * 
 */
    /*except:*/ {
      __Pyx_AddTraceback("pygame_sdl2.rwobject.python_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_8) < 0) __PYX_ERR(0, 75, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_8);

      /* "pygame_sdl2/rwobject.pyx":76
 *         f.seek(cur, 0)
 *     except:
 *         return -1             # <<<<<<<<<<<<<<
 * 
 *     return rv
 */
      __pyx_r = -1L;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6_except_return;
    }
    __pyx_L5_except_error:;

    /* "pygame_sdl2/rwobject.pyx":70
 *     f = <object> context.hidden.unknown.data1
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         cur = f.tell()
 *         f.seek(0, 2)
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "pygame_sdl2/rwobject.pyx":78
 *         return -1
 * 
 *     return rv             # <<<<<<<<<<<<<<
 * 
 * cdef Sint64 python_seek(SDL_RWops *context, Sint64 seek, int whence) with gil:
 */
  __pyx_t_9 = __Pyx_PyInt_As_int64_t(__pyx_v_rv); if (unlikely((__pyx_t_9 == ((Sint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 78, __pyx_L1_error)
  __pyx_r = __pyx_t_9;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":67
 *     SDL_SetError("%s", msg)
 * 
 * cdef Sint64 python_size(SDL_RWops *context) with gil:             # <<<<<<<<<<<<<<
 *     f = <object> context.hidden.unknown.data1
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_WriteUnraisable("pygame_sdl2.rwobject.python_size", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_cur);
  __Pyx_XDECREF(__pyx_v_rv);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":80
 *     return rv
 * 
 * cdef Sint64 python_seek(SDL_RWops *context, Sint64 seek, int whence) with gil:             # <<<<<<<<<<<<<<
 *     f = <object> context.hidden.unknown.data1
 * 
 */

static Sint64 __pyx_f_11pygame_sdl2_8rwobject_python_seek(SDL_RWops *__pyx_v_context, Sint64 __pyx_v_seek, int __pyx_v_whence) {
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_rv = NULL;
  PyObject *__pyx_v_e = NULL;
  Sint64 __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  Sint64 __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("python_seek", 0);

  /* "pygame_sdl2/rwobject.pyx":81
 * 
 * cdef Sint64 python_seek(SDL_RWops *context, Sint64 seek, int whence) with gil:
 *     f = <object> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_context->hidden.unknown.data1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_f = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":83
 *     f = <object> context.hidden.unknown.data1
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         f.seek(seek, whence)
 *         rv = f.tell()
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "pygame_sdl2/rwobject.pyx":84
 * 
 *     try:
 *         f.seek(seek, whence)             # <<<<<<<<<<<<<<
 *         rv = f.tell()
 *     except Exception as e:
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_seek); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 84, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_seek); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 84, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_whence); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 84, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 84, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_8) {
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
        __pyx_t_6 = 0;
        __pyx_t_7 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pygame_sdl2/rwobject.pyx":85
 *     try:
 *         f.seek(seek, whence)
 *         rv = f.tell()             # <<<<<<<<<<<<<<
 *     except Exception as e:
 *         set_error(e)
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_tell); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 85, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_rv = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "pygame_sdl2/rwobject.pyx":83
 *     f = <object> context.hidden.unknown.data1
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         f.seek(seek, whence)
 *         rv = f.tell()
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "pygame_sdl2/rwobject.pyx":86
 *         f.seek(seek, whence)
 *         rv = f.tell()
 *     except Exception as e:             # <<<<<<<<<<<<<<
 *         set_error(e)
 *         return -1
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_9) {
      __Pyx_AddTraceback("pygame_sdl2.rwobject.python_seek", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_5, &__pyx_t_10) < 0) __PYX_ERR(0, 86, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_e = __pyx_t_5;
      /*try:*/ {

        /* "pygame_sdl2/rwobject.pyx":87
 *         rv = f.tell()
 *     except Exception as e:
 *         set_error(e)             # <<<<<<<<<<<<<<
 *         return -1
 * 
 */
        __pyx_t_7 = __pyx_f_11pygame_sdl2_8rwobject_set_error(__pyx_v_e); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 87, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "pygame_sdl2/rwobject.pyx":88
 *     except Exception as e:
 *         set_error(e)
 *         return -1             # <<<<<<<<<<<<<<
 * 
 *     return rv
 */
        __pyx_r = -1L;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L13_return;
      }

      /* "pygame_sdl2/rwobject.pyx":86
 *         f.seek(seek, whence)
 *         rv = f.tell()
 *     except Exception as e:             # <<<<<<<<<<<<<<
 *         set_error(e)
 *         return -1
 */
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __pyx_t_9 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e);
            __pyx_v_e = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          }
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
          goto __pyx_L5_except_error;
        }
        __pyx_L13_return: {
          __pyx_t_19 = __pyx_r;
          __Pyx_DECREF(__pyx_v_e);
          __pyx_v_e = NULL;
          __pyx_r = __pyx_t_19;
          goto __pyx_L6_except_return;
        }
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pygame_sdl2/rwobject.pyx":83
 *     f = <object> context.hidden.unknown.data1
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         f.seek(seek, whence)
 *         rv = f.tell()
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "pygame_sdl2/rwobject.pyx":90
 *         return -1
 * 
 *     return rv             # <<<<<<<<<<<<<<
 * 
 * cdef size_t python_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) with gil:
 */
  __pyx_t_19 = __Pyx_PyInt_As_int64_t(__pyx_v_rv); if (unlikely((__pyx_t_19 == ((Sint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 90, __pyx_L1_error)
  __pyx_r = __pyx_t_19;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":80
 *     return rv
 * 
 * cdef Sint64 python_seek(SDL_RWops *context, Sint64 seek, int whence) with gil:             # <<<<<<<<<<<<<<
 *     f = <object> context.hidden.unknown.data1
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_WriteUnraisable("pygame_sdl2.rwobject.python_seek", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_rv);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":92
 *     return rv
 * 
 * cdef size_t python_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) with gil:             # <<<<<<<<<<<<<<
 *     f = <object> context.hidden.unknown.data1
 * 
 */

static size_t __pyx_f_11pygame_sdl2_8rwobject_python_read(SDL_RWops *__pyx_v_context, void *__pyx_v_ptr, size_t __pyx_v_size, size_t __pyx_v_maxnum) {
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_e = NULL;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  char const *__pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  size_t __pyx_t_17;
  char *__pyx_t_18;
  Py_ssize_t __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("python_read", 0);

  /* "pygame_sdl2/rwobject.pyx":93
 * 
 * cdef size_t python_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) with gil:
 *     f = <object> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_context->hidden.unknown.data1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_f = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":95
 *     f = <object> context.hidden.unknown.data1
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         data = f.read(size * maxnum)
 *     except Exception as e:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "pygame_sdl2/rwobject.pyx":96
 * 
 *     try:
 *         data = f.read(size * maxnum)             # <<<<<<<<<<<<<<
 *     except Exception as e:
 *         set_error(e)
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_read); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 96, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyInt_FromSize_t((__pyx_v_size * __pyx_v_maxnum)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 96, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_data = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "pygame_sdl2/rwobject.pyx":95
 *     f = <object> context.hidden.unknown.data1
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         data = f.read(size * maxnum)
 *     except Exception as e:
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "pygame_sdl2/rwobject.pyx":97
 *     try:
 *         data = f.read(size * maxnum)
 *     except Exception as e:             # <<<<<<<<<<<<<<
 *         set_error(e)
 *         return -1
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_8) {
      __Pyx_AddTraceback("pygame_sdl2.rwobject.python_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 97, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_e = __pyx_t_5;
      /*try:*/ {

        /* "pygame_sdl2/rwobject.pyx":98
 *         data = f.read(size * maxnum)
 *     except Exception as e:
 *         set_error(e)             # <<<<<<<<<<<<<<
 *         return -1
 * 
 */
        __pyx_t_7 = __pyx_f_11pygame_sdl2_8rwobject_set_error(__pyx_v_e); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 98, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "pygame_sdl2/rwobject.pyx":99
 *     except Exception as e:
 *         set_error(e)
 *         return -1             # <<<<<<<<<<<<<<
 * 
 *     memcpy(ptr, <void *><char *> data, len(data))
 */
        __pyx_r = -1L;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L13_return;
      }

      /* "pygame_sdl2/rwobject.pyx":97
 *     try:
 *         data = f.read(size * maxnum)
 *     except Exception as e:             # <<<<<<<<<<<<<<
 *         set_error(e)
 *         return -1
 */
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13) < 0)) __Pyx_ErrFetch(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __pyx_t_8 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_10 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e);
            __pyx_v_e = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          }
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ErrRestore(__pyx_t_11, __pyx_t_12, __pyx_t_13);
          __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_10;
          goto __pyx_L5_except_error;
        }
        __pyx_L13_return: {
          __pyx_t_17 = __pyx_r;
          __Pyx_DECREF(__pyx_v_e);
          __pyx_v_e = NULL;
          __pyx_r = __pyx_t_17;
          goto __pyx_L6_except_return;
        }
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pygame_sdl2/rwobject.pyx":95
 *     f = <object> context.hidden.unknown.data1
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         data = f.read(size * maxnum)
 *     except Exception as e:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "pygame_sdl2/rwobject.pyx":101
 *         return -1
 * 
 *     memcpy(ptr, <void *><char *> data, len(data))             # <<<<<<<<<<<<<<
 *     return len(data)
 * 
 */
  __pyx_t_18 = __Pyx_PyObject_AsWritableString(__pyx_v_data); if (unlikely((!__pyx_t_18) && PyErr_Occurred())) __PYX_ERR(0, 101, __pyx_L1_error)
  __pyx_t_19 = PyObject_Length(__pyx_v_data); if (unlikely(__pyx_t_19 == ((Py_ssize_t)-1))) __PYX_ERR(0, 101, __pyx_L1_error)
  (void)(memcpy(__pyx_v_ptr, ((void *)((char *)__pyx_t_18)), __pyx_t_19));

  /* "pygame_sdl2/rwobject.pyx":102
 * 
 *     memcpy(ptr, <void *><char *> data, len(data))
 *     return len(data)             # <<<<<<<<<<<<<<
 * 
 * cdef size_t python_write(SDL_RWops *context, const void *ptr, size_t size, size_t maxnum) with gil:
 */
  __pyx_t_19 = PyObject_Length(__pyx_v_data); if (unlikely(__pyx_t_19 == ((Py_ssize_t)-1))) __PYX_ERR(0, 102, __pyx_L1_error)
  __pyx_r = __pyx_t_19;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":92
 *     return rv
 * 
 * cdef size_t python_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) with gil:             # <<<<<<<<<<<<<<
 *     f = <object> context.hidden.unknown.data1
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_WriteUnraisable("pygame_sdl2.rwobject.python_read", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":104
 *     return len(data)
 * 
 * cdef size_t python_write(SDL_RWops *context, const void *ptr, size_t size, size_t maxnum) with gil:             # <<<<<<<<<<<<<<
 *     f = <object> context.hidden.unknown.data1
 *     data = (<char *> ptr)[:size * maxnum]
 */

static size_t __pyx_f_11pygame_sdl2_8rwobject_python_write(SDL_RWops *__pyx_v_context, void const *__pyx_v_ptr, size_t __pyx_v_size, size_t __pyx_v_maxnum) {
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_e = NULL;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  char const *__pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  size_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("python_write", 0);

  /* "pygame_sdl2/rwobject.pyx":105
 * 
 * cdef size_t python_write(SDL_RWops *context, const void *ptr, size_t size, size_t maxnum) with gil:
 *     f = <object> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 *     data = (<char *> ptr)[:size * maxnum]
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_context->hidden.unknown.data1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_f = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":106
 * cdef size_t python_write(SDL_RWops *context, const void *ptr, size_t size, size_t maxnum) with gil:
 *     f = <object> context.hidden.unknown.data1
 *     data = (<char *> ptr)[:size * maxnum]             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(((char *)__pyx_v_ptr) + 0, (__pyx_v_size * __pyx_v_maxnum) - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":108
 *     data = (<char *> ptr)[:size * maxnum]
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         f.write(data)
 *     except Exception as e:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "pygame_sdl2/rwobject.pyx":109
 * 
 *     try:
 *         f.write(data)             # <<<<<<<<<<<<<<
 *     except Exception as e:
 *         set_error(e)
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 109, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_data) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_data);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pygame_sdl2/rwobject.pyx":108
 *     data = (<char *> ptr)[:size * maxnum]
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         f.write(data)
 *     except Exception as e:
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "pygame_sdl2/rwobject.pyx":110
 *     try:
 *         f.write(data)
 *     except Exception as e:             # <<<<<<<<<<<<<<
 *         set_error(e)
 *         return -1
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_7) {
      __Pyx_AddTraceback("pygame_sdl2.rwobject.python_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 110, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_e = __pyx_t_5;
      /*try:*/ {

        /* "pygame_sdl2/rwobject.pyx":111
 *         f.write(data)
 *     except Exception as e:
 *         set_error(e)             # <<<<<<<<<<<<<<
 *         return -1
 * 
 */
        __pyx_t_8 = __pyx_f_11pygame_sdl2_8rwobject_set_error(__pyx_v_e); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 111, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "pygame_sdl2/rwobject.pyx":112
 *     except Exception as e:
 *         set_error(e)
 *         return -1             # <<<<<<<<<<<<<<
 * 
 *     return len(data)
 */
        __pyx_r = -1L;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L13_return;
      }

      /* "pygame_sdl2/rwobject.pyx":110
 *     try:
 *         f.write(data)
 *     except Exception as e:             # <<<<<<<<<<<<<<
 *         set_error(e)
 *         return -1
 */
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13) < 0)) __Pyx_ErrFetch(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __pyx_t_7 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_10 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e);
            __pyx_v_e = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          }
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ErrRestore(__pyx_t_11, __pyx_t_12, __pyx_t_13);
          __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
          __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_10;
          goto __pyx_L5_except_error;
        }
        __pyx_L13_return: {
          __pyx_t_17 = __pyx_r;
          __Pyx_DECREF(__pyx_v_e);
          __pyx_v_e = NULL;
          __pyx_r = __pyx_t_17;
          goto __pyx_L6_except_return;
        }
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pygame_sdl2/rwobject.pyx":108
 *     data = (<char *> ptr)[:size * maxnum]
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         f.write(data)
 *     except Exception as e:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "pygame_sdl2/rwobject.pyx":114
 *         return -1
 * 
 *     return len(data)             # <<<<<<<<<<<<<<
 * 
 * cdef int python_close(SDL_RWops *context) with gil:
 */
  __pyx_t_18 = PyBytes_GET_SIZE(__pyx_v_data); if (unlikely(__pyx_t_18 == ((Py_ssize_t)-1))) __PYX_ERR(0, 114, __pyx_L1_error)
  __pyx_r = __pyx_t_18;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":104
 *     return len(data)
 * 
 * cdef size_t python_write(SDL_RWops *context, const void *ptr, size_t size, size_t maxnum) with gil:             # <<<<<<<<<<<<<<
 *     f = <object> context.hidden.unknown.data1
 *     data = (<char *> ptr)[:size * maxnum]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_WriteUnraisable("pygame_sdl2.rwobject.python_write", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":116
 *     return len(data)
 * 
 * cdef int python_close(SDL_RWops *context) with gil:             # <<<<<<<<<<<<<<
 *     if context != NULL:
 *         if context.hidden.unknown.data1 != NULL:
 */

static int __pyx_f_11pygame_sdl2_8rwobject_python_close(SDL_RWops *__pyx_v_context) {
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_e = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("python_close", 0);

  /* "pygame_sdl2/rwobject.pyx":117
 * 
 * cdef int python_close(SDL_RWops *context) with gil:
 *     if context != NULL:             # <<<<<<<<<<<<<<
 *         if context.hidden.unknown.data1 != NULL:
 *             f = <object> context.hidden.unknown.data1
 */
  __pyx_t_1 = ((__pyx_v_context != NULL) != 0);
  if (__pyx_t_1) {

    /* "pygame_sdl2/rwobject.pyx":118
 * cdef int python_close(SDL_RWops *context) with gil:
 *     if context != NULL:
 *         if context.hidden.unknown.data1 != NULL:             # <<<<<<<<<<<<<<
 *             f = <object> context.hidden.unknown.data1
 * 
 */
    __pyx_t_1 = ((__pyx_v_context->hidden.unknown.data1 != NULL) != 0);
    if (__pyx_t_1) {

      /* "pygame_sdl2/rwobject.pyx":119
 *     if context != NULL:
 *         if context.hidden.unknown.data1 != NULL:
 *             f = <object> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
      __pyx_t_2 = ((PyObject *)__pyx_v_context->hidden.unknown.data1);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_v_f = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "pygame_sdl2/rwobject.pyx":121
 *             f = <object> context.hidden.unknown.data1
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 f.close()
 *             except Exception as e:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_5);
        /*try:*/ {

          /* "pygame_sdl2/rwobject.pyx":122
 * 
 *             try:
 *                 f.close()             # <<<<<<<<<<<<<<
 *             except Exception as e:
 *                 set_error(e)
 */
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_close); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 122, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
            }
          }
          __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 122, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pygame_sdl2/rwobject.pyx":121
 *             f = <object> context.hidden.unknown.data1
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 f.close()
 *             except Exception as e:
 */
        }
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L10_try_end;
        __pyx_L5_error:;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "pygame_sdl2/rwobject.pyx":123
 *             try:
 *                 f.close()
 *             except Exception as e:             # <<<<<<<<<<<<<<
 *                 set_error(e)
 *                 return -1
 */
        __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
        if (__pyx_t_8) {
          __Pyx_AddTraceback("pygame_sdl2.rwobject.python_close", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 123, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_6);
          __pyx_v_e = __pyx_t_6;
          /*try:*/ {

            /* "pygame_sdl2/rwobject.pyx":124
 *                 f.close()
 *             except Exception as e:
 *                 set_error(e)             # <<<<<<<<<<<<<<
 *                 return -1
 * 
 */
            __pyx_t_9 = __pyx_f_11pygame_sdl2_8rwobject_set_error(__pyx_v_e); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 124, __pyx_L16_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

            /* "pygame_sdl2/rwobject.pyx":125
 *             except Exception as e:
 *                 set_error(e)
 *                 return -1             # <<<<<<<<<<<<<<
 * 
 *             Py_DECREF(f)
 */
            __pyx_r = -1;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            goto __pyx_L15_return;
          }

          /* "pygame_sdl2/rwobject.pyx":123
 *             try:
 *                 f.close()
 *             except Exception as e:             # <<<<<<<<<<<<<<
 *                 set_error(e)
 *                 return -1
 */
          /*finally:*/ {
            __pyx_L16_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
              __Pyx_XGOTREF(__pyx_t_12);
              __Pyx_XGOTREF(__pyx_t_13);
              __Pyx_XGOTREF(__pyx_t_14);
              __Pyx_XGOTREF(__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_16);
              __Pyx_XGOTREF(__pyx_t_17);
              __pyx_t_8 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_e);
                __pyx_v_e = NULL;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_15);
                __Pyx_XGIVEREF(__pyx_t_16);
                __Pyx_XGIVEREF(__pyx_t_17);
                __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
              }
              __Pyx_XGIVEREF(__pyx_t_12);
              __Pyx_XGIVEREF(__pyx_t_13);
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
              __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
              __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_11;
              goto __pyx_L7_except_error;
            }
            __pyx_L15_return: {
              __pyx_t_10 = __pyx_r;
              __Pyx_DECREF(__pyx_v_e);
              __pyx_v_e = NULL;
              __pyx_r = __pyx_t_10;
              goto __pyx_L8_except_return;
            }
          }
        }
        goto __pyx_L7_except_error;
        __pyx_L7_except_error:;

        /* "pygame_sdl2/rwobject.pyx":121
 *             f = <object> context.hidden.unknown.data1
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 f.close()
 *             except Exception as e:
 */
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
        goto __pyx_L1_error;
        __pyx_L8_except_return:;
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
        goto __pyx_L0;
        __pyx_L10_try_end:;
      }

      /* "pygame_sdl2/rwobject.pyx":127
 *                 return -1
 * 
 *             Py_DECREF(f)             # <<<<<<<<<<<<<<
 * 
 *             context.hidden.unknown.data1 = NULL
 */
      Py_DECREF(__pyx_v_f);

      /* "pygame_sdl2/rwobject.pyx":129
 *             Py_DECREF(f)
 * 
 *             context.hidden.unknown.data1 = NULL             # <<<<<<<<<<<<<<
 *         SDL_FreeRW(context)
 *     return 0
 */
      __pyx_v_context->hidden.unknown.data1 = NULL;

      /* "pygame_sdl2/rwobject.pyx":118
 * cdef int python_close(SDL_RWops *context) with gil:
 *     if context != NULL:
 *         if context.hidden.unknown.data1 != NULL:             # <<<<<<<<<<<<<<
 *             f = <object> context.hidden.unknown.data1
 * 
 */
    }

    /* "pygame_sdl2/rwobject.pyx":130
 * 
 *             context.hidden.unknown.data1 = NULL
 *         SDL_FreeRW(context)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    SDL_FreeRW(__pyx_v_context);

    /* "pygame_sdl2/rwobject.pyx":117
 * 
 * cdef int python_close(SDL_RWops *context) with gil:
 *     if context != NULL:             # <<<<<<<<<<<<<<
 *         if context.hidden.unknown.data1 != NULL:
 *             f = <object> context.hidden.unknown.data1
 */
  }

  /* "pygame_sdl2/rwobject.pyx":131
 *             context.hidden.unknown.data1 = NULL
 *         SDL_FreeRW(context)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef struct SubFile:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":116
 *     return len(data)
 * 
 * cdef int python_close(SDL_RWops *context) with gil:             # <<<<<<<<<<<<<<
 *     if context != NULL:
 *         if context.hidden.unknown.data1 != NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("pygame_sdl2.rwobject.python_close", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":139
 *     Sint64 tell
 * 
 * cdef Sint64 subfile_size(SDL_RWops *context) nogil:             # <<<<<<<<<<<<<<
 *     cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1
 *     return sf.length
 */

static Sint64 __pyx_f_11pygame_sdl2_8rwobject_subfile_size(SDL_RWops *__pyx_v_context) {
  struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *__pyx_v_sf;
  Sint64 __pyx_r;

  /* "pygame_sdl2/rwobject.pyx":140
 * 
 * cdef Sint64 subfile_size(SDL_RWops *context) nogil:
 *     cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 *     return sf.length
 * 
 */
  __pyx_v_sf = ((struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *)__pyx_v_context->hidden.unknown.data1);

  /* "pygame_sdl2/rwobject.pyx":141
 * cdef Sint64 subfile_size(SDL_RWops *context) nogil:
 *     cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1
 *     return sf.length             # <<<<<<<<<<<<<<
 * 
 * cdef Sint64 subfile_seek(SDL_RWops *context, Sint64 seek, int whence) nogil:
 */
  __pyx_r = __pyx_v_sf->length;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":139
 *     Sint64 tell
 * 
 * cdef Sint64 subfile_size(SDL_RWops *context) nogil:             # <<<<<<<<<<<<<<
 *     cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1
 *     return sf.length
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":143
 *     return sf.length
 * 
 * cdef Sint64 subfile_seek(SDL_RWops *context, Sint64 seek, int whence) nogil:             # <<<<<<<<<<<<<<
 *     cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1
 * 
 */

static Sint64 __pyx_f_11pygame_sdl2_8rwobject_subfile_seek(SDL_RWops *__pyx_v_context, Sint64 __pyx_v_seek, int __pyx_v_whence) {
  struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *__pyx_v_sf;
  Sint64 __pyx_r;

  /* "pygame_sdl2/rwobject.pyx":144
 * 
 * cdef Sint64 subfile_seek(SDL_RWops *context, Sint64 seek, int whence) nogil:
 *     cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 * 
 *     if whence == RW_SEEK_SET:
 */
  __pyx_v_sf = ((struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *)__pyx_v_context->hidden.unknown.data1);

  /* "pygame_sdl2/rwobject.pyx":146
 *     cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1
 * 
 *     if whence == RW_SEEK_SET:             # <<<<<<<<<<<<<<
 *         sf.tell = SDL_RWseek(sf.rw, seek + sf.base, RW_SEEK_SET) - sf.base
 *     elif whence == RW_SEEK_CUR:
 */
  switch (__pyx_v_whence) {
    case RW_SEEK_SET:

    /* "pygame_sdl2/rwobject.pyx":147
 * 
 *     if whence == RW_SEEK_SET:
 *         sf.tell = SDL_RWseek(sf.rw, seek + sf.base, RW_SEEK_SET) - sf.base             # <<<<<<<<<<<<<<
 *     elif whence == RW_SEEK_CUR:
 *         sf.tell = SDL_RWseek(sf.rw, seek, RW_SEEK_CUR) - sf.base
 */
    __pyx_v_sf->tell = (SDL_RWseek(__pyx_v_sf->rw, (__pyx_v_seek + __pyx_v_sf->base), RW_SEEK_SET) - __pyx_v_sf->base);

    /* "pygame_sdl2/rwobject.pyx":146
 *     cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1
 * 
 *     if whence == RW_SEEK_SET:             # <<<<<<<<<<<<<<
 *         sf.tell = SDL_RWseek(sf.rw, seek + sf.base, RW_SEEK_SET) - sf.base
 *     elif whence == RW_SEEK_CUR:
 */
    break;
    case RW_SEEK_CUR:

    /* "pygame_sdl2/rwobject.pyx":149
 *         sf.tell = SDL_RWseek(sf.rw, seek + sf.base, RW_SEEK_SET) - sf.base
 *     elif whence == RW_SEEK_CUR:
 *         sf.tell = SDL_RWseek(sf.rw, seek, RW_SEEK_CUR) - sf.base             # <<<<<<<<<<<<<<
 *     elif whence == RW_SEEK_END:
 *         sf.tell = SDL_RWseek(sf.rw, sf.base + sf.length + seek, RW_SEEK_SET) - sf.base
 */
    __pyx_v_sf->tell = (SDL_RWseek(__pyx_v_sf->rw, __pyx_v_seek, RW_SEEK_CUR) - __pyx_v_sf->base);

    /* "pygame_sdl2/rwobject.pyx":148
 *     if whence == RW_SEEK_SET:
 *         sf.tell = SDL_RWseek(sf.rw, seek + sf.base, RW_SEEK_SET) - sf.base
 *     elif whence == RW_SEEK_CUR:             # <<<<<<<<<<<<<<
 *         sf.tell = SDL_RWseek(sf.rw, seek, RW_SEEK_CUR) - sf.base
 *     elif whence == RW_SEEK_END:
 */
    break;
    case RW_SEEK_END:

    /* "pygame_sdl2/rwobject.pyx":151
 *         sf.tell = SDL_RWseek(sf.rw, seek, RW_SEEK_CUR) - sf.base
 *     elif whence == RW_SEEK_END:
 *         sf.tell = SDL_RWseek(sf.rw, sf.base + sf.length + seek, RW_SEEK_SET) - sf.base             # <<<<<<<<<<<<<<
 * 
 *     return sf.tell
 */
    __pyx_v_sf->tell = (SDL_RWseek(__pyx_v_sf->rw, ((__pyx_v_sf->base + __pyx_v_sf->length) + __pyx_v_seek), RW_SEEK_SET) - __pyx_v_sf->base);

    /* "pygame_sdl2/rwobject.pyx":150
 *     elif whence == RW_SEEK_CUR:
 *         sf.tell = SDL_RWseek(sf.rw, seek, RW_SEEK_CUR) - sf.base
 *     elif whence == RW_SEEK_END:             # <<<<<<<<<<<<<<
 *         sf.tell = SDL_RWseek(sf.rw, sf.base + sf.length + seek, RW_SEEK_SET) - sf.base
 * 
 */
    break;
    default: break;
  }

  /* "pygame_sdl2/rwobject.pyx":153
 *         sf.tell = SDL_RWseek(sf.rw, sf.base + sf.length + seek, RW_SEEK_SET) - sf.base
 * 
 *     return sf.tell             # <<<<<<<<<<<<<<
 * 
 * cdef size_t subfile_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) nogil:
 */
  __pyx_r = __pyx_v_sf->tell;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":143
 *     return sf.length
 * 
 * cdef Sint64 subfile_seek(SDL_RWops *context, Sint64 seek, int whence) nogil:             # <<<<<<<<<<<<<<
 *     cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":155
 *     return sf.tell
 * 
 * cdef size_t subfile_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) nogil:             # <<<<<<<<<<<<<<
 *     cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1
 * 
 */

static size_t __pyx_f_11pygame_sdl2_8rwobject_subfile_read(SDL_RWops *__pyx_v_context, void *__pyx_v_ptr, size_t __pyx_v_size, size_t __pyx_v_maxnum) {
  struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *__pyx_v_sf;
  Sint64 __pyx_v_left;
  size_t __pyx_v_rv;
  size_t __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pygame_sdl2/rwobject.pyx":156
 * 
 * cdef size_t subfile_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) nogil:
 *     cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 * 
 *     cdef Sint64 left = sf.length - sf.tell
 */
  __pyx_v_sf = ((struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *)__pyx_v_context->hidden.unknown.data1);

  /* "pygame_sdl2/rwobject.pyx":158
 *     cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1
 * 
 *     cdef Sint64 left = sf.length - sf.tell             # <<<<<<<<<<<<<<
 *     cdef size_t rv
 * 
 */
  __pyx_v_left = (__pyx_v_sf->length - __pyx_v_sf->tell);

  /* "pygame_sdl2/rwobject.pyx":161
 *     cdef size_t rv
 * 
 *     if size * maxnum > left:             # <<<<<<<<<<<<<<
 *         maxnum = left // size
 * 
 */
  __pyx_t_1 = (((__pyx_v_size * __pyx_v_maxnum) > __pyx_v_left) != 0);
  if (__pyx_t_1) {

    /* "pygame_sdl2/rwobject.pyx":162
 * 
 *     if size * maxnum > left:
 *         maxnum = left // size             # <<<<<<<<<<<<<<
 * 
 *     if maxnum == 0:
 */
    if (unlikely(__pyx_v_size == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 162, __pyx_L1_error)
    }
    __pyx_v_maxnum = (__pyx_v_left / __pyx_v_size);

    /* "pygame_sdl2/rwobject.pyx":161
 *     cdef size_t rv
 * 
 *     if size * maxnum > left:             # <<<<<<<<<<<<<<
 *         maxnum = left // size
 * 
 */
  }

  /* "pygame_sdl2/rwobject.pyx":164
 *         maxnum = left // size
 * 
 *     if maxnum == 0:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_1 = ((__pyx_v_maxnum == 0) != 0);
  if (__pyx_t_1) {

    /* "pygame_sdl2/rwobject.pyx":165
 * 
 *     if maxnum == 0:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     rv = SDL_RWread(sf.rw, ptr, size, maxnum)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "pygame_sdl2/rwobject.pyx":164
 *         maxnum = left // size
 * 
 *     if maxnum == 0:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  }

  /* "pygame_sdl2/rwobject.pyx":167
 *         return 0
 * 
 *     rv = SDL_RWread(sf.rw, ptr, size, maxnum)             # <<<<<<<<<<<<<<
 * 
 *     if rv > 0:
 */
  __pyx_v_rv = SDL_RWread(__pyx_v_sf->rw, __pyx_v_ptr, __pyx_v_size, __pyx_v_maxnum);

  /* "pygame_sdl2/rwobject.pyx":169
 *     rv = SDL_RWread(sf.rw, ptr, size, maxnum)
 * 
 *     if rv > 0:             # <<<<<<<<<<<<<<
 *         sf.tell += size * rv
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv > 0) != 0);
  if (__pyx_t_1) {

    /* "pygame_sdl2/rwobject.pyx":170
 * 
 *     if rv > 0:
 *         sf.tell += size * rv             # <<<<<<<<<<<<<<
 * 
 *     return rv
 */
    __pyx_v_sf->tell = (__pyx_v_sf->tell + (__pyx_v_size * __pyx_v_rv));

    /* "pygame_sdl2/rwobject.pyx":169
 *     rv = SDL_RWread(sf.rw, ptr, size, maxnum)
 * 
 *     if rv > 0:             # <<<<<<<<<<<<<<
 *         sf.tell += size * rv
 * 
 */
  }

  /* "pygame_sdl2/rwobject.pyx":172
 *         sf.tell += size * rv
 * 
 *     return rv             # <<<<<<<<<<<<<<
 * 
 * cdef int subfile_close(SDL_RWops *context) nogil:
 */
  __pyx_r = __pyx_v_rv;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":155
 *     return sf.tell
 * 
 * cdef size_t subfile_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) nogil:             # <<<<<<<<<<<<<<
 *     cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("pygame_sdl2.rwobject.subfile_read", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":174
 *     return rv
 * 
 * cdef int subfile_close(SDL_RWops *context) nogil:             # <<<<<<<<<<<<<<
 *     cdef SubFile *sf
 * 
 */

static int __pyx_f_11pygame_sdl2_8rwobject_subfile_close(SDL_RWops *__pyx_v_context) {
  struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *__pyx_v_sf;
  int __pyx_r;
  int __pyx_t_1;

  /* "pygame_sdl2/rwobject.pyx":177
 *     cdef SubFile *sf
 * 
 *     if context != NULL:             # <<<<<<<<<<<<<<
 *         sf = <SubFile *> context.hidden.unknown.data1
 *         if sf.rw != NULL:
 */
  __pyx_t_1 = ((__pyx_v_context != NULL) != 0);
  if (__pyx_t_1) {

    /* "pygame_sdl2/rwobject.pyx":178
 * 
 *     if context != NULL:
 *         sf = <SubFile *> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 *         if sf.rw != NULL:
 *             SDL_RWclose(sf.rw)
 */
    __pyx_v_sf = ((struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *)__pyx_v_context->hidden.unknown.data1);

    /* "pygame_sdl2/rwobject.pyx":179
 *     if context != NULL:
 *         sf = <SubFile *> context.hidden.unknown.data1
 *         if sf.rw != NULL:             # <<<<<<<<<<<<<<
 *             SDL_RWclose(sf.rw)
 *         if sf != NULL:
 */
    __pyx_t_1 = ((__pyx_v_sf->rw != NULL) != 0);
    if (__pyx_t_1) {

      /* "pygame_sdl2/rwobject.pyx":180
 *         sf = <SubFile *> context.hidden.unknown.data1
 *         if sf.rw != NULL:
 *             SDL_RWclose(sf.rw)             # <<<<<<<<<<<<<<
 *         if sf != NULL:
 *             free(sf)
 */
      (void)(SDL_RWclose(__pyx_v_sf->rw));

      /* "pygame_sdl2/rwobject.pyx":179
 *     if context != NULL:
 *         sf = <SubFile *> context.hidden.unknown.data1
 *         if sf.rw != NULL:             # <<<<<<<<<<<<<<
 *             SDL_RWclose(sf.rw)
 *         if sf != NULL:
 */
    }

    /* "pygame_sdl2/rwobject.pyx":181
 *         if sf.rw != NULL:
 *             SDL_RWclose(sf.rw)
 *         if sf != NULL:             # <<<<<<<<<<<<<<
 *             free(sf)
 *             context.hidden.unknown.data1 = NULL
 */
    __pyx_t_1 = ((__pyx_v_sf != NULL) != 0);
    if (__pyx_t_1) {

      /* "pygame_sdl2/rwobject.pyx":182
 *             SDL_RWclose(sf.rw)
 *         if sf != NULL:
 *             free(sf)             # <<<<<<<<<<<<<<
 *             context.hidden.unknown.data1 = NULL
 *         SDL_FreeRW(context)
 */
      free(__pyx_v_sf);

      /* "pygame_sdl2/rwobject.pyx":183
 *         if sf != NULL:
 *             free(sf)
 *             context.hidden.unknown.data1 = NULL             # <<<<<<<<<<<<<<
 *         SDL_FreeRW(context)
 * 
 */
      __pyx_v_context->hidden.unknown.data1 = NULL;

      /* "pygame_sdl2/rwobject.pyx":181
 *         if sf.rw != NULL:
 *             SDL_RWclose(sf.rw)
 *         if sf != NULL:             # <<<<<<<<<<<<<<
 *             free(sf)
 *             context.hidden.unknown.data1 = NULL
 */
    }

    /* "pygame_sdl2/rwobject.pyx":184
 *             free(sf)
 *             context.hidden.unknown.data1 = NULL
 *         SDL_FreeRW(context)             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    SDL_FreeRW(__pyx_v_context);

    /* "pygame_sdl2/rwobject.pyx":177
 *     cdef SubFile *sf
 * 
 *     if context != NULL:             # <<<<<<<<<<<<<<
 *         sf = <SubFile *> context.hidden.unknown.data1
 *         if sf.rw != NULL:
 */
  }

  /* "pygame_sdl2/rwobject.pyx":186
 *         SDL_FreeRW(context)
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef struct BufFile:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":174
 *     return rv
 * 
 * cdef int subfile_close(SDL_RWops *context) nogil:             # <<<<<<<<<<<<<<
 *     cdef SubFile *sf
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":194
 *     Uint8 *stop
 * 
 * cdef Sint64 buffile_size(SDL_RWops *context) nogil:             # <<<<<<<<<<<<<<
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1
 * 
 */

static Sint64 __pyx_f_11pygame_sdl2_8rwobject_buffile_size(SDL_RWops *__pyx_v_context) {
  struct __pyx_t_11pygame_sdl2_8rwobject_BufFile *__pyx_v_bf;
  Sint64 __pyx_r;

  /* "pygame_sdl2/rwobject.pyx":195
 * 
 * cdef Sint64 buffile_size(SDL_RWops *context) nogil:
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 * 
 *     return bf.stop - bf.base
 */
  __pyx_v_bf = ((struct __pyx_t_11pygame_sdl2_8rwobject_BufFile *)__pyx_v_context->hidden.unknown.data1);

  /* "pygame_sdl2/rwobject.pyx":197
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1
 * 
 *     return bf.stop - bf.base             # <<<<<<<<<<<<<<
 * 
 * cdef Sint64 buffile_seek(SDL_RWops *context, Sint64 offset, int whence) nogil:
 */
  __pyx_r = (__pyx_v_bf->stop - __pyx_v_bf->base);
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":194
 *     Uint8 *stop
 * 
 * cdef Sint64 buffile_size(SDL_RWops *context) nogil:             # <<<<<<<<<<<<<<
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":199
 *     return bf.stop - bf.base
 * 
 * cdef Sint64 buffile_seek(SDL_RWops *context, Sint64 offset, int whence) nogil:             # <<<<<<<<<<<<<<
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1
 * 
 */

static Sint64 __pyx_f_11pygame_sdl2_8rwobject_buffile_seek(SDL_RWops *__pyx_v_context, Sint64 __pyx_v_offset, int __pyx_v_whence) {
  struct __pyx_t_11pygame_sdl2_8rwobject_BufFile *__pyx_v_bf;
  Uint8 *__pyx_v_newpos;
  Sint64 __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Uint8 *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("buffile_seek", 1);

  /* "pygame_sdl2/rwobject.pyx":200
 * 
 * cdef Sint64 buffile_seek(SDL_RWops *context, Sint64 offset, int whence) nogil:
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 * 
 *     cdef Uint8 *newpos
 */
  /*try:*/ {
    __pyx_v_bf = ((struct __pyx_t_11pygame_sdl2_8rwobject_BufFile *)__pyx_v_context->hidden.unknown.data1);

    /* "pygame_sdl2/rwobject.pyx":204
 *     cdef Uint8 *newpos
 * 
 *     if whence == RW_SEEK_SET:             # <<<<<<<<<<<<<<
 *         newpos = bf.base + offset
 *     elif whence == RW_SEEK_CUR:
 */
    switch (__pyx_v_whence) {
      case RW_SEEK_SET:

      /* "pygame_sdl2/rwobject.pyx":205
 * 
 *     if whence == RW_SEEK_SET:
 *         newpos = bf.base + offset             # <<<<<<<<<<<<<<
 *     elif whence == RW_SEEK_CUR:
 *         newpos = bf.here + offset
 */
      __pyx_v_newpos = (__pyx_v_bf->base + __pyx_v_offset);

      /* "pygame_sdl2/rwobject.pyx":204
 *     cdef Uint8 *newpos
 * 
 *     if whence == RW_SEEK_SET:             # <<<<<<<<<<<<<<
 *         newpos = bf.base + offset
 *     elif whence == RW_SEEK_CUR:
 */
      break;
      case RW_SEEK_CUR:

      /* "pygame_sdl2/rwobject.pyx":207
 *         newpos = bf.base + offset
 *     elif whence == RW_SEEK_CUR:
 *         newpos = bf.here + offset             # <<<<<<<<<<<<<<
 *     elif whence == RW_SEEK_END:
 *         newpos = bf.stop + offset
 */
      __pyx_v_newpos = (__pyx_v_bf->here + __pyx_v_offset);

      /* "pygame_sdl2/rwobject.pyx":206
 *     if whence == RW_SEEK_SET:
 *         newpos = bf.base + offset
 *     elif whence == RW_SEEK_CUR:             # <<<<<<<<<<<<<<
 *         newpos = bf.here + offset
 *     elif whence == RW_SEEK_END:
 */
      break;
      case RW_SEEK_END:

      /* "pygame_sdl2/rwobject.pyx":209
 *         newpos = bf.here + offset
 *     elif whence == RW_SEEK_END:
 *         newpos = bf.stop + offset             # <<<<<<<<<<<<<<
 *     else:
 *         with gil:
 */
      __pyx_v_newpos = (__pyx_v_bf->stop + __pyx_v_offset);

      /* "pygame_sdl2/rwobject.pyx":208
 *     elif whence == RW_SEEK_CUR:
 *         newpos = bf.here + offset
 *     elif whence == RW_SEEK_END:             # <<<<<<<<<<<<<<
 *         newpos = bf.stop + offset
 *     else:
 */
      break;
      default:

      /* "pygame_sdl2/rwobject.pyx":211
 *         newpos = bf.stop + offset
 *     else:
 *         with gil:             # <<<<<<<<<<<<<<
 *             set_error("Unknown value for 'whence'")
 *         return -1
 */
      {
          #ifdef WITH_THREAD
          PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
          #endif
          /*try:*/ {

            /* "pygame_sdl2/rwobject.pyx":212
 *     else:
 *         with gil:
 *             set_error("Unknown value for 'whence'")             # <<<<<<<<<<<<<<
 *         return -1
 *     if newpos < bf.base:
 */
            __pyx_t_1 = __pyx_f_11pygame_sdl2_8rwobject_set_error(__pyx_kp_s_Unknown_value_for_whence); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }

          /* "pygame_sdl2/rwobject.pyx":211
 *         newpos = bf.stop + offset
 *     else:
 *         with gil:             # <<<<<<<<<<<<<<
 *             set_error("Unknown value for 'whence'")
 *         return -1
 */
          /*finally:*/ {
            /*normal exit:*/{
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              goto __pyx_L8;
            }
            __pyx_L7_error: {
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              goto __pyx_L4_error;
            }
            __pyx_L8:;
          }
      }

      /* "pygame_sdl2/rwobject.pyx":213
 *         with gil:
 *             set_error("Unknown value for 'whence'")
 *         return -1             # <<<<<<<<<<<<<<
 *     if newpos < bf.base:
 *         newpos = bf.base
 */
      __pyx_r = -1L;
      goto __pyx_L3_return;
      break;
    }

    /* "pygame_sdl2/rwobject.pyx":214
 *             set_error("Unknown value for 'whence'")
 *         return -1
 *     if newpos < bf.base:             # <<<<<<<<<<<<<<
 *         newpos = bf.base
 *     if newpos > bf.stop:
 */
    __pyx_t_2 = ((__pyx_v_newpos < __pyx_v_bf->base) != 0);
    if (__pyx_t_2) {

      /* "pygame_sdl2/rwobject.pyx":215
 *         return -1
 *     if newpos < bf.base:
 *         newpos = bf.base             # <<<<<<<<<<<<<<
 *     if newpos > bf.stop:
 *         newpos = bf.stop
 */
      __pyx_t_3 = __pyx_v_bf->base;
      __pyx_v_newpos = __pyx_t_3;

      /* "pygame_sdl2/rwobject.pyx":214
 *             set_error("Unknown value for 'whence'")
 *         return -1
 *     if newpos < bf.base:             # <<<<<<<<<<<<<<
 *         newpos = bf.base
 *     if newpos > bf.stop:
 */
    }

    /* "pygame_sdl2/rwobject.pyx":216
 *     if newpos < bf.base:
 *         newpos = bf.base
 *     if newpos > bf.stop:             # <<<<<<<<<<<<<<
 *         newpos = bf.stop
 *     bf.here = newpos
 */
    __pyx_t_2 = ((__pyx_v_newpos > __pyx_v_bf->stop) != 0);
    if (__pyx_t_2) {

      /* "pygame_sdl2/rwobject.pyx":217
 *         newpos = bf.base
 *     if newpos > bf.stop:
 *         newpos = bf.stop             # <<<<<<<<<<<<<<
 *     bf.here = newpos
 * 
 */
      __pyx_t_3 = __pyx_v_bf->stop;
      __pyx_v_newpos = __pyx_t_3;

      /* "pygame_sdl2/rwobject.pyx":216
 *     if newpos < bf.base:
 *         newpos = bf.base
 *     if newpos > bf.stop:             # <<<<<<<<<<<<<<
 *         newpos = bf.stop
 *     bf.here = newpos
 */
    }

    /* "pygame_sdl2/rwobject.pyx":218
 *     if newpos > bf.stop:
 *         newpos = bf.stop
 *     bf.here = newpos             # <<<<<<<<<<<<<<
 * 
 *     return bf.here - bf.base
 */
    __pyx_v_bf->here = __pyx_v_newpos;

    /* "pygame_sdl2/rwobject.pyx":220
 *     bf.here = newpos
 * 
 *     return bf.here - bf.base             # <<<<<<<<<<<<<<
 * 
 * cdef size_t buffile_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) nogil:
 */
    __pyx_r = (__pyx_v_bf->here - __pyx_v_bf->base);
    goto __pyx_L3_return;
  }

  /* "pygame_sdl2/rwobject.pyx":200
 * 
 * cdef Sint64 buffile_seek(SDL_RWops *context, Sint64 offset, int whence) nogil:
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 * 
 *     cdef Uint8 *newpos
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "pygame_sdl2/rwobject.pyx":199
 *     return bf.stop - bf.base
 * 
 * cdef Sint64 buffile_seek(SDL_RWops *context, Sint64 offset, int whence) nogil:             # <<<<<<<<<<<<<<
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("pygame_sdl2.rwobject.buffile_seek", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":222
 *     return bf.here - bf.base
 * 
 * cdef size_t buffile_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) nogil:             # <<<<<<<<<<<<<<
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1
 *     cdef size_t total_bytes = 0
 */

static size_t __pyx_f_11pygame_sdl2_8rwobject_buffile_read(SDL_RWops *__pyx_v_context, void *__pyx_v_ptr, size_t __pyx_v_size, size_t __pyx_v_maxnum) {
  struct __pyx_t_11pygame_sdl2_8rwobject_BufFile *__pyx_v_bf;
  size_t __pyx_v_total_bytes;
  size_t __pyx_v_mem_available;
  size_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pygame_sdl2/rwobject.pyx":223
 * 
 * cdef size_t buffile_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) nogil:
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 *     cdef size_t total_bytes = 0
 *     cdef size_t mem_available = 0
 */
  __pyx_v_bf = ((struct __pyx_t_11pygame_sdl2_8rwobject_BufFile *)__pyx_v_context->hidden.unknown.data1);

  /* "pygame_sdl2/rwobject.pyx":224
 * cdef size_t buffile_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) nogil:
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1
 *     cdef size_t total_bytes = 0             # <<<<<<<<<<<<<<
 *     cdef size_t mem_available = 0
 * 
 */
  __pyx_v_total_bytes = 0;

  /* "pygame_sdl2/rwobject.pyx":225
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1
 *     cdef size_t total_bytes = 0
 *     cdef size_t mem_available = 0             # <<<<<<<<<<<<<<
 * 
 *     total_bytes = maxnum * size
 */
  __pyx_v_mem_available = 0;

  /* "pygame_sdl2/rwobject.pyx":227
 *     cdef size_t mem_available = 0
 * 
 *     total_bytes = maxnum * size             # <<<<<<<<<<<<<<
 *     if (maxnum == 0) or (size == 0) or ((total_bytes // maxnum) != size):
 *         return 0
 */
  __pyx_v_total_bytes = (__pyx_v_maxnum * __pyx_v_size);

  /* "pygame_sdl2/rwobject.pyx":228
 * 
 *     total_bytes = maxnum * size
 *     if (maxnum == 0) or (size == 0) or ((total_bytes // maxnum) != size):             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_2 = ((__pyx_v_maxnum == 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_size == 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  if (unlikely(__pyx_v_maxnum == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 228, __pyx_L1_error)
  }
  __pyx_t_2 = (((__pyx_v_total_bytes / __pyx_v_maxnum) != __pyx_v_size) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "pygame_sdl2/rwobject.pyx":229
 *     total_bytes = maxnum * size
 *     if (maxnum == 0) or (size == 0) or ((total_bytes // maxnum) != size):
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     mem_available = bf.stop - bf.here
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "pygame_sdl2/rwobject.pyx":228
 * 
 *     total_bytes = maxnum * size
 *     if (maxnum == 0) or (size == 0) or ((total_bytes // maxnum) != size):             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  }

  /* "pygame_sdl2/rwobject.pyx":231
 *         return 0
 * 
 *     mem_available = bf.stop - bf.here             # <<<<<<<<<<<<<<
 *     if total_bytes > mem_available:
 *         total_bytes = mem_available
 */
  __pyx_v_mem_available = (__pyx_v_bf->stop - __pyx_v_bf->here);

  /* "pygame_sdl2/rwobject.pyx":232
 * 
 *     mem_available = bf.stop - bf.here
 *     if total_bytes > mem_available:             # <<<<<<<<<<<<<<
 *         total_bytes = mem_available
 * 
 */
  __pyx_t_1 = ((__pyx_v_total_bytes > __pyx_v_mem_available) != 0);
  if (__pyx_t_1) {

    /* "pygame_sdl2/rwobject.pyx":233
 *     mem_available = bf.stop - bf.here
 *     if total_bytes > mem_available:
 *         total_bytes = mem_available             # <<<<<<<<<<<<<<
 * 
 *     SDL_memcpy(ptr, bf.here, total_bytes)
 */
    __pyx_v_total_bytes = __pyx_v_mem_available;

    /* "pygame_sdl2/rwobject.pyx":232
 * 
 *     mem_available = bf.stop - bf.here
 *     if total_bytes > mem_available:             # <<<<<<<<<<<<<<
 *         total_bytes = mem_available
 * 
 */
  }

  /* "pygame_sdl2/rwobject.pyx":235
 *         total_bytes = mem_available
 * 
 *     SDL_memcpy(ptr, bf.here, total_bytes)             # <<<<<<<<<<<<<<
 *     bf.here += total_bytes
 * 
 */
  (void)(SDL_memcpy(__pyx_v_ptr, __pyx_v_bf->here, __pyx_v_total_bytes));

  /* "pygame_sdl2/rwobject.pyx":236
 * 
 *     SDL_memcpy(ptr, bf.here, total_bytes)
 *     bf.here += total_bytes             # <<<<<<<<<<<<<<
 * 
 *     return (total_bytes // size)
 */
  __pyx_v_bf->here = (__pyx_v_bf->here + __pyx_v_total_bytes);

  /* "pygame_sdl2/rwobject.pyx":238
 *     bf.here += total_bytes
 * 
 *     return (total_bytes // size)             # <<<<<<<<<<<<<<
 * 
 * cdef size_t buffile_write(SDL_RWops *context, const void *ptr, size_t size, size_t num) nogil:
 */
  if (unlikely(__pyx_v_size == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 238, __pyx_L1_error)
  }
  __pyx_r = (__pyx_v_total_bytes / __pyx_v_size);
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":222
 *     return bf.here - bf.base
 * 
 * cdef size_t buffile_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) nogil:             # <<<<<<<<<<<<<<
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1
 *     cdef size_t total_bytes = 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("pygame_sdl2.rwobject.buffile_read", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":240
 *     return (total_bytes // size)
 * 
 * cdef size_t buffile_write(SDL_RWops *context, const void *ptr, size_t size, size_t num) nogil:             # <<<<<<<<<<<<<<
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1
 * 
 */

static size_t __pyx_f_11pygame_sdl2_8rwobject_buffile_write(SDL_RWops *__pyx_v_context, void const *__pyx_v_ptr, size_t __pyx_v_size, size_t __pyx_v_num) {
  struct __pyx_t_11pygame_sdl2_8rwobject_BufFile *__pyx_v_bf;
  size_t __pyx_r;
  int __pyx_t_1;
  ptrdiff_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pygame_sdl2/rwobject.pyx":241
 * 
 * cdef size_t buffile_write(SDL_RWops *context, const void *ptr, size_t size, size_t num) nogil:
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 * 
 *     if bf.view.readonly != 0:
 */
  __pyx_v_bf = ((struct __pyx_t_11pygame_sdl2_8rwobject_BufFile *)__pyx_v_context->hidden.unknown.data1);

  /* "pygame_sdl2/rwobject.pyx":243
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1
 * 
 *     if bf.view.readonly != 0:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_1 = ((__pyx_v_bf->view.readonly != 0) != 0);
  if (__pyx_t_1) {

    /* "pygame_sdl2/rwobject.pyx":244
 * 
 *     if bf.view.readonly != 0:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     if (bf.here + (num * size)) > bf.stop:
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "pygame_sdl2/rwobject.pyx":243
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1
 * 
 *     if bf.view.readonly != 0:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  }

  /* "pygame_sdl2/rwobject.pyx":246
 *         return 0
 * 
 *     if (bf.here + (num * size)) > bf.stop:             # <<<<<<<<<<<<<<
 *         num = (bf.stop - bf.here) // size
 *     SDL_memcpy(bf.here, ptr, num * size)
 */
  __pyx_t_1 = (((__pyx_v_bf->here + (__pyx_v_num * __pyx_v_size)) > __pyx_v_bf->stop) != 0);
  if (__pyx_t_1) {

    /* "pygame_sdl2/rwobject.pyx":247
 * 
 *     if (bf.here + (num * size)) > bf.stop:
 *         num = (bf.stop - bf.here) // size             # <<<<<<<<<<<<<<
 *     SDL_memcpy(bf.here, ptr, num * size)
 *     bf.here += num * size
 */
    __pyx_t_2 = (__pyx_v_bf->stop - __pyx_v_bf->here);
    if (unlikely(__pyx_v_size == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 247, __pyx_L1_error)
    }
    else if (sizeof(ptrdiff_t) == sizeof(long) && (!(((size_t)-1) > 0)) && unlikely(__pyx_v_size == (size_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_t_2))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 247, __pyx_L1_error)
    }
    __pyx_v_num = __Pyx_div_ptrdiff_t(__pyx_t_2, __pyx_v_size);

    /* "pygame_sdl2/rwobject.pyx":246
 *         return 0
 * 
 *     if (bf.here + (num * size)) > bf.stop:             # <<<<<<<<<<<<<<
 *         num = (bf.stop - bf.here) // size
 *     SDL_memcpy(bf.here, ptr, num * size)
 */
  }

  /* "pygame_sdl2/rwobject.pyx":248
 *     if (bf.here + (num * size)) > bf.stop:
 *         num = (bf.stop - bf.here) // size
 *     SDL_memcpy(bf.here, ptr, num * size)             # <<<<<<<<<<<<<<
 *     bf.here += num * size
 * 
 */
  (void)(SDL_memcpy(__pyx_v_bf->here, __pyx_v_ptr, (__pyx_v_num * __pyx_v_size)));

  /* "pygame_sdl2/rwobject.pyx":249
 *         num = (bf.stop - bf.here) // size
 *     SDL_memcpy(bf.here, ptr, num * size)
 *     bf.here += num * size             # <<<<<<<<<<<<<<
 * 
 *     return num
 */
  __pyx_v_bf->here = (__pyx_v_bf->here + (__pyx_v_num * __pyx_v_size));

  /* "pygame_sdl2/rwobject.pyx":251
 *     bf.here += num * size
 * 
 *     return num             # <<<<<<<<<<<<<<
 * 
 * cdef int buffile_close(SDL_RWops *context) with gil:
 */
  __pyx_r = __pyx_v_num;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":240
 *     return (total_bytes // size)
 * 
 * cdef size_t buffile_write(SDL_RWops *context, const void *ptr, size_t size, size_t num) nogil:             # <<<<<<<<<<<<<<
 *     cdef BufFile *bf = <BufFile *> context.hidden.unknown.data1
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("pygame_sdl2.rwobject.buffile_write", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":253
 *     return num
 * 
 * cdef int buffile_close(SDL_RWops *context) with gil:             # <<<<<<<<<<<<<<
 *     cdef BufFile *bf
 * 
 */

static int __pyx_f_11pygame_sdl2_8rwobject_buffile_close(SDL_RWops *__pyx_v_context) {
  struct __pyx_t_11pygame_sdl2_8rwobject_BufFile *__pyx_v_bf;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("buffile_close", 0);

  /* "pygame_sdl2/rwobject.pyx":256
 *     cdef BufFile *bf
 * 
 *     if context != NULL:             # <<<<<<<<<<<<<<
 *         bf = <BufFile *> context.hidden.unknown.data1
 *         if bf != NULL:
 */
  __pyx_t_1 = ((__pyx_v_context != NULL) != 0);
  if (__pyx_t_1) {

    /* "pygame_sdl2/rwobject.pyx":257
 * 
 *     if context != NULL:
 *         bf = <BufFile *> context.hidden.unknown.data1             # <<<<<<<<<<<<<<
 *         if bf != NULL:
 *             PyBuffer_Release(&bf.view)
 */
    __pyx_v_bf = ((struct __pyx_t_11pygame_sdl2_8rwobject_BufFile *)__pyx_v_context->hidden.unknown.data1);

    /* "pygame_sdl2/rwobject.pyx":258
 *     if context != NULL:
 *         bf = <BufFile *> context.hidden.unknown.data1
 *         if bf != NULL:             # <<<<<<<<<<<<<<
 *             PyBuffer_Release(&bf.view)
 *             free(bf)
 */
    __pyx_t_1 = ((__pyx_v_bf != NULL) != 0);
    if (__pyx_t_1) {

      /* "pygame_sdl2/rwobject.pyx":259
 *         bf = <BufFile *> context.hidden.unknown.data1
 *         if bf != NULL:
 *             PyBuffer_Release(&bf.view)             # <<<<<<<<<<<<<<
 *             free(bf)
 *             bf = NULL
 */
      PyBuffer_Release((&__pyx_v_bf->view));

      /* "pygame_sdl2/rwobject.pyx":260
 *         if bf != NULL:
 *             PyBuffer_Release(&bf.view)
 *             free(bf)             # <<<<<<<<<<<<<<
 *             bf = NULL
 *         SDL_FreeRW(context)
 */
      free(__pyx_v_bf);

      /* "pygame_sdl2/rwobject.pyx":261
 *             PyBuffer_Release(&bf.view)
 *             free(bf)
 *             bf = NULL             # <<<<<<<<<<<<<<
 *         SDL_FreeRW(context)
 * 
 */
      __pyx_v_bf = NULL;

      /* "pygame_sdl2/rwobject.pyx":258
 *     if context != NULL:
 *         bf = <BufFile *> context.hidden.unknown.data1
 *         if bf != NULL:             # <<<<<<<<<<<<<<
 *             PyBuffer_Release(&bf.view)
 *             free(bf)
 */
    }

    /* "pygame_sdl2/rwobject.pyx":262
 *             free(bf)
 *             bf = NULL
 *         SDL_FreeRW(context)             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    SDL_FreeRW(__pyx_v_context);

    /* "pygame_sdl2/rwobject.pyx":256
 *     cdef BufFile *bf
 * 
 *     if context != NULL:             # <<<<<<<<<<<<<<
 *         bf = <BufFile *> context.hidden.unknown.data1
 *         if bf != NULL:
 */
  }

  /* "pygame_sdl2/rwobject.pyx":264
 *         SDL_FreeRW(context)
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef SDL_RWops *to_rwops(filelike, mode="rb") except NULL:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":253
 *     return num
 * 
 * cdef int buffile_close(SDL_RWops *context) with gil:             # <<<<<<<<<<<<<<
 *     cdef BufFile *bf
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":266
 *     return 0
 * 
 * cdef SDL_RWops *to_rwops(filelike, mode="rb") except NULL:             # <<<<<<<<<<<<<<
 * 
 *     cdef FILE *f
 */

static SDL_RWops *__pyx_f_11pygame_sdl2_8rwobject_to_rwops(PyObject *__pyx_v_filelike, struct __pyx_opt_args_11pygame_sdl2_8rwobject_to_rwops *__pyx_optional_args) {
  PyObject *__pyx_v_mode = ((PyObject *)__pyx_n_s_rb);
  struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *__pyx_v_sf;
  SDL_RWops *__pyx_v_rv;
  SDL_RWops *__pyx_v_rw;
  char *__pyx_v_cname;
  char *__pyx_v_cmode;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_dname = NULL;
  PyObject *__pyx_v_base = NULL;
  PyObject *__pyx_v_length = NULL;
  SDL_RWops *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Sint64 __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_rwops", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_mode = __pyx_optional_args->mode;
    }
  }
  __Pyx_INCREF(__pyx_v_mode);

  /* "pygame_sdl2/rwobject.pyx":275
 *     cdef char *cmode
 * 
 *     if not isinstance(mode, bytes_):             # <<<<<<<<<<<<<<
 *         mode = mode.encode("ascii")
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_bytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_mode, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_3) {

    /* "pygame_sdl2/rwobject.pyx":276
 * 
 *     if not isinstance(mode, bytes_):
 *         mode = mode.encode("ascii")             # <<<<<<<<<<<<<<
 * 
 *     name = filelike
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_mode, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_n_s_ascii) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_s_ascii);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_mode, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pygame_sdl2/rwobject.pyx":275
 *     cdef char *cmode
 * 
 *     if not isinstance(mode, bytes_):             # <<<<<<<<<<<<<<
 *         mode = mode.encode("ascii")
 * 
 */
  }

  /* "pygame_sdl2/rwobject.pyx":278
 *         mode = mode.encode("ascii")
 * 
 *     name = filelike             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(filelike, RWops):
 */
  __Pyx_INCREF(__pyx_v_filelike);
  __pyx_v_name = __pyx_v_filelike;

  /* "pygame_sdl2/rwobject.pyx":280
 *     name = filelike
 * 
 *     if isinstance(filelike, RWops):             # <<<<<<<<<<<<<<
 *         rv = (<RWopsImpl>filelike._holder).get_rwops()
 *         if rv == NULL:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_RWops); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_filelike, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "pygame_sdl2/rwobject.pyx":281
 * 
 *     if isinstance(filelike, RWops):
 *         rv = (<RWopsImpl>filelike._holder).get_rwops()             # <<<<<<<<<<<<<<
 *         if rv == NULL:
 *             raise ValueError("Passed in RWops object is closed")
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_filelike, __pyx_n_s_holder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_rv = ((struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *)((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1)->__pyx_vtab)->get_rwops(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1));
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pygame_sdl2/rwobject.pyx":282
 *     if isinstance(filelike, RWops):
 *         rv = (<RWopsImpl>filelike._holder).get_rwops()
 *         if rv == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Passed in RWops object is closed")
 *         (<RWopsImpl>filelike._holder).clear_rwops()
 */
    __pyx_t_2 = ((__pyx_v_rv == NULL) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "pygame_sdl2/rwobject.pyx":283
 *         rv = (<RWopsImpl>filelike._holder).get_rwops()
 *         if rv == NULL:
 *             raise ValueError("Passed in RWops object is closed")             # <<<<<<<<<<<<<<
 *         (<RWopsImpl>filelike._holder).clear_rwops()
 *         return rv
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 283, __pyx_L1_error)

      /* "pygame_sdl2/rwobject.pyx":282
 *     if isinstance(filelike, RWops):
 *         rv = (<RWopsImpl>filelike._holder).get_rwops()
 *         if rv == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Passed in RWops object is closed")
 *         (<RWopsImpl>filelike._holder).clear_rwops()
 */
    }

    /* "pygame_sdl2/rwobject.pyx":284
 *         if rv == NULL:
 *             raise ValueError("Passed in RWops object is closed")
 *         (<RWopsImpl>filelike._holder).clear_rwops()             # <<<<<<<<<<<<<<
 *         return rv
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_filelike, __pyx_n_s_holder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = ((struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *)((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1)->__pyx_vtab)->clear_rwops(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pygame_sdl2/rwobject.pyx":285
 *             raise ValueError("Passed in RWops object is closed")
 *         (<RWopsImpl>filelike._holder).clear_rwops()
 *         return rv             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(filelike, (file_type, io.IOBase)) and mode == "rb":
 */
    __pyx_r = __pyx_v_rv;
    goto __pyx_L0;

    /* "pygame_sdl2/rwobject.pyx":280
 *     name = filelike
 * 
 *     if isinstance(filelike, RWops):             # <<<<<<<<<<<<<<
 *         rv = (<RWopsImpl>filelike._holder).get_rwops()
 *         if rv == NULL:
 */
  }

  /* "pygame_sdl2/rwobject.pyx":287
 *         return rv
 * 
 *     if isinstance(filelike, (file_type, io.IOBase)) and mode == "rb":             # <<<<<<<<<<<<<<
 *         name = getattr(filelike, "name", None)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_file_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_io); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_IOBase); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = PyObject_IsInstance(__pyx_v_filelike, __pyx_t_4); 
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_3 = __pyx_t_7;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_7 = PyObject_IsInstance(__pyx_v_filelike, __pyx_t_5); 
  __pyx_t_6 = (__pyx_t_7 != 0);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L9_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = (__pyx_t_3 != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_rb, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 287, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_6;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {

    /* "pygame_sdl2/rwobject.pyx":288
 * 
 *     if isinstance(filelike, (file_type, io.IOBase)) and mode == "rb":
 *         name = getattr(filelike, "name", None)             # <<<<<<<<<<<<<<
 * 
 *     # Try to open as a file.
 */
    __pyx_t_4 = __Pyx_GetAttr3(__pyx_v_filelike, __pyx_n_s_name, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_name, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "pygame_sdl2/rwobject.pyx":287
 *         return rv
 * 
 *     if isinstance(filelike, (file_type, io.IOBase)) and mode == "rb":             # <<<<<<<<<<<<<<
 *         name = getattr(filelike, "name", None)
 * 
 */
  }

  /* "pygame_sdl2/rwobject.pyx":291
 * 
 *     # Try to open as a file.
 *     if isinstance(name, bytes_):             # <<<<<<<<<<<<<<
 *         name = name.decode(fsencoding)
 *     elif isinstance(name, unicode_):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_bytes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_name, __pyx_t_4); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = (__pyx_t_2 != 0);
  if (__pyx_t_6) {

    /* "pygame_sdl2/rwobject.pyx":292
 *     # Try to open as a file.
 *     if isinstance(name, bytes_):
 *         name = name.decode(fsencoding)             # <<<<<<<<<<<<<<
 *     elif isinstance(name, unicode_):
 *         pass
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_decode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_fsencoding); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_name, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "pygame_sdl2/rwobject.pyx":291
 * 
 *     # Try to open as a file.
 *     if isinstance(name, bytes_):             # <<<<<<<<<<<<<<
 *         name = name.decode(fsencoding)
 *     elif isinstance(name, unicode_):
 */
    goto __pyx_L11;
  }

  /* "pygame_sdl2/rwobject.pyx":293
 *     if isinstance(name, bytes_):
 *         name = name.decode(fsencoding)
 *     elif isinstance(name, unicode_):             # <<<<<<<<<<<<<<
 *         pass
 *     else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyObject_IsInstance(__pyx_v_name, __pyx_t_4); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = (__pyx_t_6 != 0);
  if (__pyx_t_2) {
    goto __pyx_L11;
  }

  /* "pygame_sdl2/rwobject.pyx":296
 *         pass
 *     else:
 *         name = None             # <<<<<<<<<<<<<<
 * 
 *     if (mode == b"rb") and (name is not None):
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_name, Py_None);
  }
  __pyx_L11:;

  /* "pygame_sdl2/rwobject.pyx":298
 *         name = None
 * 
 *     if (mode == b"rb") and (name is not None):             # <<<<<<<<<<<<<<
 * 
 *         dname = name.encode("utf-8")
 */
  __pyx_t_6 = (__Pyx_PyBytes_Equals(__pyx_v_mode, __pyx_n_b_rb, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 298, __pyx_L1_error)
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L13_bool_binop_done;
  }
  __pyx_t_6 = (__pyx_v_name != Py_None);
  __pyx_t_3 = (__pyx_t_6 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L13_bool_binop_done:;
  if (__pyx_t_2) {

    /* "pygame_sdl2/rwobject.pyx":300
 *     if (mode == b"rb") and (name is not None):
 * 
 *         dname = name.encode("utf-8")             # <<<<<<<<<<<<<<
 *         cname = dname
 *         cmode = mode
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_s_utf_8);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dname = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "pygame_sdl2/rwobject.pyx":301
 * 
 *         dname = name.encode("utf-8")
 *         cname = dname             # <<<<<<<<<<<<<<
 *         cmode = mode
 * 
 */
    __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_v_dname); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 301, __pyx_L1_error)
    __pyx_v_cname = __pyx_t_9;

    /* "pygame_sdl2/rwobject.pyx":302
 *         dname = name.encode("utf-8")
 *         cname = dname
 *         cmode = mode             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
 */
    __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_v_mode); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 302, __pyx_L1_error)
    __pyx_v_cmode = __pyx_t_9;

    /* "pygame_sdl2/rwobject.pyx":304
 *         cmode = mode
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             rv = SDL_RWFromFile(cname, cmode)
 * 
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {

          /* "pygame_sdl2/rwobject.pyx":305
 * 
 *         with nogil:
 *             rv = SDL_RWFromFile(cname, cmode)             # <<<<<<<<<<<<<<
 * 
 *         if rv == NULL:
 */
          __pyx_v_rv = SDL_RWFromFile(__pyx_v_cname, __pyx_v_cmode);
        }

        /* "pygame_sdl2/rwobject.pyx":304
 *         cmode = mode
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             rv = SDL_RWFromFile(cname, cmode)
 * 
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L17;
          }
          __pyx_L17:;
        }
    }

    /* "pygame_sdl2/rwobject.pyx":307
 *             rv = SDL_RWFromFile(cname, cmode)
 * 
 *         if rv == NULL:             # <<<<<<<<<<<<<<
 *             raise IOError("Could not open {!r}: {}".format(name, SDL_GetError()))
 * 
 */
    __pyx_t_2 = ((__pyx_v_rv == NULL) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "pygame_sdl2/rwobject.pyx":308
 * 
 *         if rv == NULL:
 *             raise IOError("Could not open {!r}: {}".format(name, SDL_GetError()))             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Could_not_open_r, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = __Pyx_PyBytes_FromString(SDL_GetError()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_10 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_name, __pyx_t_1};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 308, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_name, __pyx_t_1};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 308, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 308, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_8) {
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_8); __pyx_t_8 = NULL;
        }
        __Pyx_INCREF(__pyx_v_name);
        __Pyx_GIVEREF(__pyx_v_name);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_v_name);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_t_1);
        __pyx_t_1 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 308, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IOError, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 308, __pyx_L1_error)

      /* "pygame_sdl2/rwobject.pyx":307
 *             rv = SDL_RWFromFile(cname, cmode)
 * 
 *         if rv == NULL:             # <<<<<<<<<<<<<<
 *             raise IOError("Could not open {!r}: {}".format(name, SDL_GetError()))
 * 
 */
    }

    /* "pygame_sdl2/rwobject.pyx":310
 *             raise IOError("Could not open {!r}: {}".format(name, SDL_GetError()))
 * 
 *         try:             # <<<<<<<<<<<<<<
 * 
 *             # If we have these properties, we're either an APK asset or a Ren'Py-style
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      /*try:*/ {

        /* "pygame_sdl2/rwobject.pyx":314
 *             # If we have these properties, we're either an APK asset or a Ren'Py-style
 *             # subfile, so use an optimized path.
 *             base = filelike.base             # <<<<<<<<<<<<<<
 *             length = filelike.length
 * 
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_filelike, __pyx_n_s_base); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 314, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_v_base = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "pygame_sdl2/rwobject.pyx":315
 *             # subfile, so use an optimized path.
 *             base = filelike.base
 *             length = filelike.length             # <<<<<<<<<<<<<<
 * 
 *             rw = rv
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_filelike, __pyx_n_s_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 315, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_v_length = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "pygame_sdl2/rwobject.pyx":317
 *             length = filelike.length
 * 
 *             rw = rv             # <<<<<<<<<<<<<<
 * 
 *             SDL_RWseek(rw, base, RW_SEEK_SET);
 */
        __pyx_v_rw = __pyx_v_rv;

        /* "pygame_sdl2/rwobject.pyx":319
 *             rw = rv
 * 
 *             SDL_RWseek(rw, base, RW_SEEK_SET);             # <<<<<<<<<<<<<<
 * 
 *             sf = <SubFile *> calloc(sizeof(SubFile), 1)
 */
        __pyx_t_15 = __Pyx_PyInt_As_int64_t(__pyx_v_base); if (unlikely((__pyx_t_15 == ((Sint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 319, __pyx_L19_error)
        (void)(SDL_RWseek(__pyx_v_rw, __pyx_t_15, RW_SEEK_SET));

        /* "pygame_sdl2/rwobject.pyx":321
 *             SDL_RWseek(rw, base, RW_SEEK_SET);
 * 
 *             sf = <SubFile *> calloc(sizeof(SubFile), 1)             # <<<<<<<<<<<<<<
 *             sf.rw = rw
 *             sf.base = base
 */
        __pyx_v_sf = ((struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *)calloc((sizeof(struct __pyx_t_11pygame_sdl2_8rwobject_SubFile)), 1));

        /* "pygame_sdl2/rwobject.pyx":322
 * 
 *             sf = <SubFile *> calloc(sizeof(SubFile), 1)
 *             sf.rw = rw             # <<<<<<<<<<<<<<
 *             sf.base = base
 *             sf.length = length
 */
        __pyx_v_sf->rw = __pyx_v_rw;

        /* "pygame_sdl2/rwobject.pyx":323
 *             sf = <SubFile *> calloc(sizeof(SubFile), 1)
 *             sf.rw = rw
 *             sf.base = base             # <<<<<<<<<<<<<<
 *             sf.length = length
 *             sf.tell = 0;
 */
        __pyx_t_15 = __Pyx_PyInt_As_int64_t(__pyx_v_base); if (unlikely((__pyx_t_15 == ((Sint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 323, __pyx_L19_error)
        __pyx_v_sf->base = __pyx_t_15;

        /* "pygame_sdl2/rwobject.pyx":324
 *             sf.rw = rw
 *             sf.base = base
 *             sf.length = length             # <<<<<<<<<<<<<<
 *             sf.tell = 0;
 * 
 */
        __pyx_t_15 = __Pyx_PyInt_As_int64_t(__pyx_v_length); if (unlikely((__pyx_t_15 == ((Sint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 324, __pyx_L19_error)
        __pyx_v_sf->length = __pyx_t_15;

        /* "pygame_sdl2/rwobject.pyx":325
 *             sf.base = base
 *             sf.length = length
 *             sf.tell = 0;             # <<<<<<<<<<<<<<
 * 
 *             SDL_RWseek(rw, base, RW_SEEK_SET)
 */
        __pyx_v_sf->tell = 0;

        /* "pygame_sdl2/rwobject.pyx":327
 *             sf.tell = 0;
 * 
 *             SDL_RWseek(rw, base, RW_SEEK_SET)             # <<<<<<<<<<<<<<
 * 
 *             sf = <SubFile *> calloc(sizeof(SubFile), 1)
 */
        __pyx_t_15 = __Pyx_PyInt_As_int64_t(__pyx_v_base); if (unlikely((__pyx_t_15 == ((Sint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 327, __pyx_L19_error)
        (void)(SDL_RWseek(__pyx_v_rw, __pyx_t_15, RW_SEEK_SET));

        /* "pygame_sdl2/rwobject.pyx":329
 *             SDL_RWseek(rw, base, RW_SEEK_SET)
 * 
 *             sf = <SubFile *> calloc(sizeof(SubFile), 1)             # <<<<<<<<<<<<<<
 *             sf.rw = rw
 *             sf.base = base
 */
        __pyx_v_sf = ((struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *)calloc((sizeof(struct __pyx_t_11pygame_sdl2_8rwobject_SubFile)), 1));

        /* "pygame_sdl2/rwobject.pyx":330
 * 
 *             sf = <SubFile *> calloc(sizeof(SubFile), 1)
 *             sf.rw = rw             # <<<<<<<<<<<<<<
 *             sf.base = base
 *             sf.length = length
 */
        __pyx_v_sf->rw = __pyx_v_rw;

        /* "pygame_sdl2/rwobject.pyx":331
 *             sf = <SubFile *> calloc(sizeof(SubFile), 1)
 *             sf.rw = rw
 *             sf.base = base             # <<<<<<<<<<<<<<
 *             sf.length = length
 *             sf.tell = 0
 */
        __pyx_t_15 = __Pyx_PyInt_As_int64_t(__pyx_v_base); if (unlikely((__pyx_t_15 == ((Sint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 331, __pyx_L19_error)
        __pyx_v_sf->base = __pyx_t_15;

        /* "pygame_sdl2/rwobject.pyx":332
 *             sf.rw = rw
 *             sf.base = base
 *             sf.length = length             # <<<<<<<<<<<<<<
 *             sf.tell = 0
 * 
 */
        __pyx_t_15 = __Pyx_PyInt_As_int64_t(__pyx_v_length); if (unlikely((__pyx_t_15 == ((Sint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 332, __pyx_L19_error)
        __pyx_v_sf->length = __pyx_t_15;

        /* "pygame_sdl2/rwobject.pyx":333
 *             sf.base = base
 *             sf.length = length
 *             sf.tell = 0             # <<<<<<<<<<<<<<
 * 
 *             rv = SDL_AllocRW()
 */
        __pyx_v_sf->tell = 0;

        /* "pygame_sdl2/rwobject.pyx":335
 *             sf.tell = 0
 * 
 *             rv = SDL_AllocRW()             # <<<<<<<<<<<<<<
 *             rv.size = subfile_size
 *             rv.seek = subfile_seek
 */
        __pyx_v_rv = SDL_AllocRW();

        /* "pygame_sdl2/rwobject.pyx":336
 * 
 *             rv = SDL_AllocRW()
 *             rv.size = subfile_size             # <<<<<<<<<<<<<<
 *             rv.seek = subfile_seek
 *             rv.read = subfile_read
 */
        __pyx_v_rv->size = __pyx_f_11pygame_sdl2_8rwobject_subfile_size;

        /* "pygame_sdl2/rwobject.pyx":337
 *             rv = SDL_AllocRW()
 *             rv.size = subfile_size
 *             rv.seek = subfile_seek             # <<<<<<<<<<<<<<
 *             rv.read = subfile_read
 *             rv.write = NULL
 */
        __pyx_v_rv->seek = __pyx_f_11pygame_sdl2_8rwobject_subfile_seek;

        /* "pygame_sdl2/rwobject.pyx":338
 *             rv.size = subfile_size
 *             rv.seek = subfile_seek
 *             rv.read = subfile_read             # <<<<<<<<<<<<<<
 *             rv.write = NULL
 *             rv.close = subfile_close
 */
        __pyx_v_rv->read = __pyx_f_11pygame_sdl2_8rwobject_subfile_read;

        /* "pygame_sdl2/rwobject.pyx":339
 *             rv.seek = subfile_seek
 *             rv.read = subfile_read
 *             rv.write = NULL             # <<<<<<<<<<<<<<
 *             rv.close = subfile_close
 *             rv.type = 0
 */
        __pyx_v_rv->write = NULL;

        /* "pygame_sdl2/rwobject.pyx":340
 *             rv.read = subfile_read
 *             rv.write = NULL
 *             rv.close = subfile_close             # <<<<<<<<<<<<<<
 *             rv.type = 0
 *             rv.hidden.unknown.data1 = <void *> sf
 */
        __pyx_v_rv->close = __pyx_f_11pygame_sdl2_8rwobject_subfile_close;

        /* "pygame_sdl2/rwobject.pyx":341
 *             rv.write = NULL
 *             rv.close = subfile_close
 *             rv.type = 0             # <<<<<<<<<<<<<<
 *             rv.hidden.unknown.data1 = <void *> sf
 * 
 */
        __pyx_v_rv->type = 0;

        /* "pygame_sdl2/rwobject.pyx":342
 *             rv.close = subfile_close
 *             rv.type = 0
 *             rv.hidden.unknown.data1 = <void *> sf             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
        __pyx_v_rv->hidden.unknown.data1 = ((void *)__pyx_v_sf);

        /* "pygame_sdl2/rwobject.pyx":344
 *             rv.hidden.unknown.data1 = <void *> sf
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 filelike.close()
 *             except:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          /*try:*/ {

            /* "pygame_sdl2/rwobject.pyx":345
 * 
 *             try:
 *                 filelike.close()             # <<<<<<<<<<<<<<
 *             except:
 *                 pass
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_filelike, __pyx_n_s_close); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 345, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_11 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_11)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_11);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            __pyx_t_5 = (__pyx_t_11) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_11) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 345, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

            /* "pygame_sdl2/rwobject.pyx":344
 *             rv.hidden.unknown.data1 = <void *> sf
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 filelike.close()
 *             except:
 */
          }
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          goto __pyx_L30_try_end;
          __pyx_L25_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "pygame_sdl2/rwobject.pyx":346
 *             try:
 *                 filelike.close()
 *             except:             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          /*except:*/ {
            __Pyx_ErrRestore(0,0,0);
            goto __pyx_L26_exception_handled;
          }
          __pyx_L26_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          __pyx_L30_try_end:;
        }

        /* "pygame_sdl2/rwobject.pyx":349
 *                 pass
 * 
 *             return rv             # <<<<<<<<<<<<<<
 * 
 *         except AttributeError:
 */
        __pyx_r = __pyx_v_rv;
        goto __pyx_L23_try_return;

        /* "pygame_sdl2/rwobject.pyx":310
 *             raise IOError("Could not open {!r}: {}".format(name, SDL_GetError()))
 * 
 *         try:             # <<<<<<<<<<<<<<
 * 
 *             # If we have these properties, we're either an APK asset or a Ren'Py-style
 */
      }
      __pyx_L19_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "pygame_sdl2/rwobject.pyx":351
 *             return rv
 * 
 *         except AttributeError:             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
      if (__pyx_t_10) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L20_exception_handled;
      }
      goto __pyx_L21_except_error;
      __pyx_L21_except_error:;

      /* "pygame_sdl2/rwobject.pyx":310
 *             raise IOError("Could not open {!r}: {}".format(name, SDL_GetError()))
 * 
 *         try:             # <<<<<<<<<<<<<<
 * 
 *             # If we have these properties, we're either an APK asset or a Ren'Py-style
 */
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      goto __pyx_L1_error;
      __pyx_L23_try_return:;
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      goto __pyx_L0;
      __pyx_L20_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
    }

    /* "pygame_sdl2/rwobject.pyx":354
 *             pass
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             filelike.close()
 *         except:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_14, &__pyx_t_13, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_12);
      /*try:*/ {

        /* "pygame_sdl2/rwobject.pyx":355
 * 
 *         try:
 *             filelike.close()             # <<<<<<<<<<<<<<
 *         except:
 *             pass
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_filelike, __pyx_n_s_close); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 355, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_5 = (__pyx_t_11) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_11) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 355, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "pygame_sdl2/rwobject.pyx":354
 *             pass
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             filelike.close()
 *         except:
 */
      }
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      goto __pyx_L36_try_end;
      __pyx_L31_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "pygame_sdl2/rwobject.pyx":356
 *         try:
 *             filelike.close()
 *         except:             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
      /*except:*/ {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L32_exception_handled;
      }
      __pyx_L32_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_13, __pyx_t_12);
      __pyx_L36_try_end:;
    }

    /* "pygame_sdl2/rwobject.pyx":359
 *             pass
 * 
 *         return rv             # <<<<<<<<<<<<<<
 * 
 *     if not (hasattr(filelike, "read") or hasattr(filelike, "write")):
 */
    __pyx_r = __pyx_v_rv;
    goto __pyx_L0;

    /* "pygame_sdl2/rwobject.pyx":298
 *         name = None
 * 
 *     if (mode == b"rb") and (name is not None):             # <<<<<<<<<<<<<<
 * 
 *         dname = name.encode("utf-8")
 */
  }

  /* "pygame_sdl2/rwobject.pyx":361
 *         return rv
 * 
 *     if not (hasattr(filelike, "read") or hasattr(filelike, "write")):             # <<<<<<<<<<<<<<
 *         raise IOError("{!r} is not a filename or file-like object.".format(filelike))
 * 
 */
  __pyx_t_3 = __Pyx_HasAttr(__pyx_v_filelike, __pyx_n_s_read); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 361, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_3 != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L38_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(__pyx_v_filelike, __pyx_n_s_write); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 361, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_6 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L38_bool_binop_done:;
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "pygame_sdl2/rwobject.pyx":362
 * 
 *     if not (hasattr(filelike, "read") or hasattr(filelike, "write")):
 *         raise IOError("{!r} is not a filename or file-like object.".format(filelike))             # <<<<<<<<<<<<<<
 * 
 *     Py_INCREF(filelike)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_r_is_not_a_filename_or_file_lik, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_5 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_11, __pyx_v_filelike) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_filelike);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IOError, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 362, __pyx_L1_error)

    /* "pygame_sdl2/rwobject.pyx":361
 *         return rv
 * 
 *     if not (hasattr(filelike, "read") or hasattr(filelike, "write")):             # <<<<<<<<<<<<<<
 *         raise IOError("{!r} is not a filename or file-like object.".format(filelike))
 * 
 */
  }

  /* "pygame_sdl2/rwobject.pyx":364
 *         raise IOError("{!r} is not a filename or file-like object.".format(filelike))
 * 
 *     Py_INCREF(filelike)             # <<<<<<<<<<<<<<
 * 
 *     rv = SDL_AllocRW()
 */
  Py_INCREF(__pyx_v_filelike);

  /* "pygame_sdl2/rwobject.pyx":366
 *     Py_INCREF(filelike)
 * 
 *     rv = SDL_AllocRW()             # <<<<<<<<<<<<<<
 *     rv.size = python_size
 *     rv.seek = python_seek
 */
  __pyx_v_rv = SDL_AllocRW();

  /* "pygame_sdl2/rwobject.pyx":367
 * 
 *     rv = SDL_AllocRW()
 *     rv.size = python_size             # <<<<<<<<<<<<<<
 *     rv.seek = python_seek
 *     rv.read = python_read
 */
  __pyx_v_rv->size = __pyx_f_11pygame_sdl2_8rwobject_python_size;

  /* "pygame_sdl2/rwobject.pyx":368
 *     rv = SDL_AllocRW()
 *     rv.size = python_size
 *     rv.seek = python_seek             # <<<<<<<<<<<<<<
 *     rv.read = python_read
 *     rv.write = python_write
 */
  __pyx_v_rv->seek = __pyx_f_11pygame_sdl2_8rwobject_python_seek;

  /* "pygame_sdl2/rwobject.pyx":369
 *     rv.size = python_size
 *     rv.seek = python_seek
 *     rv.read = python_read             # <<<<<<<<<<<<<<
 *     rv.write = python_write
 *     rv.close = python_close
 */
  __pyx_v_rv->read = __pyx_f_11pygame_sdl2_8rwobject_python_read;

  /* "pygame_sdl2/rwobject.pyx":370
 *     rv.seek = python_seek
 *     rv.read = python_read
 *     rv.write = python_write             # <<<<<<<<<<<<<<
 *     rv.close = python_close
 *     rv.type = 0
 */
  __pyx_v_rv->write = __pyx_f_11pygame_sdl2_8rwobject_python_write;

  /* "pygame_sdl2/rwobject.pyx":371
 *     rv.read = python_read
 *     rv.write = python_write
 *     rv.close = python_close             # <<<<<<<<<<<<<<
 *     rv.type = 0
 *     rv.hidden.unknown.data1 = <void *> filelike
 */
  __pyx_v_rv->close = __pyx_f_11pygame_sdl2_8rwobject_python_close;

  /* "pygame_sdl2/rwobject.pyx":372
 *     rv.write = python_write
 *     rv.close = python_close
 *     rv.type = 0             # <<<<<<<<<<<<<<
 *     rv.hidden.unknown.data1 = <void *> filelike
 * 
 */
  __pyx_v_rv->type = 0;

  /* "pygame_sdl2/rwobject.pyx":373
 *     rv.close = python_close
 *     rv.type = 0
 *     rv.hidden.unknown.data1 = <void *> filelike             # <<<<<<<<<<<<<<
 * 
 *     return rv
 */
  __pyx_v_rv->hidden.unknown.data1 = ((void *)__pyx_v_filelike);

  /* "pygame_sdl2/rwobject.pyx":375
 *     rv.hidden.unknown.data1 = <void *> filelike
 * 
 *     return rv             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_rv;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":266
 *     return 0
 * 
 * cdef SDL_RWops *to_rwops(filelike, mode="rb") except NULL:             # <<<<<<<<<<<<<<
 * 
 *     cdef FILE *f
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.to_rwops", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_dname);
  __Pyx_XDECREF(__pyx_v_base);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XDECREF(__pyx_v_mode);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":378
 * 
 * 
 * cdef api SDL_RWops *RWopsFromPython(filelike) except NULL:             # <<<<<<<<<<<<<<
 *     return to_rwops(filelike)
 * 
 */

static SDL_RWops *__pyx_f_11pygame_sdl2_8rwobject_RWopsFromPython(PyObject *__pyx_v_filelike) {
  SDL_RWops *__pyx_r;
  __Pyx_RefNannyDeclarations
  SDL_RWops *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("RWopsFromPython", 0);

  /* "pygame_sdl2/rwobject.pyx":379
 * 
 * cdef api SDL_RWops *RWopsFromPython(filelike) except NULL:
 *     return to_rwops(filelike)             # <<<<<<<<<<<<<<
 * 
 * cdef class RWopsImpl(object):
 */
  __pyx_t_1 = __pyx_f_11pygame_sdl2_8rwobject_to_rwops(__pyx_v_filelike, NULL); if (unlikely(__pyx_t_1 == ((SDL_RWops *)NULL))) __PYX_ERR(0, 379, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":378
 * 
 * 
 * cdef api SDL_RWops *RWopsFromPython(filelike) except NULL:             # <<<<<<<<<<<<<<
 *     return to_rwops(filelike)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWopsFromPython", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":385
 *     cdef int closed
 * 
 *     def check_closed(self):             # <<<<<<<<<<<<<<
 *         """
 *         Internal: raise a ValueError if this RWops object is closed
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_1check_closed(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_check_closed[] = "RWopsImpl.check_closed(self)\n\n        Internal: raise a ValueError if this RWops object is closed\n        ";
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_1check_closed(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_closed (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_check_closed(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_check_closed(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_closed", 0);

  /* "pygame_sdl2/rwobject.pyx":389
 *         Internal: raise a ValueError if this RWops object is closed
 *         """
 *         if self.closed != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("I/O operation on closed RWops.")
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->closed != 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pygame_sdl2/rwobject.pyx":390
 *         """
 *         if self.closed != 0:
 *             raise ValueError("I/O operation on closed RWops.")             # <<<<<<<<<<<<<<
 * 
 *     def get_closed(self):
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 390, __pyx_L1_error)

    /* "pygame_sdl2/rwobject.pyx":389
 *         Internal: raise a ValueError if this RWops object is closed
 *         """
 *         if self.closed != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("I/O operation on closed RWops.")
 * 
 */
  }

  /* "pygame_sdl2/rwobject.pyx":385
 *     cdef int closed
 * 
 *     def check_closed(self):             # <<<<<<<<<<<<<<
 *         """
 *         Internal: raise a ValueError if this RWops object is closed
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWopsImpl.check_closed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":392
 *             raise ValueError("I/O operation on closed RWops.")
 * 
 *     def get_closed(self):             # <<<<<<<<<<<<<<
 *         return self.closed != 0
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_3get_closed(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_2get_closed[] = "RWopsImpl.get_closed(self)";
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_3get_closed(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_closed (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_2get_closed(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_2get_closed(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_closed", 0);

  /* "pygame_sdl2/rwobject.pyx":393
 * 
 *     def get_closed(self):
 *         return self.closed != 0             # <<<<<<<<<<<<<<
 * 
 *     cdef SDL_RWops *get_rwops(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->closed != 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":392
 *             raise ValueError("I/O operation on closed RWops.")
 * 
 *     def get_closed(self):             # <<<<<<<<<<<<<<
 *         return self.closed != 0
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWopsImpl.get_closed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":395
 *         return self.closed != 0
 * 
 *     cdef SDL_RWops *get_rwops(self):             # <<<<<<<<<<<<<<
 *         return self.ops
 * 
 */

static SDL_RWops *__pyx_f_11pygame_sdl2_8rwobject_9RWopsImpl_get_rwops(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self) {
  SDL_RWops *__pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_rwops", 0);

  /* "pygame_sdl2/rwobject.pyx":396
 * 
 *     cdef SDL_RWops *get_rwops(self):
 *         return self.ops             # <<<<<<<<<<<<<<
 * 
 *     cdef set_rwops(self, SDL_RWops * ops):
 */
  __pyx_r = __pyx_v_self->ops;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":395
 *         return self.closed != 0
 * 
 *     cdef SDL_RWops *get_rwops(self):             # <<<<<<<<<<<<<<
 *         return self.ops
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":398
 *         return self.ops
 * 
 *     cdef set_rwops(self, SDL_RWops * ops):             # <<<<<<<<<<<<<<
 *         """
 *         Internal: Set the rwops object.
 */

static PyObject *__pyx_f_11pygame_sdl2_8rwobject_9RWopsImpl_set_rwops(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self, SDL_RWops *__pyx_v_ops) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_rwops", 0);

  /* "pygame_sdl2/rwobject.pyx":402
 *         Internal: Set the rwops object.
 *         """
 *         self.ops = ops             # <<<<<<<<<<<<<<
 *         self.closed = 0
 * 
 */
  __pyx_v_self->ops = __pyx_v_ops;

  /* "pygame_sdl2/rwobject.pyx":403
 *         """
 *         self.ops = ops
 *         self.closed = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef clear_rwops(self):
 */
  __pyx_v_self->closed = 0;

  /* "pygame_sdl2/rwobject.pyx":398
 *         return self.ops
 * 
 *     cdef set_rwops(self, SDL_RWops * ops):             # <<<<<<<<<<<<<<
 *         """
 *         Internal: Set the rwops object.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":405
 *         self.closed = 0
 * 
 *     cdef clear_rwops(self):             # <<<<<<<<<<<<<<
 *         """
 *         Internal: Clear the rwops object.
 */

static PyObject *__pyx_f_11pygame_sdl2_8rwobject_9RWopsImpl_clear_rwops(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear_rwops", 0);

  /* "pygame_sdl2/rwobject.pyx":409
 *         Internal: Clear the rwops object.
 *         """
 *         self.ops = NULL             # <<<<<<<<<<<<<<
 *         self.closed = 1
 * 
 */
  __pyx_v_self->ops = NULL;

  /* "pygame_sdl2/rwobject.pyx":410
 *         """
 *         self.ops = NULL
 *         self.closed = 1             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
  __pyx_v_self->closed = 1;

  /* "pygame_sdl2/rwobject.pyx":405
 *         self.closed = 0
 * 
 *     cdef clear_rwops(self):             # <<<<<<<<<<<<<<
 *         """
 *         Internal: Clear the rwops object.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":412
 *         self.closed = 1
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         # Allow close to be called multiple times without raising an exception.
 *         if self.closed != 0:
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_5close(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_4close[] = "RWopsImpl.close(self)";
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_5close(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_4close(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_4close(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self) {
  SDL_RWops *__pyx_v_ops;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  SDL_RWops *__pyx_t_2;
  __Pyx_RefNannySetupContext("close", 0);

  /* "pygame_sdl2/rwobject.pyx":414
 *     def close(self):
 *         # Allow close to be called multiple times without raising an exception.
 *         if self.closed != 0:             # <<<<<<<<<<<<<<
 *             return
 *         self.closed = 1
 */
  __pyx_t_1 = ((__pyx_v_self->closed != 0) != 0);
  if (__pyx_t_1) {

    /* "pygame_sdl2/rwobject.pyx":415
 *         # Allow close to be called multiple times without raising an exception.
 *         if self.closed != 0:
 *             return             # <<<<<<<<<<<<<<
 *         self.closed = 1
 *         ops = self.ops
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "pygame_sdl2/rwobject.pyx":414
 *     def close(self):
 *         # Allow close to be called multiple times without raising an exception.
 *         if self.closed != 0:             # <<<<<<<<<<<<<<
 *             return
 *         self.closed = 1
 */
  }

  /* "pygame_sdl2/rwobject.pyx":416
 *         if self.closed != 0:
 *             return
 *         self.closed = 1             # <<<<<<<<<<<<<<
 *         ops = self.ops
 *         self.ops = NULL
 */
  __pyx_v_self->closed = 1;

  /* "pygame_sdl2/rwobject.pyx":417
 *             return
 *         self.closed = 1
 *         ops = self.ops             # <<<<<<<<<<<<<<
 *         self.ops = NULL
 *         SDL_RWclose(ops)
 */
  __pyx_t_2 = __pyx_v_self->ops;
  __pyx_v_ops = __pyx_t_2;

  /* "pygame_sdl2/rwobject.pyx":418
 *         self.closed = 1
 *         ops = self.ops
 *         self.ops = NULL             # <<<<<<<<<<<<<<
 *         SDL_RWclose(ops)
 * 
 */
  __pyx_v_self->ops = NULL;

  /* "pygame_sdl2/rwobject.pyx":419
 *         ops = self.ops
 *         self.ops = NULL
 *         SDL_RWclose(ops)             # <<<<<<<<<<<<<<
 * 
 *     def seek(self, offset, whence):
 */
  (void)(SDL_RWclose(__pyx_v_ops));

  /* "pygame_sdl2/rwobject.pyx":412
 *         self.closed = 1
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         # Allow close to be called multiple times without raising an exception.
 *         if self.closed != 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":421
 *         SDL_RWclose(ops)
 * 
 *     def seek(self, offset, whence):             # <<<<<<<<<<<<<<
 *         whence_rw = RW_SEEK_SET
 *         whence_mapping = {
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_7seek(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_6seek[] = "RWopsImpl.seek(self, offset, whence)";
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_7seek(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_offset = 0;
  PyObject *__pyx_v_whence = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("seek (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_offset,&__pyx_n_s_whence,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_offset)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_whence)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("seek", 1, 2, 2, 1); __PYX_ERR(0, 421, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "seek") < 0)) __PYX_ERR(0, 421, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_offset = values[0];
    __pyx_v_whence = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("seek", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 421, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWopsImpl.seek", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_6seek(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_v_self), __pyx_v_offset, __pyx_v_whence);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_6seek(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self, PyObject *__pyx_v_offset, PyObject *__pyx_v_whence) {
  PyObject *__pyx_v_whence_rw = NULL;
  PyObject *__pyx_v_whence_mapping = NULL;
  Sint64 __pyx_v_rv;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  Sint64 __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("seek", 0);

  /* "pygame_sdl2/rwobject.pyx":422
 * 
 *     def seek(self, offset, whence):
 *         whence_rw = RW_SEEK_SET             # <<<<<<<<<<<<<<
 *         whence_mapping = {
 *             io.SEEK_SET : RW_SEEK_SET,
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(RW_SEEK_SET); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_whence_rw = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":424
 *         whence_rw = RW_SEEK_SET
 *         whence_mapping = {
 *             io.SEEK_SET : RW_SEEK_SET,             # <<<<<<<<<<<<<<
 *             io.SEEK_CUR : RW_SEEK_CUR,
 *             io.SEEK_END : RW_SEEK_END,
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_io); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SEEK_SET); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(RW_SEEK_SET); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_3, __pyx_t_2) < 0) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":425
 *         whence_mapping = {
 *             io.SEEK_SET : RW_SEEK_SET,
 *             io.SEEK_CUR : RW_SEEK_CUR,             # <<<<<<<<<<<<<<
 *             io.SEEK_END : RW_SEEK_END,
 *         }
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_io); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SEEK_CUR); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(RW_SEEK_CUR); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_3, __pyx_t_2) < 0) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":426
 *             io.SEEK_SET : RW_SEEK_SET,
 *             io.SEEK_CUR : RW_SEEK_CUR,
 *             io.SEEK_END : RW_SEEK_END,             # <<<<<<<<<<<<<<
 *         }
 *         if whence in whence_mapping:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_io); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SEEK_END); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(RW_SEEK_END); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_3, __pyx_t_2) < 0) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_whence_mapping = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":428
 *             io.SEEK_END : RW_SEEK_END,
 *         }
 *         if whence in whence_mapping:             # <<<<<<<<<<<<<<
 *             whence_rw = whence_mapping[whence]
 *         rv = SDL_RWseek(self.ops, offset, whence_rw)
 */
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_v_whence, __pyx_v_whence_mapping, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 428, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "pygame_sdl2/rwobject.pyx":429
 *         }
 *         if whence in whence_mapping:
 *             whence_rw = whence_mapping[whence]             # <<<<<<<<<<<<<<
 *         rv = SDL_RWseek(self.ops, offset, whence_rw)
 *         if rv < 0:
 */
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_whence_mapping, __pyx_v_whence); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_whence_rw, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pygame_sdl2/rwobject.pyx":428
 *             io.SEEK_END : RW_SEEK_END,
 *         }
 *         if whence in whence_mapping:             # <<<<<<<<<<<<<<
 *             whence_rw = whence_mapping[whence]
 *         rv = SDL_RWseek(self.ops, offset, whence_rw)
 */
  }

  /* "pygame_sdl2/rwobject.pyx":430
 *         if whence in whence_mapping:
 *             whence_rw = whence_mapping[whence]
 *         rv = SDL_RWseek(self.ops, offset, whence_rw)             # <<<<<<<<<<<<<<
 *         if rv < 0:
 *             raise IOError("Could not seek: {}".format(SDL_GetError()))
 */
  __pyx_t_6 = __Pyx_PyInt_As_int64_t(__pyx_v_offset); if (unlikely((__pyx_t_6 == ((Sint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 430, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_whence_rw); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 430, __pyx_L1_error)
  __pyx_v_rv = SDL_RWseek(__pyx_v_self->ops, __pyx_t_6, __pyx_t_7);

  /* "pygame_sdl2/rwobject.pyx":431
 *             whence_rw = whence_mapping[whence]
 *         rv = SDL_RWseek(self.ops, offset, whence_rw)
 *         if rv < 0:             # <<<<<<<<<<<<<<
 *             raise IOError("Could not seek: {}".format(SDL_GetError()))
 *         return rv
 */
  __pyx_t_5 = ((__pyx_v_rv < 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pygame_sdl2/rwobject.pyx":432
 *         rv = SDL_RWseek(self.ops, offset, whence_rw)
 *         if rv < 0:
 *             raise IOError("Could not seek: {}".format(SDL_GetError()))             # <<<<<<<<<<<<<<
 *         return rv
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Could_not_seek, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyBytes_FromString(SDL_GetError()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IOError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 432, __pyx_L1_error)

    /* "pygame_sdl2/rwobject.pyx":431
 *             whence_rw = whence_mapping[whence]
 *         rv = SDL_RWseek(self.ops, offset, whence_rw)
 *         if rv < 0:             # <<<<<<<<<<<<<<
 *             raise IOError("Could not seek: {}".format(SDL_GetError()))
 *         return rv
 */
  }

  /* "pygame_sdl2/rwobject.pyx":433
 *         if rv < 0:
 *             raise IOError("Could not seek: {}".format(SDL_GetError()))
 *         return rv             # <<<<<<<<<<<<<<
 * 
 *     def readinto(self, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_rv); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":421
 *         SDL_RWclose(ops)
 * 
 *     def seek(self, offset, whence):             # <<<<<<<<<<<<<<
 *         whence_rw = RW_SEEK_SET
 *         whence_mapping = {
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWopsImpl.seek", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_whence_rw);
  __Pyx_XDECREF(__pyx_v_whence_mapping);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":435
 *         return rv
 * 
 *     def readinto(self, b):             # <<<<<<<<<<<<<<
 *         cdef Py_buffer view
 *         rv = 0
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_9readinto(PyObject *__pyx_v_self, PyObject *__pyx_v_b); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_8readinto[] = "RWopsImpl.readinto(self, b)";
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_9readinto(PyObject *__pyx_v_self, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("readinto (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_8readinto(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_v_self), ((PyObject *)__pyx_v_b));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_8readinto(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self, PyObject *__pyx_v_b) {
  Py_buffer __pyx_v_view;
  size_t __pyx_v_rv;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  char const *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("readinto", 0);

  /* "pygame_sdl2/rwobject.pyx":437
 *     def readinto(self, b):
 *         cdef Py_buffer view
 *         rv = 0             # <<<<<<<<<<<<<<
 * 
 *         if not PyObject_CheckBuffer(b):
 */
  __pyx_v_rv = 0;

  /* "pygame_sdl2/rwobject.pyx":439
 *         rv = 0
 * 
 *         if not PyObject_CheckBuffer(b):             # <<<<<<<<<<<<<<
 *             raise ValueError("Passed in object does not support buffer protocol")
 *         try:
 */
  __pyx_t_1 = ((!(PyObject_CheckBuffer(__pyx_v_b) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pygame_sdl2/rwobject.pyx":440
 * 
 *         if not PyObject_CheckBuffer(b):
 *             raise ValueError("Passed in object does not support buffer protocol")             # <<<<<<<<<<<<<<
 *         try:
 *             PyObject_GetBuffer(b, &view, PyBUF_CONTIG)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 440, __pyx_L1_error)

    /* "pygame_sdl2/rwobject.pyx":439
 *         rv = 0
 * 
 *         if not PyObject_CheckBuffer(b):             # <<<<<<<<<<<<<<
 *             raise ValueError("Passed in object does not support buffer protocol")
 *         try:
 */
  }

  /* "pygame_sdl2/rwobject.pyx":441
 *         if not PyObject_CheckBuffer(b):
 *             raise ValueError("Passed in object does not support buffer protocol")
 *         try:             # <<<<<<<<<<<<<<
 *             PyObject_GetBuffer(b, &view, PyBUF_CONTIG)
 *             rv = SDL_RWread(self.ops, view.buf, 1, view.len)
 */
  /*try:*/ {

    /* "pygame_sdl2/rwobject.pyx":442
 *             raise ValueError("Passed in object does not support buffer protocol")
 *         try:
 *             PyObject_GetBuffer(b, &view, PyBUF_CONTIG)             # <<<<<<<<<<<<<<
 *             rv = SDL_RWread(self.ops, view.buf, 1, view.len)
 *         finally:
 */
    __pyx_t_3 = PyObject_GetBuffer(__pyx_v_b, (&__pyx_v_view), PyBUF_CONTIG); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 442, __pyx_L5_error)

    /* "pygame_sdl2/rwobject.pyx":443
 *         try:
 *             PyObject_GetBuffer(b, &view, PyBUF_CONTIG)
 *             rv = SDL_RWread(self.ops, view.buf, 1, view.len)             # <<<<<<<<<<<<<<
 *         finally:
 *             PyBuffer_Release(&view)
 */
    __pyx_v_rv = SDL_RWread(__pyx_v_self->ops, __pyx_v_view.buf, 1, __pyx_v_view.len);
  }

  /* "pygame_sdl2/rwobject.pyx":445
 *             rv = SDL_RWread(self.ops, view.buf, 1, view.len)
 *         finally:
 *             PyBuffer_Release(&view)             # <<<<<<<<<<<<<<
 *         if rv < 0:
 *             raise IOError("Could not read: {}".format(SDL_GetError()))
 */
  /*finally:*/ {
    /*normal exit:*/{
      PyBuffer_Release((&__pyx_v_view));
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __pyx_t_3 = __pyx_lineno; __pyx_t_4 = __pyx_clineno; __pyx_t_5 = __pyx_filename;
      {
        PyBuffer_Release((&__pyx_v_view));
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0;
      __pyx_lineno = __pyx_t_3; __pyx_clineno = __pyx_t_4; __pyx_filename = __pyx_t_5;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "pygame_sdl2/rwobject.pyx":446
 *         finally:
 *             PyBuffer_Release(&view)
 *         if rv < 0:             # <<<<<<<<<<<<<<
 *             raise IOError("Could not read: {}".format(SDL_GetError()))
 *         return rv
 */
  __pyx_t_1 = ((__pyx_v_rv < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pygame_sdl2/rwobject.pyx":447
 *             PyBuffer_Release(&view)
 *         if rv < 0:
 *             raise IOError("Could not read: {}".format(SDL_GetError()))             # <<<<<<<<<<<<<<
 *         return rv
 * 
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Could_not_read, __pyx_n_s_format); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_13 = __Pyx_PyBytes_FromString(SDL_GetError()); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
      }
    }
    __pyx_t_2 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_14, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_13);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IOError, __pyx_t_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_12, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __PYX_ERR(0, 447, __pyx_L1_error)

    /* "pygame_sdl2/rwobject.pyx":446
 *         finally:
 *             PyBuffer_Release(&view)
 *         if rv < 0:             # <<<<<<<<<<<<<<
 *             raise IOError("Could not read: {}".format(SDL_GetError()))
 *         return rv
 */
  }

  /* "pygame_sdl2/rwobject.pyx":448
 *         if rv < 0:
 *             raise IOError("Could not read: {}".format(SDL_GetError()))
 *         return rv             # <<<<<<<<<<<<<<
 * 
 *     def write(self, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_12 = __Pyx_PyInt_FromSize_t(__pyx_v_rv); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_r = __pyx_t_12;
  __pyx_t_12 = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":435
 *         return rv
 * 
 *     def readinto(self, b):             # <<<<<<<<<<<<<<
 *         cdef Py_buffer view
 *         rv = 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWopsImpl.readinto", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":450
 *         return rv
 * 
 *     def write(self, b):             # <<<<<<<<<<<<<<
 *         cdef Py_buffer view
 *         rv = 0
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_11write(PyObject *__pyx_v_self, PyObject *__pyx_v_b); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_10write[] = "RWopsImpl.write(self, b)";
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_11write(PyObject *__pyx_v_self, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_10write(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_v_self), ((PyObject *)__pyx_v_b));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_10write(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self, PyObject *__pyx_v_b) {
  Py_buffer __pyx_v_view;
  size_t __pyx_v_rv;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  char const *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write", 0);

  /* "pygame_sdl2/rwobject.pyx":452
 *     def write(self, b):
 *         cdef Py_buffer view
 *         rv = 0             # <<<<<<<<<<<<<<
 * 
 *         if not PyObject_CheckBuffer(b):
 */
  __pyx_v_rv = 0;

  /* "pygame_sdl2/rwobject.pyx":454
 *         rv = 0
 * 
 *         if not PyObject_CheckBuffer(b):             # <<<<<<<<<<<<<<
 *             raise ValueError("Passed in object does not support buffer protocol")
 *         try:
 */
  __pyx_t_1 = ((!(PyObject_CheckBuffer(__pyx_v_b) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pygame_sdl2/rwobject.pyx":455
 * 
 *         if not PyObject_CheckBuffer(b):
 *             raise ValueError("Passed in object does not support buffer protocol")             # <<<<<<<<<<<<<<
 *         try:
 *             PyObject_GetBuffer(b, &view, PyBUF_CONTIG_RO)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 455, __pyx_L1_error)

    /* "pygame_sdl2/rwobject.pyx":454
 *         rv = 0
 * 
 *         if not PyObject_CheckBuffer(b):             # <<<<<<<<<<<<<<
 *             raise ValueError("Passed in object does not support buffer protocol")
 *         try:
 */
  }

  /* "pygame_sdl2/rwobject.pyx":456
 *         if not PyObject_CheckBuffer(b):
 *             raise ValueError("Passed in object does not support buffer protocol")
 *         try:             # <<<<<<<<<<<<<<
 *             PyObject_GetBuffer(b, &view, PyBUF_CONTIG_RO)
 *             rv = SDL_RWwrite(self.ops, view.buf, 1, view.len)
 */
  /*try:*/ {

    /* "pygame_sdl2/rwobject.pyx":457
 *             raise ValueError("Passed in object does not support buffer protocol")
 *         try:
 *             PyObject_GetBuffer(b, &view, PyBUF_CONTIG_RO)             # <<<<<<<<<<<<<<
 *             rv = SDL_RWwrite(self.ops, view.buf, 1, view.len)
 *         finally:
 */
    __pyx_t_3 = PyObject_GetBuffer(__pyx_v_b, (&__pyx_v_view), PyBUF_CONTIG_RO); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 457, __pyx_L5_error)

    /* "pygame_sdl2/rwobject.pyx":458
 *         try:
 *             PyObject_GetBuffer(b, &view, PyBUF_CONTIG_RO)
 *             rv = SDL_RWwrite(self.ops, view.buf, 1, view.len)             # <<<<<<<<<<<<<<
 *         finally:
 *             PyBuffer_Release(&view)
 */
    __pyx_v_rv = SDL_RWwrite(__pyx_v_self->ops, __pyx_v_view.buf, 1, __pyx_v_view.len);
  }

  /* "pygame_sdl2/rwobject.pyx":460
 *             rv = SDL_RWwrite(self.ops, view.buf, 1, view.len)
 *         finally:
 *             PyBuffer_Release(&view)             # <<<<<<<<<<<<<<
 *         if rv < 0:
 *             raise IOError("Could not write: {}".format(SDL_GetError()))
 */
  /*finally:*/ {
    /*normal exit:*/{
      PyBuffer_Release((&__pyx_v_view));
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __pyx_t_3 = __pyx_lineno; __pyx_t_4 = __pyx_clineno; __pyx_t_5 = __pyx_filename;
      {
        PyBuffer_Release((&__pyx_v_view));
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0;
      __pyx_lineno = __pyx_t_3; __pyx_clineno = __pyx_t_4; __pyx_filename = __pyx_t_5;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "pygame_sdl2/rwobject.pyx":461
 *         finally:
 *             PyBuffer_Release(&view)
 *         if rv < 0:             # <<<<<<<<<<<<<<
 *             raise IOError("Could not write: {}".format(SDL_GetError()))
 *         return rv
 */
  __pyx_t_1 = ((__pyx_v_rv < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pygame_sdl2/rwobject.pyx":462
 *             PyBuffer_Release(&view)
 *         if rv < 0:
 *             raise IOError("Could not write: {}".format(SDL_GetError()))             # <<<<<<<<<<<<<<
 *         return rv
 * 
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Could_not_write, __pyx_n_s_format); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 462, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_13 = __Pyx_PyBytes_FromString(SDL_GetError()); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 462, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
      }
    }
    __pyx_t_2 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_14, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_13);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 462, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IOError, __pyx_t_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 462, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_12, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __PYX_ERR(0, 462, __pyx_L1_error)

    /* "pygame_sdl2/rwobject.pyx":461
 *         finally:
 *             PyBuffer_Release(&view)
 *         if rv < 0:             # <<<<<<<<<<<<<<
 *             raise IOError("Could not write: {}".format(SDL_GetError()))
 *         return rv
 */
  }

  /* "pygame_sdl2/rwobject.pyx":463
 *         if rv < 0:
 *             raise IOError("Could not write: {}".format(SDL_GetError()))
 *         return rv             # <<<<<<<<<<<<<<
 * 
 *     def get_sdl_rwops_pointer(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_12 = __Pyx_PyInt_FromSize_t(__pyx_v_rv); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_r = __pyx_t_12;
  __pyx_t_12 = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":450
 *         return rv
 * 
 *     def write(self, b):             # <<<<<<<<<<<<<<
 *         cdef Py_buffer view
 *         rv = 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWopsImpl.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":465
 *         return rv
 * 
 *     def get_sdl_rwops_pointer(self):             # <<<<<<<<<<<<<<
 *         import ctypes
 *         return ctypes.c_void_p(<uintptr_t> self.ops)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_13get_sdl_rwops_pointer(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_12get_sdl_rwops_pointer[] = "RWopsImpl.get_sdl_rwops_pointer(self)";
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_13get_sdl_rwops_pointer(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_sdl_rwops_pointer (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_12get_sdl_rwops_pointer(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_12get_sdl_rwops_pointer(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self) {
  PyObject *__pyx_v_ctypes = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_sdl_rwops_pointer", 0);

  /* "pygame_sdl2/rwobject.pyx":466
 * 
 *     def get_sdl_rwops_pointer(self):
 *         import ctypes             # <<<<<<<<<<<<<<
 *         return ctypes.c_void_p(<uintptr_t> self.ops)
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_ctypes, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ctypes = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":467
 *     def get_sdl_rwops_pointer(self):
 *         import ctypes
 *         return ctypes.c_void_p(<uintptr_t> self.ops)             # <<<<<<<<<<<<<<
 * 
 * class RWops(io.RawIOBase):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_ctypes, __pyx_n_s_c_void_p); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(((uintptr_t)__pyx_v_self->ops)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":465
 *         return rv
 * 
 *     def get_sdl_rwops_pointer(self):             # <<<<<<<<<<<<<<
 *         import ctypes
 *         return ctypes.c_void_p(<uintptr_t> self.ops)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWopsImpl.get_sdl_rwops_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ctypes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.ops cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_14__reduce_cython__[] = "RWopsImpl.__reduce_cython__(self)";
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_14__reduce_cython__(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.ops cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.ops cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.ops cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWopsImpl.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.ops cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.ops cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_16__setstate_cython__[] = "RWopsImpl.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_16__setstate_cython__(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_9RWopsImpl_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.ops cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.ops cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.ops cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.ops cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWopsImpl.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":470
 * 
 * class RWops(io.RawIOBase):
 *     def __init__(self, name=None):             # <<<<<<<<<<<<<<
 *         io.RawIOBase.__init__(self)
 *         self._holder = RWopsImpl()
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_5RWops___init__[] = "RWops.__init__(self, name=None)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_5RWops_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11pygame_sdl2_8rwobject_5RWops_1__init__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11pygame_sdl2_8rwobject_5RWops___init__};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 470, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 470, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_5RWops___init__(__pyx_self, __pyx_v_self, __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pygame_sdl2/rwobject.pyx":471
 * class RWops(io.RawIOBase):
 *     def __init__(self, name=None):
 *         io.RawIOBase.__init__(self)             # <<<<<<<<<<<<<<
 *         self._holder = RWopsImpl()
 *         self.name = name
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_io); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_RawIOBase); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":472
 *     def __init__(self, name=None):
 *         io.RawIOBase.__init__(self)
 *         self._holder = RWopsImpl()             # <<<<<<<<<<<<<<
 *         self.name = name
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_11pygame_sdl2_8rwobject_RWopsImpl)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_holder, __pyx_t_1) < 0) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":473
 *         io.RawIOBase.__init__(self)
 *         self._holder = RWopsImpl()
 *         self.name = name             # <<<<<<<<<<<<<<
 * 
 *     # Implemented class: io.IOBase
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_v_name) < 0) __PYX_ERR(0, 473, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":470
 * 
 * class RWops(io.RawIOBase):
 *     def __init__(self, name=None):             # <<<<<<<<<<<<<<
 *         io.RawIOBase.__init__(self)
 *         self._holder = RWopsImpl()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":477
 *     # Implemented class: io.IOBase
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         self._holder.close()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_3close(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_5RWops_2close[] = "RWops.close(self)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_5RWops_3close = {"close", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_5RWops_3close, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_5RWops_2close};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_3close(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_5RWops_2close(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_2close(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);

  /* "pygame_sdl2/rwobject.pyx":478
 * 
 *     def close(self):
 *         self._holder.close()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_holder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_close); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":477
 *     # Implemented class: io.IOBase
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         self._holder.close()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":481
 * 
 *     @property
 *     def closed(self):             # <<<<<<<<<<<<<<
 *         return self._holder.get_closed()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_5closed(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_5RWops_4closed[] = "RWops.closed(self)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_5RWops_5closed = {"closed", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_5RWops_5closed, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_5RWops_4closed};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_5closed(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("closed (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_5RWops_4closed(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_4closed(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("closed", 0);

  /* "pygame_sdl2/rwobject.pyx":482
 *     @property
 *     def closed(self):
 *         return self._holder.get_closed()             # <<<<<<<<<<<<<<
 * 
 *     def fileno(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_holder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get_closed); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":481
 * 
 *     @property
 *     def closed(self):             # <<<<<<<<<<<<<<
 *         return self._holder.get_closed()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops.closed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":484
 *         return self._holder.get_closed()
 * 
 *     def fileno(self):             # <<<<<<<<<<<<<<
 *         raise OSError()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_7fileno(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_5RWops_6fileno[] = "RWops.fileno(self)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_5RWops_7fileno = {"fileno", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_5RWops_7fileno, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_5RWops_6fileno};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_7fileno(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fileno (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_5RWops_6fileno(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_6fileno(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fileno", 0);

  /* "pygame_sdl2/rwobject.pyx":485
 * 
 *     def fileno(self):
 *         raise OSError()             # <<<<<<<<<<<<<<
 * 
 *     # inherited flush is used
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_OSError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 485, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":484
 *         return self._holder.get_closed()
 * 
 *     def fileno(self):             # <<<<<<<<<<<<<<
 *         raise OSError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops.fileno", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":491
 *     # inherited isatty is used
 * 
 *     def readable(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_9readable(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_5RWops_8readable[] = "RWops.readable(self)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_5RWops_9readable = {"readable", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_5RWops_9readable, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_5RWops_8readable};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_9readable(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("readable (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_5RWops_8readable(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_8readable(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("readable", 0);

  /* "pygame_sdl2/rwobject.pyx":492
 * 
 *     def readable(self):
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     # inherited readline is used
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":491
 *     # inherited isatty is used
 * 
 *     def readable(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":498
 *     # inherited readlines is used
 * 
 *     def seek(self, offset, whence=0):             # <<<<<<<<<<<<<<
 *         self._holder.check_closed()
 *         return self._holder.seek(offset, whence)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_11seek(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_5RWops_10seek[] = "RWops.seek(self, offset, whence=0)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_5RWops_11seek = {"seek", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11pygame_sdl2_8rwobject_5RWops_11seek, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11pygame_sdl2_8rwobject_5RWops_10seek};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_11seek(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_offset = 0;
  PyObject *__pyx_v_whence = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("seek (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_offset,&__pyx_n_s_whence,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)__pyx_int_0));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("seek", 0, 2, 3, 1); __PYX_ERR(0, 498, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_whence);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "seek") < 0)) __PYX_ERR(0, 498, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_offset = values[1];
    __pyx_v_whence = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("seek", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 498, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops.seek", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_5RWops_10seek(__pyx_self, __pyx_v_self, __pyx_v_offset, __pyx_v_whence);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_10seek(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_offset, PyObject *__pyx_v_whence) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("seek", 0);

  /* "pygame_sdl2/rwobject.pyx":499
 * 
 *     def seek(self, offset, whence=0):
 *         self._holder.check_closed()             # <<<<<<<<<<<<<<
 *         return self._holder.seek(offset, whence)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_holder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_check_closed); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":500
 *     def seek(self, offset, whence=0):
 *         self._holder.check_closed()
 *         return self._holder.seek(offset, whence)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_holder); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_seek); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_offset, __pyx_v_whence};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_offset, __pyx_v_whence};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_offset);
    __Pyx_GIVEREF(__pyx_v_offset);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_offset);
    __Pyx_INCREF(__pyx_v_whence);
    __Pyx_GIVEREF(__pyx_v_whence);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_whence);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":498
 *     # inherited readlines is used
 * 
 *     def seek(self, offset, whence=0):             # <<<<<<<<<<<<<<
 *         self._holder.check_closed()
 *         return self._holder.seek(offset, whence)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops.seek", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":503
 * 
 * 
 *     def seekable(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_13seekable(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_5RWops_12seekable[] = "RWops.seekable(self)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_5RWops_13seekable = {"seekable", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_5RWops_13seekable, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_5RWops_12seekable};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_13seekable(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("seekable (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_5RWops_12seekable(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_12seekable(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("seekable", 0);

  /* "pygame_sdl2/rwobject.pyx":504
 * 
 *     def seekable(self):
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     # inherited tell is used
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":503
 * 
 * 
 *     def seekable(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":508
 *     # inherited tell is used
 * 
 *     def truncate(self, size=None):             # <<<<<<<<<<<<<<
 *         raise OSError()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_15truncate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_5RWops_14truncate[] = "RWops.truncate(self, size=None)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_5RWops_15truncate = {"truncate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11pygame_sdl2_8rwobject_5RWops_15truncate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11pygame_sdl2_8rwobject_5RWops_14truncate};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_15truncate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_size = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("truncate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_size,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "truncate") < 0)) __PYX_ERR(0, 508, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_size = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("truncate", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 508, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops.truncate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_5RWops_14truncate(__pyx_self, __pyx_v_self, __pyx_v_size);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_14truncate(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("truncate", 0);

  /* "pygame_sdl2/rwobject.pyx":509
 * 
 *     def truncate(self, size=None):
 *         raise OSError()             # <<<<<<<<<<<<<<
 * 
 *     def writable(self):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_OSError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 509, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":508
 *     # inherited tell is used
 * 
 *     def truncate(self, size=None):             # <<<<<<<<<<<<<<
 *         raise OSError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops.truncate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":511
 *         raise OSError()
 * 
 *     def writable(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_17writable(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_5RWops_16writable[] = "RWops.writable(self)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_5RWops_17writable = {"writable", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_5RWops_17writable, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_5RWops_16writable};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_17writable(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writable (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_5RWops_16writable(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_16writable(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writable", 0);

  /* "pygame_sdl2/rwobject.pyx":512
 * 
 *     def writable(self):
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     # inherited writelines is used
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":511
 *         raise OSError()
 * 
 *     def writable(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":524
 *     # inherited readall is used
 * 
 *     def readinto(self, b):             # <<<<<<<<<<<<<<
 *         self._holder.check_closed()
 *         return self._holder.readinto(b)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_19readinto(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_5RWops_18readinto[] = "RWops.readinto(self, b)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_5RWops_19readinto = {"readinto", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11pygame_sdl2_8rwobject_5RWops_19readinto, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11pygame_sdl2_8rwobject_5RWops_18readinto};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_19readinto(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_b = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("readinto (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_b,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("readinto", 1, 2, 2, 1); __PYX_ERR(0, 524, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "readinto") < 0)) __PYX_ERR(0, 524, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_b = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("readinto", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 524, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops.readinto", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_5RWops_18readinto(__pyx_self, __pyx_v_self, __pyx_v_b);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_18readinto(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("readinto", 0);

  /* "pygame_sdl2/rwobject.pyx":525
 * 
 *     def readinto(self, b):
 *         self._holder.check_closed()             # <<<<<<<<<<<<<<
 *         return self._holder.readinto(b)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_holder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_check_closed); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":526
 *     def readinto(self, b):
 *         self._holder.check_closed()
 *         return self._holder.readinto(b)             # <<<<<<<<<<<<<<
 * 
 *     def write(self, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_holder); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_readinto); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_b) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_b);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":524
 *     # inherited readall is used
 * 
 *     def readinto(self, b):             # <<<<<<<<<<<<<<
 *         self._holder.check_closed()
 *         return self._holder.readinto(b)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops.readinto", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":528
 *         return self._holder.readinto(b)
 * 
 *     def write(self, b):             # <<<<<<<<<<<<<<
 *         self._holder.check_closed()
 *         return self._holder.write(b)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_21write(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_5RWops_20write[] = "RWops.write(self, b)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_5RWops_21write = {"write", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11pygame_sdl2_8rwobject_5RWops_21write, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11pygame_sdl2_8rwobject_5RWops_20write};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_21write(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_b = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_b,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("write", 1, 2, 2, 1); __PYX_ERR(0, 528, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "write") < 0)) __PYX_ERR(0, 528, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_b = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("write", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 528, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_5RWops_20write(__pyx_self, __pyx_v_self, __pyx_v_b);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_20write(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write", 0);

  /* "pygame_sdl2/rwobject.pyx":529
 * 
 *     def write(self, b):
 *         self._holder.check_closed()             # <<<<<<<<<<<<<<
 *         return self._holder.write(b)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_holder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_check_closed); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":530
 *     def write(self, b):
 *         self._holder.check_closed()
 *         return self._holder.write(b)             # <<<<<<<<<<<<<<
 * 
 *     def get_sdl_rwops_pointer(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_holder); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_b) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_b);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":528
 *         return self._holder.readinto(b)
 * 
 *     def write(self, b):             # <<<<<<<<<<<<<<
 *         self._holder.check_closed()
 *         return self._holder.write(b)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":532
 *         return self._holder.write(b)
 * 
 *     def get_sdl_rwops_pointer(self):             # <<<<<<<<<<<<<<
 *         return self._holder.get_sdl_rwops_pointer()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_23get_sdl_rwops_pointer(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_5RWops_22get_sdl_rwops_pointer[] = "RWops.get_sdl_rwops_pointer(self)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_5RWops_23get_sdl_rwops_pointer = {"get_sdl_rwops_pointer", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_5RWops_23get_sdl_rwops_pointer, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_5RWops_22get_sdl_rwops_pointer};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_23get_sdl_rwops_pointer(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_sdl_rwops_pointer (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_5RWops_22get_sdl_rwops_pointer(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_5RWops_22get_sdl_rwops_pointer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_sdl_rwops_pointer", 0);

  /* "pygame_sdl2/rwobject.pyx":533
 * 
 *     def get_sdl_rwops_pointer(self):
 *         return self._holder.get_sdl_rwops_pointer()             # <<<<<<<<<<<<<<
 * 
 * def RWops_from_file(name, mode="rb"):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_holder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get_sdl_rwops_pointer); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":532
 *         return self._holder.write(b)
 * 
 *     def get_sdl_rwops_pointer(self):             # <<<<<<<<<<<<<<
 *         return self._holder.get_sdl_rwops_pointer()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops.get_sdl_rwops_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":535
 *         return self._holder.get_sdl_rwops_pointer()
 * 
 * def RWops_from_file(name, mode="rb"):             # <<<<<<<<<<<<<<
 *     cdef SDL_RWops *rw
 *     cdef char *cname
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_1RWops_from_file(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_RWops_from_file[] = "RWops_from_file(name, mode='rb')";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_1RWops_from_file = {"RWops_from_file", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11pygame_sdl2_8rwobject_1RWops_from_file, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11pygame_sdl2_8rwobject_RWops_from_file};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_1RWops_from_file(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("RWops_from_file (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_mode,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)__pyx_n_s_rb);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "RWops_from_file") < 0)) __PYX_ERR(0, 535, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = values[0];
    __pyx_v_mode = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("RWops_from_file", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 535, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops_from_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_RWops_from_file(__pyx_self, __pyx_v_name, __pyx_v_mode);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_RWops_from_file(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_mode) {
  SDL_RWops *__pyx_v_rw;
  char *__pyx_v_cname;
  char *__pyx_v_cmode;
  PyObject *__pyx_v_dname = NULL;
  PyObject *__pyx_v_rv = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("RWops_from_file", 0);
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_INCREF(__pyx_v_mode);

  /* "pygame_sdl2/rwobject.pyx":540
 *     cdef char *cmode
 * 
 *     if not isinstance(mode, bytes_):             # <<<<<<<<<<<<<<
 *         mode = mode.encode("ascii")
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_bytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_mode, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_3) {

    /* "pygame_sdl2/rwobject.pyx":541
 * 
 *     if not isinstance(mode, bytes_):
 *         mode = mode.encode("ascii")             # <<<<<<<<<<<<<<
 * 
 *     # Try to open as a file.
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_mode, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_n_s_ascii) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_s_ascii);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_mode, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pygame_sdl2/rwobject.pyx":540
 *     cdef char *cmode
 * 
 *     if not isinstance(mode, bytes_):             # <<<<<<<<<<<<<<
 *         mode = mode.encode("ascii")
 * 
 */
  }

  /* "pygame_sdl2/rwobject.pyx":544
 * 
 *     # Try to open as a file.
 *     if isinstance(name, bytes_):             # <<<<<<<<<<<<<<
 *         name = name.decode(fsencoding)
 *     elif isinstance(name, unicode_):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_bytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_name, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "pygame_sdl2/rwobject.pyx":545
 *     # Try to open as a file.
 *     if isinstance(name, bytes_):
 *         name = name.decode(fsencoding)             # <<<<<<<<<<<<<<
 *     elif isinstance(name, unicode_):
 *         pass
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_decode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_fsencoding); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_name, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pygame_sdl2/rwobject.pyx":544
 * 
 *     # Try to open as a file.
 *     if isinstance(name, bytes_):             # <<<<<<<<<<<<<<
 *         name = name.decode(fsencoding)
 *     elif isinstance(name, unicode_):
 */
    goto __pyx_L4;
  }

  /* "pygame_sdl2/rwobject.pyx":546
 *     if isinstance(name, bytes_):
 *         name = name.decode(fsencoding)
 *     elif isinstance(name, unicode_):             # <<<<<<<<<<<<<<
 *         pass
 *     else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_name, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
    goto __pyx_L4;
  }

  /* "pygame_sdl2/rwobject.pyx":549
 *         pass
 *     else:
 *         name = None             # <<<<<<<<<<<<<<
 * 
 *     if name is not None:
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_name, Py_None);
  }
  __pyx_L4:;

  /* "pygame_sdl2/rwobject.pyx":551
 *         name = None
 * 
 *     if name is not None:             # <<<<<<<<<<<<<<
 * 
 *         dname = name.encode("utf-8")
 */
  __pyx_t_3 = (__pyx_v_name != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "pygame_sdl2/rwobject.pyx":553
 *     if name is not None:
 * 
 *         dname = name.encode("utf-8")             # <<<<<<<<<<<<<<
 *         cname = dname
 *         cmode = mode
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_s_utf_8);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_dname = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "pygame_sdl2/rwobject.pyx":554
 * 
 *         dname = name.encode("utf-8")
 *         cname = dname             # <<<<<<<<<<<<<<
 *         cmode = mode
 * 
 */
    __pyx_t_7 = __Pyx_PyObject_AsWritableString(__pyx_v_dname); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(0, 554, __pyx_L1_error)
    __pyx_v_cname = __pyx_t_7;

    /* "pygame_sdl2/rwobject.pyx":555
 *         dname = name.encode("utf-8")
 *         cname = dname
 *         cmode = mode             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
 */
    __pyx_t_7 = __Pyx_PyObject_AsWritableString(__pyx_v_mode); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(0, 555, __pyx_L1_error)
    __pyx_v_cmode = __pyx_t_7;

    /* "pygame_sdl2/rwobject.pyx":557
 *         cmode = mode
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             rw = SDL_RWFromFile(cname, cmode)
 * 
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {

          /* "pygame_sdl2/rwobject.pyx":558
 * 
 *         with nogil:
 *             rw = SDL_RWFromFile(cname, cmode)             # <<<<<<<<<<<<<<
 * 
 *         if rw == NULL:
 */
          __pyx_v_rw = SDL_RWFromFile(__pyx_v_cname, __pyx_v_cmode);
        }

        /* "pygame_sdl2/rwobject.pyx":557
 *         cmode = mode
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             rw = SDL_RWFromFile(cname, cmode)
 * 
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L8;
          }
          __pyx_L8:;
        }
    }

    /* "pygame_sdl2/rwobject.pyx":560
 *             rw = SDL_RWFromFile(cname, cmode)
 * 
 *         if rw == NULL:             # <<<<<<<<<<<<<<
 *             raise IOError("Could not open {!r}: {}".format(name, SDL_GetError()))
 * 
 */
    __pyx_t_2 = ((__pyx_v_rw == NULL) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "pygame_sdl2/rwobject.pyx":561
 * 
 *         if rw == NULL:
 *             raise IOError("Could not open {!r}: {}".format(name, SDL_GetError()))             # <<<<<<<<<<<<<<
 * 
 *         rv = RWops(name)
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Could_not_open_r, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 561, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyBytes_FromString(SDL_GetError()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 561, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_name, __pyx_t_5};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 561, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_name, __pyx_t_5};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 561, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 561, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_INCREF(__pyx_v_name);
        __Pyx_GIVEREF(__pyx_v_name);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_name);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 561, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IOError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 561, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 561, __pyx_L1_error)

      /* "pygame_sdl2/rwobject.pyx":560
 *             rw = SDL_RWFromFile(cname, cmode)
 * 
 *         if rw == NULL:             # <<<<<<<<<<<<<<
 *             raise IOError("Could not open {!r}: {}".format(name, SDL_GetError()))
 * 
 */
    }

    /* "pygame_sdl2/rwobject.pyx":563
 *             raise IOError("Could not open {!r}: {}".format(name, SDL_GetError()))
 * 
 *         rv = RWops(name)             # <<<<<<<<<<<<<<
 *         (<RWopsImpl>rv._holder).set_rwops(rw)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_RWops); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_9, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_name);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_rv = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "pygame_sdl2/rwobject.pyx":564
 * 
 *         rv = RWops(name)
 *         (<RWopsImpl>rv._holder).set_rwops(rw)             # <<<<<<<<<<<<<<
 * 
 *         return rv
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_rv, __pyx_n_s_holder); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = ((struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *)((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_4)->__pyx_vtab)->set_rwops(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_4), __pyx_v_rw); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pygame_sdl2/rwobject.pyx":566
 *         (<RWopsImpl>rv._holder).set_rwops(rw)
 * 
 *         return rv             # <<<<<<<<<<<<<<
 * 
 *     return ValueError("Invalid value passed in for name")
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_rv);
    __pyx_r = __pyx_v_rv;
    goto __pyx_L0;

    /* "pygame_sdl2/rwobject.pyx":551
 *         name = None
 * 
 *     if name is not None:             # <<<<<<<<<<<<<<
 * 
 *         dname = name.encode("utf-8")
 */
  }

  /* "pygame_sdl2/rwobject.pyx":568
 *         return rv
 * 
 *     return ValueError("Invalid value passed in for name")             # <<<<<<<<<<<<<<
 * 
 * def RWops_from_file_like_object(filelike):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":535
 *         return self._holder.get_sdl_rwops_pointer()
 * 
 * def RWops_from_file(name, mode="rb"):             # <<<<<<<<<<<<<<
 *     cdef SDL_RWops *rw
 *     cdef char *cname
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops_from_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dname);
  __Pyx_XDECREF(__pyx_v_rv);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_mode);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":570
 *     return ValueError("Invalid value passed in for name")
 * 
 * def RWops_from_file_like_object(filelike):             # <<<<<<<<<<<<<<
 *     cdef SDL_RWops *rw
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_3RWops_from_file_like_object(PyObject *__pyx_self, PyObject *__pyx_v_filelike); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_2RWops_from_file_like_object[] = "RWops_from_file_like_object(filelike)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_3RWops_from_file_like_object = {"RWops_from_file_like_object", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_3RWops_from_file_like_object, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_2RWops_from_file_like_object};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_3RWops_from_file_like_object(PyObject *__pyx_self, PyObject *__pyx_v_filelike) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("RWops_from_file_like_object (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_2RWops_from_file_like_object(__pyx_self, ((PyObject *)__pyx_v_filelike));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_2RWops_from_file_like_object(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_filelike) {
  SDL_RWops *__pyx_v_rw;
  PyObject *__pyx_v_rv = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  SDL_RWops *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("RWops_from_file_like_object", 0);

  /* "pygame_sdl2/rwobject.pyx":573
 *     cdef SDL_RWops *rw
 * 
 *     rw = to_rwops(filelike)             # <<<<<<<<<<<<<<
 * 
 *     rv = RWops()
 */
  __pyx_t_1 = __pyx_f_11pygame_sdl2_8rwobject_to_rwops(__pyx_v_filelike, NULL); if (unlikely(__pyx_t_1 == ((SDL_RWops *)NULL))) __PYX_ERR(0, 573, __pyx_L1_error)
  __pyx_v_rw = __pyx_t_1;

  /* "pygame_sdl2/rwobject.pyx":575
 *     rw = to_rwops(filelike)
 * 
 *     rv = RWops()             # <<<<<<<<<<<<<<
 *     (<RWopsImpl>rv._holder).set_rwops(rw)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_RWops); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_rv = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":576
 * 
 *     rv = RWops()
 *     (<RWopsImpl>rv._holder).set_rwops(rw)             # <<<<<<<<<<<<<<
 * 
 *     return rv
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rv, __pyx_n_s_holder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *)((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_2)->__pyx_vtab)->set_rwops(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_2), __pyx_v_rw); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pygame_sdl2/rwobject.pyx":578
 *     (<RWopsImpl>rv._holder).set_rwops(rw)
 * 
 *     return rv             # <<<<<<<<<<<<<<
 * 
 * def RWops_is_file_like_object(rw_in_object):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_rv);
  __pyx_r = __pyx_v_rv;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":570
 *     return ValueError("Invalid value passed in for name")
 * 
 * def RWops_from_file_like_object(filelike):             # <<<<<<<<<<<<<<
 *     cdef SDL_RWops *rw
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops_from_file_like_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rv);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":580
 *     return rv
 * 
 * def RWops_is_file_like_object(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_is_file_like_object(PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_4RWops_is_file_like_object[] = "RWops_is_file_like_object(rw_in_object)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_5RWops_is_file_like_object = {"RWops_is_file_like_object", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_5RWops_is_file_like_object, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_4RWops_is_file_like_object};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_5RWops_is_file_like_object(PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("RWops_is_file_like_object (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_4RWops_is_file_like_object(__pyx_self, ((PyObject *)__pyx_v_rw_in_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_4RWops_is_file_like_object(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object) {
  SDL_RWops *__pyx_v_rw_in;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("RWops_is_file_like_object", 0);

  /* "pygame_sdl2/rwobject.pyx":585
 *     cdef SDL_RWops *rw
 * 
 *     if not isinstance(rw_in_object, RWops):             # <<<<<<<<<<<<<<
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_RWops); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_rw_in_object, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_3) {

    /* "pygame_sdl2/rwobject.pyx":586
 * 
 *     if not isinstance(rw_in_object, RWops):
 *         return False             # <<<<<<<<<<<<<<
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "pygame_sdl2/rwobject.pyx":585
 *     cdef SDL_RWops *rw
 * 
 *     if not isinstance(rw_in_object, RWops):             # <<<<<<<<<<<<<<
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 */
  }

  /* "pygame_sdl2/rwobject.pyx":587
 *     if not isinstance(rw_in_object, RWops):
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()             # <<<<<<<<<<<<<<
 *     if rw_in == NULL:
 *         return False
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_rw_in_object, __pyx_n_s_holder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_rw_in = ((struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *)((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1)->__pyx_vtab)->get_rwops(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":588
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:             # <<<<<<<<<<<<<<
 *         return False
 *     return rw_in.read == python_read
 */
  __pyx_t_3 = ((__pyx_v_rw_in == NULL) != 0);
  if (__pyx_t_3) {

    /* "pygame_sdl2/rwobject.pyx":589
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:
 *         return False             # <<<<<<<<<<<<<<
 *     return rw_in.read == python_read
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "pygame_sdl2/rwobject.pyx":588
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:             # <<<<<<<<<<<<<<
 *         return False
 *     return rw_in.read == python_read
 */
  }

  /* "pygame_sdl2/rwobject.pyx":590
 *     if rw_in == NULL:
 *         return False
 *     return rw_in.read == python_read             # <<<<<<<<<<<<<<
 * 
 * def RWops_from_buffer(b, mode="rb"):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_rw_in->read == __pyx_f_11pygame_sdl2_8rwobject_python_read)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":580
 *     return rv
 * 
 * def RWops_is_file_like_object(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops_is_file_like_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":592
 *     return rw_in.read == python_read
 * 
 * def RWops_from_buffer(b, mode="rb"):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_7RWops_from_buffer(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_6RWops_from_buffer[] = "RWops_from_buffer(b, mode='rb')";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_7RWops_from_buffer = {"RWops_from_buffer", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11pygame_sdl2_8rwobject_7RWops_from_buffer, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11pygame_sdl2_8rwobject_6RWops_from_buffer};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_7RWops_from_buffer(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_b = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("RWops_from_buffer (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_b,&__pyx_n_s_mode,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)__pyx_n_s_rb);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "RWops_from_buffer") < 0)) __PYX_ERR(0, 592, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_b = values[0];
    __pyx_v_mode = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("RWops_from_buffer", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 592, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops_from_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_6RWops_from_buffer(__pyx_self, __pyx_v_b, __pyx_v_mode);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_6RWops_from_buffer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_b, PyObject *__pyx_v_mode) {
  Py_buffer __pyx_v_view;
  SDL_RWops *__pyx_v_rw;
  struct __pyx_t_11pygame_sdl2_8rwobject_BufFile *__pyx_v_bf;
  PyObject *__pyx_v_rv = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  Uint8 *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("RWops_from_buffer", 0);

  /* "pygame_sdl2/rwobject.pyx":596
 *     cdef SDL_RWops *rw
 * 
 *     if not PyObject_CheckBuffer(b):             # <<<<<<<<<<<<<<
 *         raise ValueError("Passed in object does not support buffer protocol")
 *     PyObject_GetBuffer(b, &view, PyBUF_CONTIG_RO if ("r" in mode) else PyBUF_CONTIG)
 */
  __pyx_t_1 = ((!(PyObject_CheckBuffer(__pyx_v_b) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pygame_sdl2/rwobject.pyx":597
 * 
 *     if not PyObject_CheckBuffer(b):
 *         raise ValueError("Passed in object does not support buffer protocol")             # <<<<<<<<<<<<<<
 *     PyObject_GetBuffer(b, &view, PyBUF_CONTIG_RO if ("r" in mode) else PyBUF_CONTIG)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 597, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 597, __pyx_L1_error)

    /* "pygame_sdl2/rwobject.pyx":596
 *     cdef SDL_RWops *rw
 * 
 *     if not PyObject_CheckBuffer(b):             # <<<<<<<<<<<<<<
 *         raise ValueError("Passed in object does not support buffer protocol")
 *     PyObject_GetBuffer(b, &view, PyBUF_CONTIG_RO if ("r" in mode) else PyBUF_CONTIG)
 */
  }

  /* "pygame_sdl2/rwobject.pyx":598
 *     if not PyObject_CheckBuffer(b):
 *         raise ValueError("Passed in object does not support buffer protocol")
 *     PyObject_GetBuffer(b, &view, PyBUF_CONTIG_RO if ("r" in mode) else PyBUF_CONTIG)             # <<<<<<<<<<<<<<
 * 
 *     bf = <BufFile *> calloc(sizeof(BufFile), 1)
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_n_s_r, __pyx_v_mode, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 598, __pyx_L1_error)
  if ((__pyx_t_1 != 0)) {
    __pyx_t_3 = PyBUF_CONTIG_RO;
  } else {
    __pyx_t_3 = PyBUF_CONTIG;
  }
  __pyx_t_4 = PyObject_GetBuffer(__pyx_v_b, (&__pyx_v_view), __pyx_t_3); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 598, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":600
 *     PyObject_GetBuffer(b, &view, PyBUF_CONTIG_RO if ("r" in mode) else PyBUF_CONTIG)
 * 
 *     bf = <BufFile *> calloc(sizeof(BufFile), 1)             # <<<<<<<<<<<<<<
 *     bf.view = view
 *     bf.base = <Uint8 *>view.buf
 */
  __pyx_v_bf = ((struct __pyx_t_11pygame_sdl2_8rwobject_BufFile *)calloc((sizeof(struct __pyx_t_11pygame_sdl2_8rwobject_BufFile)), 1));

  /* "pygame_sdl2/rwobject.pyx":601
 * 
 *     bf = <BufFile *> calloc(sizeof(BufFile), 1)
 *     bf.view = view             # <<<<<<<<<<<<<<
 *     bf.base = <Uint8 *>view.buf
 *     bf.here = bf.base
 */
  __pyx_v_bf->view = __pyx_v_view;

  /* "pygame_sdl2/rwobject.pyx":602
 *     bf = <BufFile *> calloc(sizeof(BufFile), 1)
 *     bf.view = view
 *     bf.base = <Uint8 *>view.buf             # <<<<<<<<<<<<<<
 *     bf.here = bf.base
 *     bf.stop = bf.base + view.len
 */
  __pyx_v_bf->base = ((Uint8 *)__pyx_v_view.buf);

  /* "pygame_sdl2/rwobject.pyx":603
 *     bf.view = view
 *     bf.base = <Uint8 *>view.buf
 *     bf.here = bf.base             # <<<<<<<<<<<<<<
 *     bf.stop = bf.base + view.len
 * 
 */
  __pyx_t_5 = __pyx_v_bf->base;
  __pyx_v_bf->here = __pyx_t_5;

  /* "pygame_sdl2/rwobject.pyx":604
 *     bf.base = <Uint8 *>view.buf
 *     bf.here = bf.base
 *     bf.stop = bf.base + view.len             # <<<<<<<<<<<<<<
 * 
 *     rw = SDL_AllocRW()
 */
  __pyx_v_bf->stop = (__pyx_v_bf->base + __pyx_v_view.len);

  /* "pygame_sdl2/rwobject.pyx":606
 *     bf.stop = bf.base + view.len
 * 
 *     rw = SDL_AllocRW()             # <<<<<<<<<<<<<<
 *     rw.size = buffile_size
 *     rw.seek = buffile_seek
 */
  __pyx_v_rw = SDL_AllocRW();

  /* "pygame_sdl2/rwobject.pyx":607
 * 
 *     rw = SDL_AllocRW()
 *     rw.size = buffile_size             # <<<<<<<<<<<<<<
 *     rw.seek = buffile_seek
 *     rw.read = buffile_read
 */
  __pyx_v_rw->size = __pyx_f_11pygame_sdl2_8rwobject_buffile_size;

  /* "pygame_sdl2/rwobject.pyx":608
 *     rw = SDL_AllocRW()
 *     rw.size = buffile_size
 *     rw.seek = buffile_seek             # <<<<<<<<<<<<<<
 *     rw.read = buffile_read
 *     rw.write = buffile_write
 */
  __pyx_v_rw->seek = __pyx_f_11pygame_sdl2_8rwobject_buffile_seek;

  /* "pygame_sdl2/rwobject.pyx":609
 *     rw.size = buffile_size
 *     rw.seek = buffile_seek
 *     rw.read = buffile_read             # <<<<<<<<<<<<<<
 *     rw.write = buffile_write
 *     rw.close = buffile_close
 */
  __pyx_v_rw->read = __pyx_f_11pygame_sdl2_8rwobject_buffile_read;

  /* "pygame_sdl2/rwobject.pyx":610
 *     rw.seek = buffile_seek
 *     rw.read = buffile_read
 *     rw.write = buffile_write             # <<<<<<<<<<<<<<
 *     rw.close = buffile_close
 *     rw.type = 0
 */
  __pyx_v_rw->write = __pyx_f_11pygame_sdl2_8rwobject_buffile_write;

  /* "pygame_sdl2/rwobject.pyx":611
 *     rw.read = buffile_read
 *     rw.write = buffile_write
 *     rw.close = buffile_close             # <<<<<<<<<<<<<<
 *     rw.type = 0
 *     rw.hidden.unknown.data1 = <void *> bf
 */
  __pyx_v_rw->close = __pyx_f_11pygame_sdl2_8rwobject_buffile_close;

  /* "pygame_sdl2/rwobject.pyx":612
 *     rw.write = buffile_write
 *     rw.close = buffile_close
 *     rw.type = 0             # <<<<<<<<<<<<<<
 *     rw.hidden.unknown.data1 = <void *> bf
 * 
 */
  __pyx_v_rw->type = 0;

  /* "pygame_sdl2/rwobject.pyx":613
 *     rw.close = buffile_close
 *     rw.type = 0
 *     rw.hidden.unknown.data1 = <void *> bf             # <<<<<<<<<<<<<<
 * 
 *     rv = RWops()
 */
  __pyx_v_rw->hidden.unknown.data1 = ((void *)__pyx_v_bf);

  /* "pygame_sdl2/rwobject.pyx":615
 *     rw.hidden.unknown.data1 = <void *> bf
 * 
 *     rv = RWops()             # <<<<<<<<<<<<<<
 *     (<RWopsImpl>rv._holder).set_rwops(rw)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_RWops); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_rv = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":616
 * 
 *     rv = RWops()
 *     (<RWopsImpl>rv._holder).set_rwops(rw)             # <<<<<<<<<<<<<<
 * 
 *     return rv
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rv, __pyx_n_s_holder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = ((struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *)((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_2)->__pyx_vtab)->set_rwops(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_2), __pyx_v_rw); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pygame_sdl2/rwobject.pyx":618
 *     (<RWopsImpl>rv._holder).set_rwops(rw)
 * 
 *     return rv             # <<<<<<<<<<<<<<
 * 
 * def RWops_is_buffer(rw_in_object):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_rv);
  __pyx_r = __pyx_v_rv;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":592
 *     return rw_in.read == python_read
 * 
 * def RWops_from_buffer(b, mode="rb"):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops_from_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rv);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":620
 *     return rv
 * 
 * def RWops_is_buffer(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWops_is_buffer(PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_8RWops_is_buffer[] = "RWops_is_buffer(rw_in_object)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_9RWops_is_buffer = {"RWops_is_buffer", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_9RWops_is_buffer, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_8RWops_is_buffer};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_9RWops_is_buffer(PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("RWops_is_buffer (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_8RWops_is_buffer(__pyx_self, ((PyObject *)__pyx_v_rw_in_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_8RWops_is_buffer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object) {
  SDL_RWops *__pyx_v_rw_in;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("RWops_is_buffer", 0);

  /* "pygame_sdl2/rwobject.pyx":625
 *     cdef SDL_RWops *rw
 * 
 *     if not isinstance(rw_in_object, RWops):             # <<<<<<<<<<<<<<
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_RWops); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_rw_in_object, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_3) {

    /* "pygame_sdl2/rwobject.pyx":626
 * 
 *     if not isinstance(rw_in_object, RWops):
 *         return False             # <<<<<<<<<<<<<<
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "pygame_sdl2/rwobject.pyx":625
 *     cdef SDL_RWops *rw
 * 
 *     if not isinstance(rw_in_object, RWops):             # <<<<<<<<<<<<<<
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 */
  }

  /* "pygame_sdl2/rwobject.pyx":627
 *     if not isinstance(rw_in_object, RWops):
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()             # <<<<<<<<<<<<<<
 *     if rw_in == NULL:
 *         return False
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_rw_in_object, __pyx_n_s_holder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_rw_in = ((struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *)((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1)->__pyx_vtab)->get_rwops(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":628
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:             # <<<<<<<<<<<<<<
 *         return False
 *     return rw_in.read == buffile_read
 */
  __pyx_t_3 = ((__pyx_v_rw_in == NULL) != 0);
  if (__pyx_t_3) {

    /* "pygame_sdl2/rwobject.pyx":629
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:
 *         return False             # <<<<<<<<<<<<<<
 *     return rw_in.read == buffile_read
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "pygame_sdl2/rwobject.pyx":628
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:             # <<<<<<<<<<<<<<
 *         return False
 *     return rw_in.read == buffile_read
 */
  }

  /* "pygame_sdl2/rwobject.pyx":630
 *     if rw_in == NULL:
 *         return False
 *     return rw_in.read == buffile_read             # <<<<<<<<<<<<<<
 * 
 * def RWops_create_subfile(rw_in_object, base, length):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_rw_in->read == __pyx_f_11pygame_sdl2_8rwobject_buffile_read)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":620
 *     return rv
 * 
 * def RWops_is_buffer(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops_is_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":632
 *     return rw_in.read == buffile_read
 * 
 * def RWops_create_subfile(rw_in_object, base, length):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_11RWops_create_subfile(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_10RWops_create_subfile[] = "RWops_create_subfile(rw_in_object, base, length)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_11RWops_create_subfile = {"RWops_create_subfile", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11pygame_sdl2_8rwobject_11RWops_create_subfile, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11pygame_sdl2_8rwobject_10RWops_create_subfile};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_11RWops_create_subfile(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_rw_in_object = 0;
  PyObject *__pyx_v_base = 0;
  PyObject *__pyx_v_length = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("RWops_create_subfile (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_rw_in_object,&__pyx_n_s_base,&__pyx_n_s_length,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rw_in_object)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_base)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RWops_create_subfile", 1, 3, 3, 1); __PYX_ERR(0, 632, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_length)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RWops_create_subfile", 1, 3, 3, 2); __PYX_ERR(0, 632, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "RWops_create_subfile") < 0)) __PYX_ERR(0, 632, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_rw_in_object = values[0];
    __pyx_v_base = values[1];
    __pyx_v_length = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("RWops_create_subfile", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 632, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops_create_subfile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_10RWops_create_subfile(__pyx_self, __pyx_v_rw_in_object, __pyx_v_base, __pyx_v_length);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_10RWops_create_subfile(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object, PyObject *__pyx_v_base, PyObject *__pyx_v_length) {
  SDL_RWops *__pyx_v_rw_in;
  SDL_RWops *__pyx_v_rw;
  struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *__pyx_v_sf;
  PyObject *__pyx_v_rv = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Sint64 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("RWops_create_subfile", 0);

  /* "pygame_sdl2/rwobject.pyx":637
 *     cdef SDL_RWops *rw
 * 
 *     if not isinstance(rw_in_object, RWops):             # <<<<<<<<<<<<<<
 *         raise ValueError("Passed in object should be RWops")
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_RWops); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_rw_in_object, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "pygame_sdl2/rwobject.pyx":638
 * 
 *     if not isinstance(rw_in_object, RWops):
 *         raise ValueError("Passed in object should be RWops")             # <<<<<<<<<<<<<<
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 638, __pyx_L1_error)

    /* "pygame_sdl2/rwobject.pyx":637
 *     cdef SDL_RWops *rw
 * 
 *     if not isinstance(rw_in_object, RWops):             # <<<<<<<<<<<<<<
 *         raise ValueError("Passed in object should be RWops")
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 */
  }

  /* "pygame_sdl2/rwobject.pyx":639
 *     if not isinstance(rw_in_object, RWops):
 *         raise ValueError("Passed in object should be RWops")
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()             # <<<<<<<<<<<<<<
 *     if rw_in == NULL:
 *         raise ValueError("Passed in RWops object is closed")
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_rw_in_object, __pyx_n_s_holder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_rw_in = ((struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *)((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1)->__pyx_vtab)->get_rwops(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":640
 *         raise ValueError("Passed in object should be RWops")
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:             # <<<<<<<<<<<<<<
 *         raise ValueError("Passed in RWops object is closed")
 *     (<RWopsImpl>rw_in_object._holder).clear_rwops()
 */
  __pyx_t_3 = ((__pyx_v_rw_in == NULL) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "pygame_sdl2/rwobject.pyx":641
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:
 *         raise ValueError("Passed in RWops object is closed")             # <<<<<<<<<<<<<<
 *     (<RWopsImpl>rw_in_object._holder).clear_rwops()
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 641, __pyx_L1_error)

    /* "pygame_sdl2/rwobject.pyx":640
 *         raise ValueError("Passed in object should be RWops")
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:             # <<<<<<<<<<<<<<
 *         raise ValueError("Passed in RWops object is closed")
 *     (<RWopsImpl>rw_in_object._holder).clear_rwops()
 */
  }

  /* "pygame_sdl2/rwobject.pyx":642
 *     if rw_in == NULL:
 *         raise ValueError("Passed in RWops object is closed")
 *     (<RWopsImpl>rw_in_object._holder).clear_rwops()             # <<<<<<<<<<<<<<
 * 
 *     SDL_RWseek(rw_in, base, RW_SEEK_SET)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_rw_in_object, __pyx_n_s_holder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = ((struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *)((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1)->__pyx_vtab)->clear_rwops(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pygame_sdl2/rwobject.pyx":644
 *     (<RWopsImpl>rw_in_object._holder).clear_rwops()
 * 
 *     SDL_RWseek(rw_in, base, RW_SEEK_SET)             # <<<<<<<<<<<<<<
 * 
 *     sf = <SubFile *> calloc(sizeof(SubFile), 1)
 */
  __pyx_t_5 = __Pyx_PyInt_As_int64_t(__pyx_v_base); if (unlikely((__pyx_t_5 == ((Sint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 644, __pyx_L1_error)
  (void)(SDL_RWseek(__pyx_v_rw_in, __pyx_t_5, RW_SEEK_SET));

  /* "pygame_sdl2/rwobject.pyx":646
 *     SDL_RWseek(rw_in, base, RW_SEEK_SET)
 * 
 *     sf = <SubFile *> calloc(sizeof(SubFile), 1)             # <<<<<<<<<<<<<<
 *     sf.rw = rw_in
 *     sf.base = base
 */
  __pyx_v_sf = ((struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *)calloc((sizeof(struct __pyx_t_11pygame_sdl2_8rwobject_SubFile)), 1));

  /* "pygame_sdl2/rwobject.pyx":647
 * 
 *     sf = <SubFile *> calloc(sizeof(SubFile), 1)
 *     sf.rw = rw_in             # <<<<<<<<<<<<<<
 *     sf.base = base
 *     sf.length = length
 */
  __pyx_v_sf->rw = __pyx_v_rw_in;

  /* "pygame_sdl2/rwobject.pyx":648
 *     sf = <SubFile *> calloc(sizeof(SubFile), 1)
 *     sf.rw = rw_in
 *     sf.base = base             # <<<<<<<<<<<<<<
 *     sf.length = length
 *     sf.tell = 0
 */
  __pyx_t_5 = __Pyx_PyInt_As_int64_t(__pyx_v_base); if (unlikely((__pyx_t_5 == ((Sint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 648, __pyx_L1_error)
  __pyx_v_sf->base = __pyx_t_5;

  /* "pygame_sdl2/rwobject.pyx":649
 *     sf.rw = rw_in
 *     sf.base = base
 *     sf.length = length             # <<<<<<<<<<<<<<
 *     sf.tell = 0
 * 
 */
  __pyx_t_5 = __Pyx_PyInt_As_int64_t(__pyx_v_length); if (unlikely((__pyx_t_5 == ((Sint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 649, __pyx_L1_error)
  __pyx_v_sf->length = __pyx_t_5;

  /* "pygame_sdl2/rwobject.pyx":650
 *     sf.base = base
 *     sf.length = length
 *     sf.tell = 0             # <<<<<<<<<<<<<<
 * 
 *     rw = SDL_AllocRW()
 */
  __pyx_v_sf->tell = 0;

  /* "pygame_sdl2/rwobject.pyx":652
 *     sf.tell = 0
 * 
 *     rw = SDL_AllocRW()             # <<<<<<<<<<<<<<
 *     rw.size = subfile_size
 *     rw.seek = subfile_seek
 */
  __pyx_v_rw = SDL_AllocRW();

  /* "pygame_sdl2/rwobject.pyx":653
 * 
 *     rw = SDL_AllocRW()
 *     rw.size = subfile_size             # <<<<<<<<<<<<<<
 *     rw.seek = subfile_seek
 *     rw.read = subfile_read
 */
  __pyx_v_rw->size = __pyx_f_11pygame_sdl2_8rwobject_subfile_size;

  /* "pygame_sdl2/rwobject.pyx":654
 *     rw = SDL_AllocRW()
 *     rw.size = subfile_size
 *     rw.seek = subfile_seek             # <<<<<<<<<<<<<<
 *     rw.read = subfile_read
 *     rw.write = NULL
 */
  __pyx_v_rw->seek = __pyx_f_11pygame_sdl2_8rwobject_subfile_seek;

  /* "pygame_sdl2/rwobject.pyx":655
 *     rw.size = subfile_size
 *     rw.seek = subfile_seek
 *     rw.read = subfile_read             # <<<<<<<<<<<<<<
 *     rw.write = NULL
 *     rw.close = subfile_close
 */
  __pyx_v_rw->read = __pyx_f_11pygame_sdl2_8rwobject_subfile_read;

  /* "pygame_sdl2/rwobject.pyx":656
 *     rw.seek = subfile_seek
 *     rw.read = subfile_read
 *     rw.write = NULL             # <<<<<<<<<<<<<<
 *     rw.close = subfile_close
 *     rw.type = 0
 */
  __pyx_v_rw->write = NULL;

  /* "pygame_sdl2/rwobject.pyx":657
 *     rw.read = subfile_read
 *     rw.write = NULL
 *     rw.close = subfile_close             # <<<<<<<<<<<<<<
 *     rw.type = 0
 *     rw.hidden.unknown.data1 = <void *> sf
 */
  __pyx_v_rw->close = __pyx_f_11pygame_sdl2_8rwobject_subfile_close;

  /* "pygame_sdl2/rwobject.pyx":658
 *     rw.write = NULL
 *     rw.close = subfile_close
 *     rw.type = 0             # <<<<<<<<<<<<<<
 *     rw.hidden.unknown.data1 = <void *> sf
 * 
 */
  __pyx_v_rw->type = 0;

  /* "pygame_sdl2/rwobject.pyx":659
 *     rw.close = subfile_close
 *     rw.type = 0
 *     rw.hidden.unknown.data1 = <void *> sf             # <<<<<<<<<<<<<<
 * 
 *     rv = RWops()
 */
  __pyx_v_rw->hidden.unknown.data1 = ((void *)__pyx_v_sf);

  /* "pygame_sdl2/rwobject.pyx":661
 *     rw.hidden.unknown.data1 = <void *> sf
 * 
 *     rv = RWops()             # <<<<<<<<<<<<<<
 *     (<RWopsImpl>rv._holder).set_rwops(rw)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_RWops); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rv = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pygame_sdl2/rwobject.pyx":662
 * 
 *     rv = RWops()
 *     (<RWopsImpl>rv._holder).set_rwops(rw)             # <<<<<<<<<<<<<<
 * 
 *     return rv
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_rv, __pyx_n_s_holder); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = ((struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *)((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_4)->__pyx_vtab)->set_rwops(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_4), __pyx_v_rw); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":664
 *     (<RWopsImpl>rv._holder).set_rwops(rw)
 * 
 *     return rv             # <<<<<<<<<<<<<<
 * 
 * def RWops_is_subfile(rw_in_object):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_rv);
  __pyx_r = __pyx_v_rv;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":632
 *     return rw_in.read == buffile_read
 * 
 * def RWops_create_subfile(rw_in_object, base, length):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops_create_subfile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rv);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":666
 *     return rv
 * 
 * def RWops_is_subfile(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_13RWops_is_subfile(PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_12RWops_is_subfile[] = "RWops_is_subfile(rw_in_object)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_13RWops_is_subfile = {"RWops_is_subfile", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_13RWops_is_subfile, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_12RWops_is_subfile};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_13RWops_is_subfile(PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("RWops_is_subfile (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_12RWops_is_subfile(__pyx_self, ((PyObject *)__pyx_v_rw_in_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_12RWops_is_subfile(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object) {
  SDL_RWops *__pyx_v_rw_in;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("RWops_is_subfile", 0);

  /* "pygame_sdl2/rwobject.pyx":671
 *     cdef SDL_RWops *rw
 * 
 *     if not isinstance(rw_in_object, RWops):             # <<<<<<<<<<<<<<
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_RWops); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_rw_in_object, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_3) {

    /* "pygame_sdl2/rwobject.pyx":672
 * 
 *     if not isinstance(rw_in_object, RWops):
 *         return False             # <<<<<<<<<<<<<<
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "pygame_sdl2/rwobject.pyx":671
 *     cdef SDL_RWops *rw
 * 
 *     if not isinstance(rw_in_object, RWops):             # <<<<<<<<<<<<<<
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 */
  }

  /* "pygame_sdl2/rwobject.pyx":673
 *     if not isinstance(rw_in_object, RWops):
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()             # <<<<<<<<<<<<<<
 *     if rw_in == NULL:
 *         return False
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_rw_in_object, __pyx_n_s_holder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_rw_in = ((struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *)((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1)->__pyx_vtab)->get_rwops(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":674
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:             # <<<<<<<<<<<<<<
 *         return False
 *     return rw_in.read == subfile_read
 */
  __pyx_t_3 = ((__pyx_v_rw_in == NULL) != 0);
  if (__pyx_t_3) {

    /* "pygame_sdl2/rwobject.pyx":675
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:
 *         return False             # <<<<<<<<<<<<<<
 *     return rw_in.read == subfile_read
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "pygame_sdl2/rwobject.pyx":674
 *         return False
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:             # <<<<<<<<<<<<<<
 *         return False
 *     return rw_in.read == subfile_read
 */
  }

  /* "pygame_sdl2/rwobject.pyx":676
 *     if rw_in == NULL:
 *         return False
 *     return rw_in.read == subfile_read             # <<<<<<<<<<<<<<
 * 
 * def RWops_unwrap_subfile(rw_in_object):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_rw_in->read == __pyx_f_11pygame_sdl2_8rwobject_subfile_read)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":666
 *     return rv
 * 
 * def RWops_is_subfile(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops_is_subfile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pygame_sdl2/rwobject.pyx":678
 *     return rw_in.read == subfile_read
 * 
 * def RWops_unwrap_subfile(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */

/* Python wrapper */
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_15RWops_unwrap_subfile(PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object); /*proto*/
static char __pyx_doc_11pygame_sdl2_8rwobject_14RWops_unwrap_subfile[] = "RWops_unwrap_subfile(rw_in_object)";
static PyMethodDef __pyx_mdef_11pygame_sdl2_8rwobject_15RWops_unwrap_subfile = {"RWops_unwrap_subfile", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_15RWops_unwrap_subfile, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_14RWops_unwrap_subfile};
static PyObject *__pyx_pw_11pygame_sdl2_8rwobject_15RWops_unwrap_subfile(PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("RWops_unwrap_subfile (wrapper)", 0);
  __pyx_r = __pyx_pf_11pygame_sdl2_8rwobject_14RWops_unwrap_subfile(__pyx_self, ((PyObject *)__pyx_v_rw_in_object));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11pygame_sdl2_8rwobject_14RWops_unwrap_subfile(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_rw_in_object) {
  SDL_RWops *__pyx_v_rw_in;
  SDL_RWops *__pyx_v_rw;
  struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *__pyx_v_sf;
  PyObject *__pyx_v_rv = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  SDL_RWops *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("RWops_unwrap_subfile", 0);

  /* "pygame_sdl2/rwobject.pyx":684
 *     cdef SubFile *sf
 * 
 *     if not isinstance(rw_in_object, RWops):             # <<<<<<<<<<<<<<
 *         raise ValueError("Passed in object should be RWops")
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_RWops); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_rw_in_object, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "pygame_sdl2/rwobject.pyx":685
 * 
 *     if not isinstance(rw_in_object, RWops):
 *         raise ValueError("Passed in object should be RWops")             # <<<<<<<<<<<<<<
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 685, __pyx_L1_error)

    /* "pygame_sdl2/rwobject.pyx":684
 *     cdef SubFile *sf
 * 
 *     if not isinstance(rw_in_object, RWops):             # <<<<<<<<<<<<<<
 *         raise ValueError("Passed in object should be RWops")
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 */
  }

  /* "pygame_sdl2/rwobject.pyx":686
 *     if not isinstance(rw_in_object, RWops):
 *         raise ValueError("Passed in object should be RWops")
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()             # <<<<<<<<<<<<<<
 *     if rw_in == NULL:
 *         raise ValueError("Passed in RWops object is closed")
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_rw_in_object, __pyx_n_s_holder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_rw_in = ((struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *)((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1)->__pyx_vtab)->get_rwops(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":687
 *         raise ValueError("Passed in object should be RWops")
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:             # <<<<<<<<<<<<<<
 *         raise ValueError("Passed in RWops object is closed")
 *     if rw_in.read != subfile_read:
 */
  __pyx_t_3 = ((__pyx_v_rw_in == NULL) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "pygame_sdl2/rwobject.pyx":688
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:
 *         raise ValueError("Passed in RWops object is closed")             # <<<<<<<<<<<<<<
 *     if rw_in.read != subfile_read:
 *         raise ValueError("Passed in RWops object is not SubFile")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 688, __pyx_L1_error)

    /* "pygame_sdl2/rwobject.pyx":687
 *         raise ValueError("Passed in object should be RWops")
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:             # <<<<<<<<<<<<<<
 *         raise ValueError("Passed in RWops object is closed")
 *     if rw_in.read != subfile_read:
 */
  }

  /* "pygame_sdl2/rwobject.pyx":689
 *     if rw_in == NULL:
 *         raise ValueError("Passed in RWops object is closed")
 *     if rw_in.read != subfile_read:             # <<<<<<<<<<<<<<
 *         raise ValueError("Passed in RWops object is not SubFile")
 *     (<RWopsImpl>rw_in_object._holder).clear_rwops()
 */
  __pyx_t_3 = ((__pyx_v_rw_in->read != __pyx_f_11pygame_sdl2_8rwobject_subfile_read) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "pygame_sdl2/rwobject.pyx":690
 *         raise ValueError("Passed in RWops object is closed")
 *     if rw_in.read != subfile_read:
 *         raise ValueError("Passed in RWops object is not SubFile")             # <<<<<<<<<<<<<<
 *     (<RWopsImpl>rw_in_object._holder).clear_rwops()
 *     sf = <SubFile *>rw_in_object.hidden.unknown.data1
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 690, __pyx_L1_error)

    /* "pygame_sdl2/rwobject.pyx":689
 *     if rw_in == NULL:
 *         raise ValueError("Passed in RWops object is closed")
 *     if rw_in.read != subfile_read:             # <<<<<<<<<<<<<<
 *         raise ValueError("Passed in RWops object is not SubFile")
 *     (<RWopsImpl>rw_in_object._holder).clear_rwops()
 */
  }

  /* "pygame_sdl2/rwobject.pyx":691
 *     if rw_in.read != subfile_read:
 *         raise ValueError("Passed in RWops object is not SubFile")
 *     (<RWopsImpl>rw_in_object._holder).clear_rwops()             # <<<<<<<<<<<<<<
 *     sf = <SubFile *>rw_in_object.hidden.unknown.data1
 *     rw = sf.rw
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_rw_in_object, __pyx_n_s_holder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = ((struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *)((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1)->__pyx_vtab)->clear_rwops(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pygame_sdl2/rwobject.pyx":692
 *         raise ValueError("Passed in RWops object is not SubFile")
 *     (<RWopsImpl>rw_in_object._holder).clear_rwops()
 *     sf = <SubFile *>rw_in_object.hidden.unknown.data1             # <<<<<<<<<<<<<<
 *     rw = sf.rw
 *     sf.rw = NULL
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_rw_in_object, __pyx_n_s_hidden); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_unknown); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_data1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sf = ((struct __pyx_t_11pygame_sdl2_8rwobject_SubFile *)__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pygame_sdl2/rwobject.pyx":693
 *     (<RWopsImpl>rw_in_object._holder).clear_rwops()
 *     sf = <SubFile *>rw_in_object.hidden.unknown.data1
 *     rw = sf.rw             # <<<<<<<<<<<<<<
 *     sf.rw = NULL
 *     SDL_RWclose(rw_in)
 */
  __pyx_t_5 = __pyx_v_sf->rw;
  __pyx_v_rw = __pyx_t_5;

  /* "pygame_sdl2/rwobject.pyx":694
 *     sf = <SubFile *>rw_in_object.hidden.unknown.data1
 *     rw = sf.rw
 *     sf.rw = NULL             # <<<<<<<<<<<<<<
 *     SDL_RWclose(rw_in)
 * 
 */
  __pyx_v_sf->rw = NULL;

  /* "pygame_sdl2/rwobject.pyx":695
 *     rw = sf.rw
 *     sf.rw = NULL
 *     SDL_RWclose(rw_in)             # <<<<<<<<<<<<<<
 * 
 *     SDL_RWseek(rw, 0, RW_SEEK_SET)
 */
  (void)(SDL_RWclose(__pyx_v_rw_in));

  /* "pygame_sdl2/rwobject.pyx":697
 *     SDL_RWclose(rw_in)
 * 
 *     SDL_RWseek(rw, 0, RW_SEEK_SET)             # <<<<<<<<<<<<<<
 * 
 *     rv = RWops()
 */
  (void)(SDL_RWseek(__pyx_v_rw, 0, RW_SEEK_SET));

  /* "pygame_sdl2/rwobject.pyx":699
 *     SDL_RWseek(rw, 0, RW_SEEK_SET)
 * 
 *     rv = RWops()             # <<<<<<<<<<<<<<
 *     (<RWopsImpl>rv._holder).set_rwops(rw)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_RWops); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rv = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pygame_sdl2/rwobject.pyx":700
 * 
 *     rv = RWops()
 *     (<RWopsImpl>rv._holder).set_rwops(rw)             # <<<<<<<<<<<<<<
 * 
 *     return rv
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_rv, __pyx_n_s_holder); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = ((struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl *)((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_4)->__pyx_vtab)->set_rwops(((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)__pyx_t_4), __pyx_v_rw); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pygame_sdl2/rwobject.pyx":702
 *     (<RWopsImpl>rv._holder).set_rwops(rw)
 * 
 *     return rv             # <<<<<<<<<<<<<<
 * 
 * init_python_threads()
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_rv);
  __pyx_r = __pyx_v_rv;
  goto __pyx_L0;

  /* "pygame_sdl2/rwobject.pyx":678
 *     return rw_in.read == subfile_read
 * 
 * def RWops_unwrap_subfile(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pygame_sdl2.rwobject.RWops_unwrap_subfile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rv);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_11pygame_sdl2_8rwobject_RWopsImpl __pyx_vtable_11pygame_sdl2_8rwobject_RWopsImpl;

static PyObject *__pyx_tp_new_11pygame_sdl2_8rwobject_RWopsImpl(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *)o);
  p->__pyx_vtab = __pyx_vtabptr_11pygame_sdl2_8rwobject_RWopsImpl;
  return o;
}

static void __pyx_tp_dealloc_11pygame_sdl2_8rwobject_RWopsImpl(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_11pygame_sdl2_8rwobject_RWopsImpl[] = {
  {"check_closed", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_1check_closed, METH_NOARGS, __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_check_closed},
  {"get_closed", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_3get_closed, METH_NOARGS, __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_2get_closed},
  {"close", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_5close, METH_NOARGS, __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_4close},
  {"seek", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_7seek, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_6seek},
  {"readinto", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_9readinto, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_8readinto},
  {"write", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_11write, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_10write},
  {"get_sdl_rwops_pointer", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_13get_sdl_rwops_pointer, METH_NOARGS, __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_12get_sdl_rwops_pointer},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_15__reduce_cython__, METH_NOARGS, __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_14__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_11pygame_sdl2_8rwobject_9RWopsImpl_17__setstate_cython__, METH_O, __pyx_doc_11pygame_sdl2_8rwobject_9RWopsImpl_16__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_11pygame_sdl2_8rwobject_RWopsImpl = {
  PyVarObject_HEAD_INIT(0, 0)
  "pygame_sdl2.rwobject.RWopsImpl", /*tp_name*/
  sizeof(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_11pygame_sdl2_8rwobject_RWopsImpl, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_11pygame_sdl2_8rwobject_RWopsImpl, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_11pygame_sdl2_8rwobject_RWopsImpl, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_rwobject(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_rwobject},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "rwobject",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Could_not_open_r, __pyx_k_Could_not_open_r, sizeof(__pyx_k_Could_not_open_r), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_read, __pyx_k_Could_not_read, sizeof(__pyx_k_Could_not_read), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_seek, __pyx_k_Could_not_seek, sizeof(__pyx_k_Could_not_seek), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_write, __pyx_k_Could_not_write, sizeof(__pyx_k_Could_not_write), 0, 0, 1, 0},
  {&__pyx_n_s_IOBase, __pyx_k_IOBase, sizeof(__pyx_k_IOBase), 0, 0, 1, 1},
  {&__pyx_n_s_IOError, __pyx_k_IOError, sizeof(__pyx_k_IOError), 0, 0, 1, 1},
  {&__pyx_kp_s_I_O_operation_on_closed_RWops, __pyx_k_I_O_operation_on_closed_RWops, sizeof(__pyx_k_I_O_operation_on_closed_RWops), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_value_passed_in_for_name, __pyx_k_Invalid_value_passed_in_for_name, sizeof(__pyx_k_Invalid_value_passed_in_for_name), 0, 0, 1, 0},
  {&__pyx_n_s_OSError, __pyx_k_OSError, sizeof(__pyx_k_OSError), 0, 0, 1, 1},
  {&__pyx_kp_s_Passed_in_RWops_object_is_closed, __pyx_k_Passed_in_RWops_object_is_closed, sizeof(__pyx_k_Passed_in_RWops_object_is_closed), 0, 0, 1, 0},
  {&__pyx_kp_s_Passed_in_RWops_object_is_not_Su, __pyx_k_Passed_in_RWops_object_is_not_Su, sizeof(__pyx_k_Passed_in_RWops_object_is_not_Su), 0, 0, 1, 0},
  {&__pyx_kp_s_Passed_in_object_does_not_suppor, __pyx_k_Passed_in_object_does_not_suppor, sizeof(__pyx_k_Passed_in_object_does_not_suppor), 0, 0, 1, 0},
  {&__pyx_kp_s_Passed_in_object_should_be_RWops, __pyx_k_Passed_in_object_should_be_RWops, sizeof(__pyx_k_Passed_in_object_should_be_RWops), 0, 0, 1, 0},
  {&__pyx_n_s_RWops, __pyx_k_RWops, sizeof(__pyx_k_RWops), 0, 0, 1, 1},
  {&__pyx_n_s_RWopsImpl, __pyx_k_RWopsImpl, sizeof(__pyx_k_RWopsImpl), 0, 0, 1, 1},
  {&__pyx_n_s_RWops___init, __pyx_k_RWops___init, sizeof(__pyx_k_RWops___init), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_close, __pyx_k_RWops_close, sizeof(__pyx_k_RWops_close), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_closed, __pyx_k_RWops_closed, sizeof(__pyx_k_RWops_closed), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_create_subfile, __pyx_k_RWops_create_subfile, sizeof(__pyx_k_RWops_create_subfile), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_fileno, __pyx_k_RWops_fileno, sizeof(__pyx_k_RWops_fileno), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_from_buffer, __pyx_k_RWops_from_buffer, sizeof(__pyx_k_RWops_from_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_from_file, __pyx_k_RWops_from_file, sizeof(__pyx_k_RWops_from_file), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_from_file_like_object, __pyx_k_RWops_from_file_like_object, sizeof(__pyx_k_RWops_from_file_like_object), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_get_sdl_rwops_pointer, __pyx_k_RWops_get_sdl_rwops_pointer, sizeof(__pyx_k_RWops_get_sdl_rwops_pointer), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_is_buffer, __pyx_k_RWops_is_buffer, sizeof(__pyx_k_RWops_is_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_is_file_like_object, __pyx_k_RWops_is_file_like_object, sizeof(__pyx_k_RWops_is_file_like_object), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_is_subfile, __pyx_k_RWops_is_subfile, sizeof(__pyx_k_RWops_is_subfile), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_readable, __pyx_k_RWops_readable, sizeof(__pyx_k_RWops_readable), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_readinto, __pyx_k_RWops_readinto, sizeof(__pyx_k_RWops_readinto), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_seek, __pyx_k_RWops_seek, sizeof(__pyx_k_RWops_seek), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_seekable, __pyx_k_RWops_seekable, sizeof(__pyx_k_RWops_seekable), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_truncate, __pyx_k_RWops_truncate, sizeof(__pyx_k_RWops_truncate), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_unwrap_subfile, __pyx_k_RWops_unwrap_subfile, sizeof(__pyx_k_RWops_unwrap_subfile), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_writable, __pyx_k_RWops_writable, sizeof(__pyx_k_RWops_writable), 0, 0, 1, 1},
  {&__pyx_n_s_RWops_write, __pyx_k_RWops_write, sizeof(__pyx_k_RWops_write), 0, 0, 1, 1},
  {&__pyx_n_s_RawIOBase, __pyx_k_RawIOBase, sizeof(__pyx_k_RawIOBase), 0, 0, 1, 1},
  {&__pyx_n_s_SEEK_CUR, __pyx_k_SEEK_CUR, sizeof(__pyx_k_SEEK_CUR), 0, 0, 1, 1},
  {&__pyx_n_s_SEEK_END, __pyx_k_SEEK_END, sizeof(__pyx_k_SEEK_END), 0, 0, 1, 1},
  {&__pyx_n_s_SEEK_SET, __pyx_k_SEEK_SET, sizeof(__pyx_k_SEEK_SET), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unknown_value_for_whence, __pyx_k_Unknown_value_for_whence, sizeof(__pyx_k_Unknown_value_for_whence), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_ascii, __pyx_k_ascii, sizeof(__pyx_k_ascii), 0, 0, 1, 1},
  {&__pyx_n_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_bf, __pyx_k_bf, sizeof(__pyx_k_bf), 0, 0, 1, 1},
  {&__pyx_n_s_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_c_void_p, __pyx_k_c_void_p, sizeof(__pyx_k_c_void_p), 0, 0, 1, 1},
  {&__pyx_n_s_check_closed, __pyx_k_check_closed, sizeof(__pyx_k_check_closed), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_closed, __pyx_k_closed, sizeof(__pyx_k_closed), 0, 0, 1, 1},
  {&__pyx_n_s_cmode, __pyx_k_cmode, sizeof(__pyx_k_cmode), 0, 0, 1, 1},
  {&__pyx_n_s_cname, __pyx_k_cname, sizeof(__pyx_k_cname), 0, 0, 1, 1},
  {&__pyx_n_s_ctypes, __pyx_k_ctypes, sizeof(__pyx_k_ctypes), 0, 0, 1, 1},
  {&__pyx_n_s_data1, __pyx_k_data1, sizeof(__pyx_k_data1), 0, 0, 1, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_n_s_dname, __pyx_k_dname, sizeof(__pyx_k_dname), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_file_type, __pyx_k_file_type, sizeof(__pyx_k_file_type), 0, 0, 1, 1},
  {&__pyx_n_s_filelike, __pyx_k_filelike, sizeof(__pyx_k_filelike), 0, 0, 1, 1},
  {&__pyx_n_s_fileno, __pyx_k_fileno, sizeof(__pyx_k_fileno), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fsencoding, __pyx_k_fsencoding, sizeof(__pyx_k_fsencoding), 0, 0, 1, 1},
  {&__pyx_n_s_get_closed, __pyx_k_get_closed, sizeof(__pyx_k_get_closed), 0, 0, 1, 1},
  {&__pyx_n_s_get_sdl_rwops_pointer, __pyx_k_get_sdl_rwops_pointer, sizeof(__pyx_k_get_sdl_rwops_pointer), 0, 0, 1, 1},
  {&__pyx_n_s_getfilesystemencoding, __pyx_k_getfilesystemencoding, sizeof(__pyx_k_getfilesystemencoding), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_hidden, __pyx_k_hidden, sizeof(__pyx_k_hidden), 0, 0, 1, 1},
  {&__pyx_n_s_holder, __pyx_k_holder, sizeof(__pyx_k_holder), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_io, __pyx_k_io, sizeof(__pyx_k_io), 0, 0, 1, 1},
  {&__pyx_n_s_length, __pyx_k_length, sizeof(__pyx_k_length), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_offset, __pyx_k_offset, sizeof(__pyx_k_offset), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_property, __pyx_k_property, sizeof(__pyx_k_property), 0, 0, 1, 1},
  {&__pyx_n_s_pygame_sdl2_compat, __pyx_k_pygame_sdl2_compat, sizeof(__pyx_k_pygame_sdl2_compat), 0, 0, 1, 1},
  {&__pyx_n_s_pygame_sdl2_rwobject, __pyx_k_pygame_sdl2_rwobject, sizeof(__pyx_k_pygame_sdl2_rwobject), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 1, 1},
  {&__pyx_kp_s_r_is_not_a_filename_or_file_lik, __pyx_k_r_is_not_a_filename_or_file_lik, sizeof(__pyx_k_r_is_not_a_filename_or_file_lik), 0, 0, 1, 0},
  {&__pyx_n_b_rb, __pyx_k_rb, sizeof(__pyx_k_rb), 0, 0, 0, 1},
  {&__pyx_n_s_rb, __pyx_k_rb, sizeof(__pyx_k_rb), 0, 0, 1, 1},
  {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
  {&__pyx_n_s_readable, __pyx_k_readable, sizeof(__pyx_k_readable), 0, 0, 1, 1},
  {&__pyx_n_s_readinto, __pyx_k_readinto, sizeof(__pyx_k_readinto), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_rv, __pyx_k_rv, sizeof(__pyx_k_rv), 0, 0, 1, 1},
  {&__pyx_n_s_rw, __pyx_k_rw, sizeof(__pyx_k_rw), 0, 0, 1, 1},
  {&__pyx_n_s_rw_in, __pyx_k_rw_in, sizeof(__pyx_k_rw_in), 0, 0, 1, 1},
  {&__pyx_n_s_rw_in_object, __pyx_k_rw_in_object, sizeof(__pyx_k_rw_in_object), 0, 0, 1, 1},
  {&__pyx_n_s_seek, __pyx_k_seek, sizeof(__pyx_k_seek), 0, 0, 1, 1},
  {&__pyx_n_s_seekable, __pyx_k_seekable, sizeof(__pyx_k_seekable), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_kp_s_self_ops_cannot_be_converted_to, __pyx_k_self_ops_cannot_be_converted_to, sizeof(__pyx_k_self_ops_cannot_be_converted_to), 0, 0, 1, 0},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_sf, __pyx_k_sf, sizeof(__pyx_k_sf), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_k_src_pygame_sdl2_rwobject_pyx, sizeof(__pyx_k_src_pygame_sdl2_rwobject_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_tell, __pyx_k_tell, sizeof(__pyx_k_tell), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_truncate, __pyx_k_truncate, sizeof(__pyx_k_truncate), 0, 0, 1, 1},
  {&__pyx_n_s_unicode, __pyx_k_unicode, sizeof(__pyx_k_unicode), 0, 0, 1, 1},
  {&__pyx_n_s_unknown, __pyx_k_unknown, sizeof(__pyx_k_unknown), 0, 0, 1, 1},
  {&__pyx_kp_s_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 0, 1, 0},
  {&__pyx_n_s_view, __pyx_k_view, sizeof(__pyx_k_view), 0, 0, 1, 1},
  {&__pyx_n_s_whence, __pyx_k_whence, sizeof(__pyx_k_whence), 0, 0, 1, 1},
  {&__pyx_n_s_writable, __pyx_k_writable, sizeof(__pyx_k_writable), 0, 0, 1, 1},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_property = __Pyx_GetBuiltinName(__pyx_n_s_property); if (!__pyx_builtin_property) __PYX_ERR(0, 480, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 283, __pyx_L1_error)
  __pyx_builtin_IOError = __Pyx_GetBuiltinName(__pyx_n_s_IOError); if (!__pyx_builtin_IOError) __PYX_ERR(0, 308, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 351, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_OSError = __Pyx_GetBuiltinName(__pyx_n_s_OSError); if (!__pyx_builtin_OSError) __PYX_ERR(0, 485, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "pygame_sdl2/rwobject.pyx":72
 *     try:
 *         cur = f.tell()
 *         f.seek(0, 2)             # <<<<<<<<<<<<<<
 *         rv = f.tell()
 *         f.seek(cur, 0)
 */
  __pyx_tuple_ = PyTuple_Pack(2, __pyx_int_0, __pyx_int_2); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "pygame_sdl2/rwobject.pyx":283
 *         rv = (<RWopsImpl>filelike._holder).get_rwops()
 *         if rv == NULL:
 *             raise ValueError("Passed in RWops object is closed")             # <<<<<<<<<<<<<<
 *         (<RWopsImpl>filelike._holder).clear_rwops()
 *         return rv
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_Passed_in_RWops_object_is_closed); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "pygame_sdl2/rwobject.pyx":390
 *         """
 *         if self.closed != 0:
 *             raise ValueError("I/O operation on closed RWops.")             # <<<<<<<<<<<<<<
 * 
 *     def get_closed(self):
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_I_O_operation_on_closed_RWops); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "pygame_sdl2/rwobject.pyx":440
 * 
 *         if not PyObject_CheckBuffer(b):
 *             raise ValueError("Passed in object does not support buffer protocol")             # <<<<<<<<<<<<<<
 *         try:
 *             PyObject_GetBuffer(b, &view, PyBUF_CONTIG)
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_Passed_in_object_does_not_suppor); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.ops cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.ops cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_self_ops_cannot_be_converted_to); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":4
 *     raise TypeError("self.ops cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.ops cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_self_ops_cannot_be_converted_to); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "pygame_sdl2/rwobject.pyx":568
 *         return rv
 * 
 *     return ValueError("Invalid value passed in for name")             # <<<<<<<<<<<<<<
 * 
 * def RWops_from_file_like_object(filelike):
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_Invalid_value_passed_in_for_name); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "pygame_sdl2/rwobject.pyx":638
 * 
 *     if not isinstance(rw_in_object, RWops):
 *         raise ValueError("Passed in object should be RWops")             # <<<<<<<<<<<<<<
 *     rw_in = (<RWopsImpl>rw_in_object._holder).get_rwops()
 *     if rw_in == NULL:
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_Passed_in_object_should_be_RWops); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "pygame_sdl2/rwobject.pyx":690
 *         raise ValueError("Passed in RWops object is closed")
 *     if rw_in.read != subfile_read:
 *         raise ValueError("Passed in RWops object is not SubFile")             # <<<<<<<<<<<<<<
 *     (<RWopsImpl>rw_in_object._holder).clear_rwops()
 *     sf = <SubFile *>rw_in_object.hidden.unknown.data1
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_Passed_in_RWops_object_is_not_Su); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "pygame_sdl2/rwobject.pyx":470
 * 
 * class RWops(io.RawIOBase):
 *     def __init__(self, name=None):             # <<<<<<<<<<<<<<
 *         io.RawIOBase.__init__(self)
 *         self._holder = RWopsImpl()
 */
  __pyx_tuple__10 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_name); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_codeobj__11 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_init, 470, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__11)) __PYX_ERR(0, 470, __pyx_L1_error)
  __pyx_tuple__12 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "pygame_sdl2/rwobject.pyx":477
 *     # Implemented class: io.IOBase
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         self._holder.close()
 * 
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);
  __pyx_codeobj__14 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__13, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_close, 477, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__14)) __PYX_ERR(0, 477, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":481
 * 
 *     @property
 *     def closed(self):             # <<<<<<<<<<<<<<
 *         return self._holder.get_closed()
 * 
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);
  __pyx_codeobj__16 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__15, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_closed, 481, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__16)) __PYX_ERR(0, 481, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":484
 *         return self._holder.get_closed()
 * 
 *     def fileno(self):             # <<<<<<<<<<<<<<
 *         raise OSError()
 * 
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__17, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_fileno, 484, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(0, 484, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":491
 *     # inherited isatty is used
 * 
 *     def readable(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);
  __pyx_codeobj__20 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__19, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_readable, 491, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__20)) __PYX_ERR(0, 491, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":498
 *     # inherited readlines is used
 * 
 *     def seek(self, offset, whence=0):             # <<<<<<<<<<<<<<
 *         self._holder.check_closed()
 *         return self._holder.seek(offset, whence)
 */
  __pyx_tuple__21 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_offset, __pyx_n_s_whence); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_seek, 498, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(0, 498, __pyx_L1_error)
  __pyx_tuple__23 = PyTuple_Pack(1, ((PyObject *)__pyx_int_0)); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "pygame_sdl2/rwobject.pyx":503
 * 
 * 
 *     def seekable(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_seekable, 503, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 503, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":508
 *     # inherited tell is used
 * 
 *     def truncate(self, size=None):             # <<<<<<<<<<<<<<
 *         raise OSError()
 * 
 */
  __pyx_tuple__26 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_size); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_truncate, 508, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 508, __pyx_L1_error)
  __pyx_tuple__28 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "pygame_sdl2/rwobject.pyx":511
 *         raise OSError()
 * 
 *     def writable(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_writable, 511, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(0, 511, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":524
 *     # inherited readall is used
 * 
 *     def readinto(self, b):             # <<<<<<<<<<<<<<
 *         self._holder.check_closed()
 *         return self._holder.readinto(b)
 */
  __pyx_tuple__31 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_b); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_readinto, 524, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 524, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":528
 *         return self._holder.readinto(b)
 * 
 *     def write(self, b):             # <<<<<<<<<<<<<<
 *         self._holder.check_closed()
 *         return self._holder.write(b)
 */
  __pyx_tuple__33 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_b); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_write, 528, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 528, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":532
 *         return self._holder.write(b)
 * 
 *     def get_sdl_rwops_pointer(self):             # <<<<<<<<<<<<<<
 *         return self._holder.get_sdl_rwops_pointer()
 * 
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_get_sdl_rwops_pointer, 532, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 532, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":535
 *         return self._holder.get_sdl_rwops_pointer()
 * 
 * def RWops_from_file(name, mode="rb"):             # <<<<<<<<<<<<<<
 *     cdef SDL_RWops *rw
 *     cdef char *cname
 */
  __pyx_tuple__37 = PyTuple_Pack(7, __pyx_n_s_name, __pyx_n_s_mode, __pyx_n_s_rw, __pyx_n_s_cname, __pyx_n_s_cmode, __pyx_n_s_dname, __pyx_n_s_rv); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(2, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_RWops_from_file, 535, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 535, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":570
 *     return ValueError("Invalid value passed in for name")
 * 
 * def RWops_from_file_like_object(filelike):             # <<<<<<<<<<<<<<
 *     cdef SDL_RWops *rw
 * 
 */
  __pyx_tuple__39 = PyTuple_Pack(3, __pyx_n_s_filelike, __pyx_n_s_rw, __pyx_n_s_rv); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_RWops_from_file_like_object, 570, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 570, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":580
 *     return rv
 * 
 * def RWops_is_file_like_object(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */
  __pyx_tuple__41 = PyTuple_Pack(4, __pyx_n_s_rw_in_object, __pyx_n_s_view, __pyx_n_s_rw_in, __pyx_n_s_rw); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_RWops_is_file_like_object, 580, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 580, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":592
 *     return rw_in.read == python_read
 * 
 * def RWops_from_buffer(b, mode="rb"):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw
 */
  __pyx_tuple__43 = PyTuple_Pack(6, __pyx_n_s_b, __pyx_n_s_mode, __pyx_n_s_view, __pyx_n_s_rw, __pyx_n_s_bf, __pyx_n_s_rv); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(2, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_RWops_from_buffer, 592, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 592, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":620
 *     return rv
 * 
 * def RWops_is_buffer(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */
  __pyx_tuple__45 = PyTuple_Pack(4, __pyx_n_s_rw_in_object, __pyx_n_s_view, __pyx_n_s_rw_in, __pyx_n_s_rw); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__45, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_RWops_is_buffer, 620, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 620, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":632
 *     return rw_in.read == buffile_read
 * 
 * def RWops_create_subfile(rw_in_object, base, length):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */
  __pyx_tuple__47 = PyTuple_Pack(8, __pyx_n_s_rw_in_object, __pyx_n_s_base, __pyx_n_s_length, __pyx_n_s_view, __pyx_n_s_rw_in, __pyx_n_s_rw, __pyx_n_s_sf, __pyx_n_s_rv); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(3, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_RWops_create_subfile, 632, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 632, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":666
 *     return rv
 * 
 * def RWops_is_subfile(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */
  __pyx_tuple__49 = PyTuple_Pack(4, __pyx_n_s_rw_in_object, __pyx_n_s_view, __pyx_n_s_rw_in, __pyx_n_s_rw); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_RWops_is_subfile, 666, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 666, __pyx_L1_error)

  /* "pygame_sdl2/rwobject.pyx":678
 *     return rw_in.read == subfile_read
 * 
 * def RWops_unwrap_subfile(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */
  __pyx_tuple__51 = PyTuple_Pack(6, __pyx_n_s_rw_in_object, __pyx_n_s_view, __pyx_n_s_rw_in, __pyx_n_s_rw, __pyx_n_s_sf, __pyx_n_s_rv); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__51, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_pygame_sdl2_rwobject_pyx, __pyx_n_s_RWops_unwrap_subfile, 678, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("to_rwops", (void (*)(void))__pyx_f_11pygame_sdl2_8rwobject_to_rwops, "SDL_RWops *(PyObject *, struct __pyx_opt_args_11pygame_sdl2_8rwobject_to_rwops *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("RWopsFromPython", (void (*)(void))__pyx_f_11pygame_sdl2_8rwobject_RWopsFromPython, "SDL_RWops *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_11pygame_sdl2_8rwobject_RWopsImpl = &__pyx_vtable_11pygame_sdl2_8rwobject_RWopsImpl;
  __pyx_vtable_11pygame_sdl2_8rwobject_RWopsImpl.get_rwops = (SDL_RWops *(*)(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *))__pyx_f_11pygame_sdl2_8rwobject_9RWopsImpl_get_rwops;
  __pyx_vtable_11pygame_sdl2_8rwobject_RWopsImpl.set_rwops = (PyObject *(*)(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *, SDL_RWops *))__pyx_f_11pygame_sdl2_8rwobject_9RWopsImpl_set_rwops;
  __pyx_vtable_11pygame_sdl2_8rwobject_RWopsImpl.clear_rwops = (PyObject *(*)(struct __pyx_obj_11pygame_sdl2_8rwobject_RWopsImpl *))__pyx_f_11pygame_sdl2_8rwobject_9RWopsImpl_clear_rwops;
  if (PyType_Ready(&__pyx_type_11pygame_sdl2_8rwobject_RWopsImpl) < 0) __PYX_ERR(0, 381, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_11pygame_sdl2_8rwobject_RWopsImpl.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_11pygame_sdl2_8rwobject_RWopsImpl.tp_dictoffset && __pyx_type_11pygame_sdl2_8rwobject_RWopsImpl.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_11pygame_sdl2_8rwobject_RWopsImpl.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_11pygame_sdl2_8rwobject_RWopsImpl.tp_dict, __pyx_vtabptr_11pygame_sdl2_8rwobject_RWopsImpl) < 0) __PYX_ERR(0, 381, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_RWopsImpl, (PyObject *)&__pyx_type_11pygame_sdl2_8rwobject_RWopsImpl) < 0) __PYX_ERR(0, 381, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_11pygame_sdl2_8rwobject_RWopsImpl) < 0) __PYX_ERR(0, 381, __pyx_L1_error)
  __pyx_ptype_11pygame_sdl2_8rwobject_RWopsImpl = &__pyx_type_11pygame_sdl2_8rwobject_RWopsImpl;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initrwobject(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initrwobject(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_rwobject(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_rwobject(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_rwobject(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'rwobject' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_rwobject(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("rwobject", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_pygame_sdl2__rwobject) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pygame_sdl2.rwobject")) {
      if (unlikely(PyDict_SetItemString(modules, "pygame_sdl2.rwobject", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "pygame_sdl2/rwobject.pyx":27
 * from libc.stdint cimport uintptr_t
 * 
 * from pygame_sdl2.compat import file_type, bytes_, unicode_             # <<<<<<<<<<<<<<
 * 
 * import sys
 */
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_file_type);
  __Pyx_GIVEREF(__pyx_n_s_file_type);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_file_type);
  __Pyx_INCREF(__pyx_n_s_bytes);
  __Pyx_GIVEREF(__pyx_n_s_bytes);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_bytes);
  __Pyx_INCREF(__pyx_n_s_unicode);
  __Pyx_GIVEREF(__pyx_n_s_unicode);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_unicode);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pygame_sdl2_compat, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_file_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_file_type, __pyx_t_1) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_bytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bytes, __pyx_t_1) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unicode, __pyx_t_1) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":29
 * from pygame_sdl2.compat import file_type, bytes_, unicode_
 * 
 * import sys             # <<<<<<<<<<<<<<
 * import io
 * 
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_sys, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_2) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":30
 * 
 * import sys
 * import io             # <<<<<<<<<<<<<<
 * 
 * # The fsencoding.
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_io, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_io, __pyx_t_2) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":33
 * 
 * # The fsencoding.
 * fsencoding = sys.getfilesystemencoding() or "utf-8"             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "SDL.h" nogil:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_getfilesystemencoding); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 33, __pyx_L1_error)
  if (!__pyx_t_4) {
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L2_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_kp_s_utf_8);
  __pyx_t_2 = __pyx_kp_s_utf_8;
  __pyx_L2_bool_binop_done:;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fsencoding, __pyx_t_2) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":469
 *         return ctypes.c_void_p(<uintptr_t> self.ops)
 * 
 * class RWops(io.RawIOBase):             # <<<<<<<<<<<<<<
 *     def __init__(self, name=None):
 *         io.RawIOBase.__init__(self)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_io); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_RawIOBase); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_RWops, __pyx_n_s_RWops, (PyObject *) NULL, __pyx_n_s_pygame_sdl2_rwobject, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pygame_sdl2/rwobject.pyx":470
 * 
 * class RWops(io.RawIOBase):
 *     def __init__(self, name=None):             # <<<<<<<<<<<<<<
 *         io.RawIOBase.__init__(self)
 *         self._holder = RWopsImpl()
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_11pygame_sdl2_8rwobject_5RWops_1__init__, 0, __pyx_n_s_RWops___init, NULL, __pyx_n_s_pygame_sdl2_rwobject, __pyx_d, ((PyObject *)__pyx_codeobj__11)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__12);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_init, __pyx_t_5) < 0) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pygame_sdl2/rwobject.pyx":477
 *     # Implemented class: io.IOBase
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         self._holder.close()
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_11pygame_sdl2_8rwobject_5RWops_3close, 0, __pyx_n_s_RWops_close, NULL, __pyx_n_s_pygame_sdl2_rwobject, __pyx_d, ((PyObject *)__pyx_codeobj__14)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_close, __pyx_t_5) < 0) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pygame_sdl2/rwobject.pyx":481
 * 
 *     @property
 *     def closed(self):             # <<<<<<<<<<<<<<
 *         return self._holder.get_closed()
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_11pygame_sdl2_8rwobject_5RWops_5closed, 0, __pyx_n_s_RWops_closed, NULL, __pyx_n_s_pygame_sdl2_rwobject, __pyx_d, ((PyObject *)__pyx_codeobj__16)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pygame_sdl2/rwobject.pyx":480
 *         self._holder.close()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def closed(self):
 *         return self._holder.get_closed()
 */
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_property, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_closed, __pyx_t_6) < 0) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pygame_sdl2/rwobject.pyx":484
 *         return self._holder.get_closed()
 * 
 *     def fileno(self):             # <<<<<<<<<<<<<<
 *         raise OSError()
 * 
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_11pygame_sdl2_8rwobject_5RWops_7fileno, 0, __pyx_n_s_RWops_fileno, NULL, __pyx_n_s_pygame_sdl2_rwobject, __pyx_d, ((PyObject *)__pyx_codeobj__18)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_fileno, __pyx_t_6) < 0) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pygame_sdl2/rwobject.pyx":491
 *     # inherited isatty is used
 * 
 *     def readable(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_11pygame_sdl2_8rwobject_5RWops_9readable, 0, __pyx_n_s_RWops_readable, NULL, __pyx_n_s_pygame_sdl2_rwobject, __pyx_d, ((PyObject *)__pyx_codeobj__20)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_readable, __pyx_t_6) < 0) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pygame_sdl2/rwobject.pyx":498
 *     # inherited readlines is used
 * 
 *     def seek(self, offset, whence=0):             # <<<<<<<<<<<<<<
 *         self._holder.check_closed()
 *         return self._holder.seek(offset, whence)
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_11pygame_sdl2_8rwobject_5RWops_11seek, 0, __pyx_n_s_RWops_seek, NULL, __pyx_n_s_pygame_sdl2_rwobject, __pyx_d, ((PyObject *)__pyx_codeobj__22)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_6, __pyx_tuple__23);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_seek, __pyx_t_6) < 0) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pygame_sdl2/rwobject.pyx":503
 * 
 * 
 *     def seekable(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_11pygame_sdl2_8rwobject_5RWops_13seekable, 0, __pyx_n_s_RWops_seekable, NULL, __pyx_n_s_pygame_sdl2_rwobject, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_seekable, __pyx_t_6) < 0) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pygame_sdl2/rwobject.pyx":508
 *     # inherited tell is used
 * 
 *     def truncate(self, size=None):             # <<<<<<<<<<<<<<
 *         raise OSError()
 * 
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_11pygame_sdl2_8rwobject_5RWops_15truncate, 0, __pyx_n_s_RWops_truncate, NULL, __pyx_n_s_pygame_sdl2_rwobject, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_6, __pyx_tuple__28);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_truncate, __pyx_t_6) < 0) __PYX_ERR(0, 508, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pygame_sdl2/rwobject.pyx":511
 *         raise OSError()
 * 
 *     def writable(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_11pygame_sdl2_8rwobject_5RWops_17writable, 0, __pyx_n_s_RWops_writable, NULL, __pyx_n_s_pygame_sdl2_rwobject, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_writable, __pyx_t_6) < 0) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pygame_sdl2/rwobject.pyx":524
 *     # inherited readall is used
 * 
 *     def readinto(self, b):             # <<<<<<<<<<<<<<
 *         self._holder.check_closed()
 *         return self._holder.readinto(b)
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_11pygame_sdl2_8rwobject_5RWops_19readinto, 0, __pyx_n_s_RWops_readinto, NULL, __pyx_n_s_pygame_sdl2_rwobject, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_readinto, __pyx_t_6) < 0) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pygame_sdl2/rwobject.pyx":528
 *         return self._holder.readinto(b)
 * 
 *     def write(self, b):             # <<<<<<<<<<<<<<
 *         self._holder.check_closed()
 *         return self._holder.write(b)
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_11pygame_sdl2_8rwobject_5RWops_21write, 0, __pyx_n_s_RWops_write, NULL, __pyx_n_s_pygame_sdl2_rwobject, __pyx_d, ((PyObject *)__pyx_codeobj__34)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_write, __pyx_t_6) < 0) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pygame_sdl2/rwobject.pyx":532
 *         return self._holder.write(b)
 * 
 *     def get_sdl_rwops_pointer(self):             # <<<<<<<<<<<<<<
 *         return self._holder.get_sdl_rwops_pointer()
 * 
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_11pygame_sdl2_8rwobject_5RWops_23get_sdl_rwops_pointer, 0, __pyx_n_s_RWops_get_sdl_rwops_pointer, NULL, __pyx_n_s_pygame_sdl2_rwobject, __pyx_d, ((PyObject *)__pyx_codeobj__36)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_get_sdl_rwops_pointer, __pyx_t_6) < 0) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pygame_sdl2/rwobject.pyx":469
 *         return ctypes.c_void_p(<uintptr_t> self.ops)
 * 
 * class RWops(io.RawIOBase):             # <<<<<<<<<<<<<<
 *     def __init__(self, name=None):
 *         io.RawIOBase.__init__(self)
 */
  __pyx_t_6 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_RWops, __pyx_t_2, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RWops, __pyx_t_6) < 0) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":535
 *         return self._holder.get_sdl_rwops_pointer()
 * 
 * def RWops_from_file(name, mode="rb"):             # <<<<<<<<<<<<<<
 *     cdef SDL_RWops *rw
 *     cdef char *cname
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_11pygame_sdl2_8rwobject_1RWops_from_file, NULL, __pyx_n_s_pygame_sdl2_rwobject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RWops_from_file, __pyx_t_2) < 0) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":570
 *     return ValueError("Invalid value passed in for name")
 * 
 * def RWops_from_file_like_object(filelike):             # <<<<<<<<<<<<<<
 *     cdef SDL_RWops *rw
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_11pygame_sdl2_8rwobject_3RWops_from_file_like_object, NULL, __pyx_n_s_pygame_sdl2_rwobject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RWops_from_file_like_object, __pyx_t_2) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":580
 *     return rv
 * 
 * def RWops_is_file_like_object(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_11pygame_sdl2_8rwobject_5RWops_is_file_like_object, NULL, __pyx_n_s_pygame_sdl2_rwobject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RWops_is_file_like_object, __pyx_t_2) < 0) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":592
 *     return rw_in.read == python_read
 * 
 * def RWops_from_buffer(b, mode="rb"):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_11pygame_sdl2_8rwobject_7RWops_from_buffer, NULL, __pyx_n_s_pygame_sdl2_rwobject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RWops_from_buffer, __pyx_t_2) < 0) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":620
 *     return rv
 * 
 * def RWops_is_buffer(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_11pygame_sdl2_8rwobject_9RWops_is_buffer, NULL, __pyx_n_s_pygame_sdl2_rwobject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RWops_is_buffer, __pyx_t_2) < 0) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":632
 *     return rw_in.read == buffile_read
 * 
 * def RWops_create_subfile(rw_in_object, base, length):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_11pygame_sdl2_8rwobject_11RWops_create_subfile, NULL, __pyx_n_s_pygame_sdl2_rwobject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RWops_create_subfile, __pyx_t_2) < 0) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":666
 *     return rv
 * 
 * def RWops_is_subfile(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_11pygame_sdl2_8rwobject_13RWops_is_subfile, NULL, __pyx_n_s_pygame_sdl2_rwobject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RWops_is_subfile, __pyx_t_2) < 0) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":678
 *     return rw_in.read == subfile_read
 * 
 * def RWops_unwrap_subfile(rw_in_object):             # <<<<<<<<<<<<<<
 *     cdef Py_buffer view
 *     cdef SDL_RWops *rw_in
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_11pygame_sdl2_8rwobject_15RWops_unwrap_subfile, NULL, __pyx_n_s_pygame_sdl2_rwobject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RWops_unwrap_subfile, __pyx_t_2) < 0) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pygame_sdl2/rwobject.pyx":704
 *     return rv
 * 
 * init_python_threads()             # <<<<<<<<<<<<<<
 */
  init_python_threads();

  /* "pygame_sdl2/rwobject.pyx":1
 * # Copyright 2014 Tom Rothamel <tom@rothamel.us>             # <<<<<<<<<<<<<<
 * #
 * # This software is provided 'as-is', without any express or implied
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init pygame_sdl2.rwobject", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pygame_sdl2.rwobject");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* DivInt[ptrdiff_t] */
static CYTHON_INLINE ptrdiff_t __Pyx_div_ptrdiff_t(ptrdiff_t a, ptrdiff_t b) {
    ptrdiff_t q = a / b;
    ptrdiff_t r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int64_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int64_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int64_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int64_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 2 * PyLong_SHIFT) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 3 * PyLong_SHIFT) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 4 * PyLong_SHIFT) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int64_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int64_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int64_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int64_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int64_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int64_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int64_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int64_t) -1;
        }
    } else {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyInt_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
